/* C++ codebase for statistics and numerics

 LICENSE AGREEMENT
 - - - - - - - - -
 All rights reserved.
 Use and redistributions of this code is permitted for commercial and non-commercial purposes,
 under the following conditions:

	* Redistributions must retain the above copyright notice, this list of 
	  conditions and the following disclaimer in the code itself, as well 
	  as in documentation and/or other materials provided with the code.
	* Neither the name of the original author (Stilianos Louca), nor the names 
	  of its contributors may be used to endorse or promote products derived 
	  from this code without specific prior written permission.
	* Proper attribution must be given to the original author, including a 
     reference to the peer-reviewed publication through which the code was published.

 THIS CODE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS" AND ANY 
 EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS CODE, 
 EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 - - - - - - - - - - - - - - - - - - - - - - - - - -

To be used with pybind11
To compile & bind to this codebase, somewhere at the top of your python script use the following code:

	# compile and import CPP codebase
	CPP_codebase_source = "pybind11_example.cpp" # specify the path to the CPP source file
	CPP_codebase_name 	= os.path.splitext(os.path.split(CPP_codebase_source)[1])[0]
	CPP_codebase_binary = CPP_codebase_name + subprocess.check_output(["python3-config","--extension-suffix"])[:-1].decode('utf-8')
	if((not os.path.exists(CPP_codebase_binary)) or (os.path.getmtime(CPP_codebase_binary)<os.path.getmtime(CPP_codebase_source))):
		invoke.run("c++ -O3 -Wall -Werror -shared -std=c++11 -fPIC $(python3 -m pybind11 --includes) %s -o %s -undefined dynamic_lookup"%(CPP_codebase_source,CPP_codebase_binary))
	CPP = importlib.import_module(CPP_codebase_name) #import CPP_codebase as CPP

*/

// required code for pybind11
#include <pybind11/pybind11.h>
#include <pybind11/stl.h>
#include <pybind11/numpy.h>
namespace py = pybind11;

#include <limits>
#include <cmath>
#include <cstdlib>
#include <algorithm>
#include <ctype.h>
#include <string>
#include <iostream>
#include <random>       // for default_random_engine
#include <stdexcept>

#ifndef INFTY_D
#define INFTY_D numeric_limits<double>::infinity()
#endif

#ifndef NAN_D
#define NAN_D std::numeric_limits<double>::quiet_NaN()
#endif

typedef std::vector<double> dvector;
typedef std::vector<long> lvector;
typedef std::vector<bool> bvector;
typedef std::vector<char> cvector;

using namespace std;


#pragma mark -
#pragma mark Functions for debugging
#pragma mark 

// print a 2D matrix (in row-major format) as a python list-of-lists
// Mainly used for debugging
template<class TYPE>
void print_as_python_matrix(const long NR, const long NC, const TYPE A[]){
	cout << "[";
	for(long r=0; r<NR; ++r){
		cout << (r>0 ? ", [" : "[");
		for(long c=0; c<NC; ++c){
			cout << (c>0 ? ", " : "") << A[r*NC+c];
		}
		cout << "]";
	}
	cout << "]\n";
}


template<class ARRAY_TYPE>
void print_as_python_matrix(const long NR, const long NC, const ARRAY_TYPE &A){
	cout << "[";
	for(long r=0; r<NR; ++r){
		cout << (r>0 ? ", [" : "[");
		for(long c=0; c<NC; ++c){
			cout << (c>0 ? ", " : "") << A[r*NC+c];
		}
		cout << "]";
	}
	cout << "]\n";
}

// print a 2D array stored in row-major format
// One of NR or NC may be negative, in which case it is automatically inferred from the remaining information
template<class ARRAY_TYPE>
void print_as_matrix(long NR, long NC, const ARRAY_TYPE &A){
	if(NR<0) NR = long(A.size()/NC);
	else if(NC<0) NC = long(A.size()/NR);
	for(long r=0; r<NR; ++r){
		for(long c=0; c<NC; ++c){
			cout << (c>0 ? ", " : "") << A[r*NC+c];
		}
		cout << "\n";
	}
	cout << "\n";
}


template<class TYPE>
void print_as_matrix(const long NR, const long NC, const TYPE A[]){
	for(long r=0; r<NR; ++r){
		for(long c=0; c<NC; ++c){
			cout << (c>0 ? ", " : "") << A[r*NC+c];
		}
		cout << "\n";
	}
	cout << "\n";
}



template<class TYPE>
void print_inhomogeneous_rows(const vector<TYPE> &rows){
	for(long r=0; r<rows.size(); ++r){
		for(long c=0; c<rows[r].size(); ++c){
			cout << (c>0 ? ", " : "") << rows[r][c];
		}
		cout << "\n";
	}
	cout << "\n";
}



template<class TYPE>
inline bool operator==(const vector<TYPE> &x, const vector<TYPE> &y){
	if(x.size()!=y.size()) return false;
	for(long i=0; i<x.size(); ++i){
		if(x[i]!=y[i]) return false;
	}
	return true;
}



#pragma mark -
#pragma mark Auxiliary functions
#pragma mark 


void seed_random_number_generators(long seed){
	srand(seed);
}


template<class TYPE>
inline TYPE vector_max(const std::vector<TYPE> &X){
	const long N = X.size();
	if(N==0) return NAN_D;
	TYPE maxX = X[0];
	for(long n=0; n<N; ++n){
		if(X[n]>maxX) maxX = X[n];
	}
	return maxX;
}


template<class TYPE>
inline long vector_argmax(const std::vector<TYPE> &X, const bool check_nan=false){
	const long N = X.size();
	if(N==0) return -1;
	long maxi;
	if(check_nan){
		maxi = -1;
		for(long i=0; i<N; ++i){
			if(std::isfinite(X[i]) && ((i<0) || (X[i]>X[maxi]))) maxi = i;
		}
	}else{
		maxi = 0;
		for(long i=1; i<N; ++i){
			if(X[i]>X[maxi]) maxi = i;
		}
	}
	return maxi;
}




template<class TYPE>
inline TYPE vector_mean(const std::vector<TYPE> &X){
	if(X.empty()) return NAN_D;
	const long N = X.size();
	TYPE S = 0;
	for(long i=0; i<N; ++i) S += X[i];
	return (S/double(N));
}


// get the median of a set of values in a quick-and-dirty way
// note that this function modifies the input vector in place
template<class TYPE>
inline TYPE vector_quick_median(std::vector<TYPE> &X){
	if(X.size()==0) return NAN_D;
	else if(X.size()==1) return X[0];
	const long N = X.size()/2;
    std::nth_element(X.begin(), X.begin()+N, X.end());
    return X[N];
}


// calculate the average of all entries in a single row in a matrix
// X[] must be a 2D array of size NR x NC, in row-major format
inline double row_mean(const long NR, const long NC, const dvector &X, const long row){
	if(NC==0) return NAN_D;
	double S = 0;
	for(long k=0; k<NC; ++k) S += X[row*NC+k];
	return (S/NC);
}


// calculate the average of a subset of entries in a single row in a matrix
inline double subrow_mean(	const long NC,				// (INPUT) number of columns in the matrix
							const dvector &X,			// (INPUT) 2D array of size NR x NC, in row-major format
							const long row,				// (INPUT) the focal row, must be between 0 and NR-1
							const lvector &columns){	// (INPUT) sequence of integers from 0 to NC-1, specifying which colums to consider
	if(columns.empty()) return NAN_D;
	double S = 0;
	for(long ki=0; ki<columns.size(); ++ki) S += X[row*NC+columns[ki]];
	return (S/columns.size());
}


// count the number of entries in a single row (or subset of a row) in a matrix that are above a specific threshold
inline long subrow_count_above(	const long NC,				// (INPUT) number of columns in the matrix
								const dvector &X,			// (INPUT) 2D array of size NR x NC, in row-major format
								const long row,				// (INPUT) the focal row, must be between 0 and NR-1
								const lvector &columns,		// (INPUT) sequence of integers from 0 to NC-1, specifying which colums to consider
								const double threshold){	// (INPUT) threshold value
	if(columns.empty()) return NAN_D;
	long N = 0;
	for(long ki=0; ki<columns.size(); ++ki) if(X[row*NC+columns[ki]]>threshold) ++N;
	return N;
}


// calculate the median of a subset of entries in a single row in a matrix
inline double subrow_median(const long NC,				// (INPUT) number of columns in the matrix
							const dvector &X,			// (INPUT) 2D array of size NR x NC, in row-major format
							const long row,				// (INPUT) the focal row, must be between 0 and NR-1
							const lvector &columns){	// (INPUT) sequence of integers from 0 to NC-1, specifying which colums to consider
	if(columns.empty()) return NAN_D;
	dvector subrow(columns.size());
	for(long ki=0; ki<columns.size(); ++ki){
		subrow[ki] = X[row*NC+columns[ki]];
	}
	return vector_quick_median(subrow);
}


// calculate the average of a subset of entries in a matrix, defined by a cross-product of row indices x column indices
inline double submatrix_mean(const long NC,				// (INPUT) number of columns in the matrix
							const dvector &X,			// (INPUT) 2D array of size NR x NC, in row-major format
							const lvector &rows,		// (INPUT) sequence of integers from 0 to NR-1, specifying which rows to consider
							const lvector &columns){	// (INPUT) sequence of integers from 0 to NC-1, specifying which colums to consider
	if(columns.empty() || rows.empty()) return NAN_D;
	double S = 0;
	for(long ri=0, ki; ri<rows.size(); ++ri){
		for(ki=0; ki<columns.size(); ++ki){
			S += X[rows[ri]*NC+columns[ki]];
		}
	}
	return (S/(rows.size()*columns.size()));
}


// count the number of entries in a sub-matrix that are above a threshold
// the submatrix is defined by a cross-product of row indices x column indices
inline long submatrix_count_above(const long NC,		// (INPUT) number of columns in the matrix
							const dvector &X,			// (INPUT) 2D array of size NR x NC, in row-major format
							const lvector &rows,		// (INPUT) sequence of integers from 0 to NR-1, specifying which rows to consider
							const lvector &columns,		// (INPUT) sequence of integers from 0 to NC-1, specifying which colums to consider
							const double  threshold){	
	long N = 0;
	for(long ri=0, ki; ri<rows.size(); ++ri){
		for(ki=0; ki<columns.size(); ++ki){
			if(X[rows[ri]*NC+columns[ki]]>threshold) ++N;
		}
	}
	return N;
}


// adjust the mean of a set of N numbers, for the case where one value is omitted
inline double mean_without1(const long N, const double Xmean, const double Xomitted){
	if(N<=1) return NAN_D;
	return (Xmean - Xomitted/N)*(N/(N-1.0));
}



// given a list of NG integer sets ("groups"), with each set containing non-negative integers from 0 to Nmembers-1 ("members"), construct a corresponding list of length Nmembers mapping each member to the unique group containing it.
// groups[g] should be an integer vector defining the members of group g.
// Note that each member must belong to a single group (i.e., groups cannot overlap), although some groups may not have any members at all.
// If Nmembers is -1, it is automatically inferred from the members of the provided groups.
// The returned member2group will be an integer vector of length Nmembers, with member2group[m] being a value between 0 and NG-1, specifying the group to which member m belongs
void map_members_to_unique_groups(	const std::vector<lvector> &groups, // (INPUT) groups[g] lists the members of the g-th group
									long Nmembers, 						// (INPUT) optional number of members to expect. If <0, this is automatically inferred.
									lvector &member2group){				// (OUTPUT) integer vector of size Nmembers, with values in 0,..,NG-1, specifying the group containing each member
	const long NG = groups.size();
	if(Nmembers<0){
		Nmembers = 0;
		for(long g=0; g<NG; ++g){
			if(groups[g].empty()) continue;
			Nmembers = max(Nmembers, 1+vector_max(groups[g]));
		}
	}

	member2group.resize(Nmembers);
	for(long g=0, m; g<NG; ++g){
		for(m=0; m<groups[g].size(); ++m){
			member2group[groups[g][m]] = g;
		}
	}
}



// container for storing a finite set of items, removing the oldest item if a new item is added past the container's capacity
// internally this is achieved using a simple vector and a "needle" pointing to the last inserted item, with the needle moving upward cyclically, i.e., resetting to 0 if it passes the container's capacity
// the capacity of the container is set upon construction.
template<class TYPE>
class CyclicVector{
private:
	std::vector<TYPE> data;
	long last;
	long capacity;
public:

	CyclicVector(const long _capacity){
		capacity = _capacity;
	}
	
	const long size(){
		return data.size();
	}
	
	// request an item of a certain age, with 0 being the most recently added
	TYPE& get_item(const long age){
		const long N = data.size();
		if(age>=N) throw std::invalid_argument("Requested item age is too high for the number of stored items");
		if(age<=last) return data[last-age];
		else return data[N-(age-last)];
	}

	void add_item(const TYPE& x){
		const long N = data.size();
		if(N<capacity){
			data.push_back(x);
			++last;
		}else{
			// need to effectively remove the oldest item by moving 'last' up (cyclically)
			if(last<N-1){
				++last;
			}else{
				last=0;
			}
			data[last] = x;
		}
	}
	
	const bool contains(const TYPE& x){
		for(long i=0; i<data.size(); ++i){
			if(data[i]==x) return true;
		}
		return false;
	}
};




#pragma mark -
#pragma mark Clustering
#pragma mark 


void standardize_group_mapping(	const lvector 	&member2group,	// (INPUT)
								lvector			&standardized){	// (OUTPUT)
	const long NM = member2group.size();
	const long NG = 1+vector_max(member2group);
	lvector group2first(NG,NM);
	for(long m=0, g; m<NM; ++m){
		g = member2group[m];
		group2first[g] = min(group2first[g], m);
	}
	standardized.resize(NM);
	for(long m=0; m<NM; ++m){
		standardized[m] = group2first[member2group[m]];
	}
}


// determine whether two groupings of the same set of member units are equivalent, modulo reordering of the groups
bool equivalent_groupings(const lvector &member2groupA,		// (INPUT) 1D integer array of length NM, specifying for each member the groupA that it belongs to
						  const lvector &member2groupB){	// (INPUT) 1D integer array of length NM, specifying for each member the groupB that it belongs to
	if(member2groupA.size()!=member2groupB.size()) return false;
	const long NM  = member2groupA.size();
	const long NGA = 1+vector_max(member2groupA);
	const long NGB = 1+vector_max(member2groupB);
	// determine the first member of each group
	lvector group2firstA(NGA,NM);
	lvector group2firstB(NGB,NM);
	for(long m=0, g; m<NM; ++m){
		g = member2groupA[m];
		group2firstA[g] = min(group2firstA[g], m);
	}
	for(long m=0, g; m<NM; ++m){
		g = member2groupB[m];
		group2firstB[g] = min(group2firstB[g], m);
	}
	// determine equivalency of groupings by identifying each group with its first member
	for(long m=0; m<NM; ++m){
		if(group2firstA[member2groupA[m]]!=group2firstB[member2groupB[m]]) return false;
	}
	return true;
}


// Given an affinity matrix (aka. adjacency matrix) between Nnodes nodes, estimate clusters in the network.
// Affinities are interpreted as probabilities that any two nodes are in the same cluster. Hence, affinities must be real numbers between 0 and 1.
py::dict cluster_greedy_refined(const long 		Nnodes,				// (INPUT) number of rows & columns in the affinity matrix
								const dvector 	&affinity_matrix, 	// (INPUT) 2D array of size Nnodes x Nnodes, in row-major format, specifying affinities between 0 and 1.
								const bool		count_only,			// (INPUT) whether to only estimate the number of clusters but not their memberships	
								const long		random_seed,		// (INPUT) seed for the internally used random number generator
								const long		max_iterations,		// (INPUT) maximum number of iterations to perform. Typical values are 10-100.
								const double	affinity_threshold){// (INPUT) number between 0 and 1, affinity threshold above which an item is merged into a group, or two groups are merged. Values >=0.5 are recommended.
	lvector node_order(Nnodes);
	for(long n=0; n<Nnodes; ++n) node_order[n] = n;
	std::default_random_engine RNG(random_seed);
	std::shuffle(node_order.begin(), node_order.end(),RNG); // note that the older std::random_shuffle is deprecated, and also is not influenced by the seed given to srand

	dvector affinities;
	affinities.reserve(Nnodes);
	std::vector<lvector> group_members, new_group_members, merged_groups;
	lvector member2group(Nnodes), standardized_member2group;
	CyclicVector<lvector> past_member_maps(10); // cyclic storage of past member2group maps, for detecting oscillatory behavior where a clustering matches that of some recent previou iteration but not necessarily the last one

	long nearest;
	for(long ni=0, n; ni<Nnodes; ++ni){
		n = node_order[ni];
		// check if this unit matches any of the existing groups (in terms of its average affinity), and add it to the nearest group if applicable
		// affinities[g] will be the average affinity of the focal unit to group g
		affinities.resize(group_members.size());
		for(long g=0; g<group_members.size(); ++g){
			affinities[g] = subrow_mean(Nnodes, affinity_matrix, n, group_members[g]);
		}
		nearest = vector_argmax(affinities, false);
		if((nearest>=0) && (affinities[nearest]>=affinity_threshold)){
			// add this unit to the nearest group
			group_members[nearest].push_back(n);
		}else{
			// none of the existing groups are close to this unit, so create a new group
			group_members.push_back(lvector(1,n));
		}
	}
	map_members_to_unique_groups(group_members, Nnodes, member2group);
	standardize_group_mapping(member2group,standardized_member2group);
	past_member_maps.add_item(standardized_member2group);

	// refine groups, by re-evaluating unit-to-group affinities using the current group configurations, and merging groups that appear close to each other
	long Niterations=0;
	for(long iteration=0; iteration<max_iterations; ++iteration){
		++Niterations;
		new_group_members.clear();
		new_group_members.resize(group_members.size()); // reserve space for revised groups, based on how many groups we got in the first round
		std::shuffle(node_order.begin(), node_order.end(), RNG);
		for(long ni=0, n; ni<Nnodes; ++ni){
			n = node_order[ni];
			affinities.resize(new_group_members.size());
			for(long g=0; g<group_members.size(); ++g){
				affinities[g] = subrow_mean(Nnodes, affinity_matrix, n, group_members[g]);
				if(member2group[n]==g){
					// correct mean to ignore self-comparison
					affinities[g] = mean_without1(group_members[g].size(), affinities[g], 1);				
				}	
			}	
			for(long ng=group_members.size(); ng<new_group_members.size(); ++ng){
				affinities[ng] = subrow_mean(Nnodes, affinity_matrix, n, new_group_members[ng]);
			}
			nearest = vector_argmax(affinities, true);
			if((nearest>=0) && (affinities[nearest]>=affinity_threshold)){
				// add this unit to the nearest group
				new_group_members[nearest].push_back(n);
			}else{
				// none of the existing groups are close to this unit, so create a new group
				new_group_members.push_back(lvector(1,n));
			}
		}
		// replace old groups with new groups, omitting groups that don't have any members left
		group_members.clear();
		group_members.reserve(new_group_members.size());
		for(long ng=0; ng<new_group_members.size(); ++ng){
			if(!new_group_members[ng].empty()) group_members.push_back(new_group_members[ng]);
		}
		// check if we should merge some groups, using an algorithm similar to "greedy_groups", but applied at the group level
		merged_groups.clear();
		merged_groups.reserve(group_members.size());
		for(long g=0; g<group_members.size(); ++g){
			// affinities[mg] will be the average affinity of the focal group to all members in merged_group mg
			affinities.resize(merged_groups.size());
			for(long mg=0; mg<merged_groups.size(); ++mg){
				affinities[mg] = submatrix_mean(Nnodes, affinity_matrix, group_members[g], merged_groups[mg]);
			}
			nearest = vector_argmax(affinities, false);
			if((nearest>=0) && (affinities[nearest]>=affinity_threshold)){
				// add all members of this group to the nearest merged group
				merged_groups[nearest].reserve(merged_groups[nearest].size() + group_members[g].size());
				merged_groups[nearest].insert(merged_groups[nearest].end(), group_members[g].begin(), group_members[g].end());
			}else{
				// none of the existing merged_groups are close to this group, so create a new merged_group
				merged_groups.push_back(group_members[g]);
			}
		}
		group_members = merged_groups;
		map_members_to_unique_groups(group_members, Nnodes, member2group);
		standardize_group_mapping(member2group,standardized_member2group);
		if(past_member_maps.contains(standardized_member2group)) break;
		past_member_maps.add_item(standardized_member2group);
	}

	py::dict results;
	results["Ngroups"] = group_members.size();
	results["Niterations"] = Niterations;
	if(!count_only){
		results["group_members"] = group_members;
	}
	return results;
}



// *********************************
// binding C++ code to python, 
// i.e. exposing a selection of functions to become callable from python
// *********************************


#pragma mark -
#pragma mark Python binding code
#pragma mark 


PYBIND11_MODULE(stats_cpp_codebase, m) {
    m.def("cluster_greedy_refined", &cluster_greedy_refined, py::arg("Nnodes"), py::arg("affinity_matrix"), py::arg("count_only"), py::arg("random_seed"), py::arg("max_iterations"), py::arg("affinity_threshold"));
    m.def("seed_random_number_generators", &seed_random_number_generators, py::arg("seed"));
}

