#!/usr/bin/env python3
#
# This script estimates the number of distinct groups based on pairwise predictor variables (e.g., pairwise similarities/dissimilarities) defined for pairs of units.
# For example, it can be used to estimate the number of genera among a group of MAGs based on pairwise ANIs, AAIs and gene content differences.
# The script uses supervised machine learning to train a model that can classify any two units as "in-group" or "out-group". Based on these classifications, the total number of clusters is estimated.
# The script can accommodate somewhat fuzzy data, i.e., where the machine learning model does not manage to define consistently delineated clusters.
#
# Example usage:
#	./cluster_from_pairwise_data.py -c known_classifications.tsv -p pairwise_similarities*.tsv -r report_file.txt --unit1_column "query_MAG" --unit2_column "reference_MAG" --classification_column "congeneric" -v
# Run:
#	./cluster_from_pairwise_data.py -h
# to see an overview of available options.
#
# Tested using Python 3.11, Mac OS 13.5
#
# Stilianos Louca
# Copyright 2024
# 
# LICENSE AGREEMENT
# - - - - - - - - -
# All rights reserved.
# Use and redistributions of this code is permitted for commercial and non-commercial purposes,
# under the following conditions:
#
#	* Redistributions must retain the above copyright notice, this list of 
#	  conditions and the following disclaimer in the code itself, as well 
#	  as in documentation and/or other materials provided with the code.
#	* Neither the name of the original author (Stilianos Louca), nor the names 
#	  of its contributors may be used to endorse or promote products derived 
#	  from this code without specific prior written permission.
#	* Proper attribution must be given to the original author, including a 
#     reference to the peer-reviewed publication through which the code was published.
#
# THIS CODE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS" AND ANY 
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
# IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS CODE, 
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# - - - - - - - - -


###### IMPORTS #####

import string
import sys, os
import argparse
import numpy
import math
import time
import glob
import fnmatch
import re
import numpy.random
import random
import warnings
import gzip
import sklearn
import ast
import pickle
import itertools
import scipy.cluster
import functools
from sklearn.linear_model import SGDClassifier
from sklearn.neural_network import MLPClassifier
from sklearn.ensemble import RandomForestClassifier
import sklearn.cluster, sklearn.calibration
from joblib import Parallel, delayed
import matplotlib
import matplotlib.pyplot as pyplot
from matplotlib.backends.backend_pdf import PdfPages
from matplotlib.colors import LinearSegmentedColormap
import invoke # calling system commands
import importlib


# compile and import CPP codebase
script_parent_dir 	= os.path.dirname(os.path.realpath(__file__)) # directory containing this script, after eliminating any symlinks
CPP_codebase_source = os.path.join(script_parent_dir,"stats_cpp_codebase.cpp") # specify the path to the CPP source file. Here we assume that it is located next to this script.
CPP_codebase_name 	= os.path.splitext(os.path.split(CPP_codebase_source)[1])[0]
CPP_codebase_binary = os.path.join(script_parent_dir,CPP_codebase_name + ".so")
if((not os.path.exists(CPP_codebase_binary)) or (os.path.getmtime(CPP_codebase_binary)<os.path.getmtime(CPP_codebase_source))):
	compiler_command = "c++ -O3 -Wall -shared -std=c++11 -fPIC $(python3.%d -m pybind11 --includes) %s -o %s"%(sys.version_info.minor,CPP_codebase_source,CPP_codebase_binary)
	if(sys.platform=="darwin"): compiler_command += " -undefined dynamic_lookup"
	invoke.run(compiler_command)
CPP = importlib.import_module(CPP_codebase_name) #import CPP_codebase as CPP


# define special numbers
from numpy import NaN
from numpy import Inf

# disable numpy corner printing
numpy.set_printoptions(threshold=Inf)

# make print function always flush, unless flush=False is explicitly given
print = functools.partial(print, flush=True)

warnings.filterwarnings(action="ignore", message="An input array is constant*")


SCATTER_COLOR_INGROUP  = "#0288db"
SCATTER_COLOR_OUTGROUP = "#960000"
GENERIC_CURVE_COLOR_PALETTE = numpy.array(["#0270b5", "#C70000", "#308C3E", "#916004", "#AD1096", "#159690", "#C44E00", "#2C5430", "#855724", "#e31a1c", "#6a3d9a", "#1f78b4", "#0C064B", "#481800", "#0C064B", "#5A2D00", "#550000", "#651265", "#0093F0", "#4A2816", "#D76C00", "#004141", "#616102", "#600592", "#9C1B1B", "#1608A2", "#250239", "#002F03", "#370037"], dtype=object)


################# Auxiliary #################


# convert an integer to a string, padded with as many zeros as needed for obtaining the same total width up until a specific maximum value
# for example int2zero_padded_str(3,78) yields "03", and int2zero_padded_str(3,1673) yields "0003".
def int2zero_padded_str(value,max_value):
	return ("%0"+str(1+int(math.log10(max(1,abs(max_value)))))+"d")%(value)


def normalize_downward(*values):
	S = sum(values)
	if(S>1):
		return (v/S for v in values)
	else:
		return values


def object2str(x):
	if(isinstance(x, float)):
		return "%.10g"%(x)
	else:
		return str(x)

# compute the ratio of two numbers A/B, properly dealing with cases where B is zero
def ratio_nansafe(A,B):
	if(B!=0):
		return A/B
	elif(A>0):
		return Inf
	elif(A<0):
		return -Inf
	else:
		return NaN


def float_or_nan(s):
    try:
        f = float(s)
        return f
    except ValueError:
        return NaN


numpy_ratio_nansafe = numpy.vectorize(ratio_nansafe)
numpy_object2str 	= numpy.vectorize(object2str,otypes=[object])
numpy_float_or_nan = numpy.vectorize(float_or_nan)


# compute the mean of an array X, properly handling NaNs
def mean_nansafe(X, axis=None, keepdims=False):
	if((axis is None) and (X.size==0)): return NaN
	elif((axis is None) and (not keepdims) and (not numpy.any(numpy.isfinite(X)))): return NaN
	with warnings.catch_warnings():
		warnings.filterwarnings(action='ignore', message='Mean of empty slice')
		warnings.filterwarnings(action='ignore', message='invalid value encountered in double_scalars')
		return numpy.nanmean(X, axis=axis, keepdims=keepdims)
		
# adjust the mean of a set of numbers, for the case where some value are omitted
# Xomitted[] should be a sequence (e.g. list or tuple) of values omitted from the set over which the average is requested
def mean_without(N, Xmean, Xomitted):
	M = len(Xomitted)
	if(N<=M): return NaN
	elif(M==0): return Xmean
	else: return (Xmean - numpy.sum(Xomitted)/N)*(N/(N-M))
	

# join a list/tuple of strings using a separator, omitting any empty ones
def join_nonempty(strings, sep):
	return sep.join(filter(None, strings))


# print a 1D numeric vector in standard layout, as a string
def sprint_vector(A, separator="  "):
	return separator.join(['%g'%(a) for a in A])

# print the shape of a numpy array in the format N1 x N2 x .. Nn
def sprint_shape(array):
	return sprint_vector(array.shape, separator=" x ")


def filebasename(filepath):
	basename = os.path.split(filepath)[1]
	if(basename.lower().endswith(".gz")): basename = basename[:-3]
	basename = os.path.splitext(basename)[0]
	return basename


def invert_permutation(perm):
	inv = numpy.empty_like(perm, dtype=int)
	inv[perm] = numpy.arange(len(perm))
	return inv


# execute a target function multiple times in parallel (if reasonable)
# For fast functions jobs are first grouped into batches to reduce the overhead of dispatching lots of parallel jobs
def run_parallel(function, 			# a callable (e.g., a function), to be executed multiple times. This function should take a single positional argument. If you want to run a function that takes multiple argument, use lambda as a wrapper.
				arguments, 			# a sequence (e.g., list) of length N, where N is the number of times the function should be executed. Each element of arguments[] is the single input argument for a single function call.
				Nthreads,			# positive integer, number of parallel threads to use.
				fast 	 = False,	# whether the function is fast, i.e., it takes a short time to execute a single call compared to the overhead associated with dispatching jobs to processes.
				prefer   = None,	# either None, "processes" or "threads"
				randomize= False):	# randomize the order in which tasks are executed, if parallelized. This may help with load balancing across threads.
	N 	  = len(arguments)
	order = list(range(N))
	if(randomize): random.shuffle(order)
	if((Nthreads>1) and (N>1)):
		if(fast):
			# group jobs into batches
			# for fast functions this can speed up computation by ~ 25%.
			batch_size = max(1,int(0.2*N/Nthreads))
			Nbatches   = math.ceil(N/batch_size)
			def aux_run_batch(b):
				return [function(arguments[order[a]]) for a in range(b*batch_size,min((b+1)*batch_size,N))]
			results = Parallel(n_jobs=min(N,Nthreads), prefer=prefer, verbose=0)(delayed(aux_run_batch)(b) for b in range(Nbatches))
			results = [r for rr in results for r in rr] # flatten results, i.e. unwrap batches
		else:
			# run standard parallelized
			results = Parallel(n_jobs=min(N,Nthreads), prefer=prefer, verbose=0)(delayed(function)(arguments[order[a]]) for a in range(N))
		if(randomize):
			inverse_order = invert_permutation(order)
			results = [results[k] for k in inverse_order]
		return results
	else:
		return list(map(function, arguments))



# find all duplicate values in a list L
# If indices=True, the indices of duplicates (all occurrences except the first one, in order of appearance) will be returned as a list of integers
# If indices=False, the duplicate values are returned as a list (each value is returned at most once).
def find_duplicates(L, indices=False):
	seen = set()
	if(indices):
		return [i for i,x in enumerate(L) if ((x in seen) or seen.add(x))]
	else:
		return sorted(set(x for x in L if ((x in seen) or seen.add(x))))

# given a sequence of integers (e.g. a 1D numpy array), with values in 0,..,NG-1, return a list of length NG, with each g-th element being an integer lists specifying which of the input integers were equal to g (i.e., "belonged to group g").
# For example, ints=[0,2,1,1,0,2,2] would yield [[0,4], [2,3], [1,5,6]]
def group_integers(ints, NG=None):
	if(NG is None): NG = 1+max(ints)
	group_members = [[] for g in range(NG)]
	for i,g in enumerate(ints):
		group_members[g].append(i)
	return group_members
	
	
# given a list of NG integer sets ("groups"), with each set containing non-negative integers from 0 to Nmembers-1 ("members"), construct a corresponding list of length Nmembers mapping each member to the groups containing it.
# groups[g] should be a 1D integer array or integer list defining the members of group g.
# Note that each member could belong to multiple groups, and some groups may not have any members.
# If Nmembers is None, it is automatically inferred from the members of the provided groups.
# If unique==False, then the returned member2groups will be a list of length Nmembers, with member2groups[m] being an integer list (with values in 0,..,NG-1) containing the indices of the groups that containing member m, i.e. each member2groups[m][k] will be such that groups[member2groups[m][k]] contained m.
# If unique==True, then the caller guarantees that each member belongs to exactly one group. In that case the returned member2group will be a 1D integer array of length Nmembers, with member2groups[m] being an integer in 0,..,NG-1, specifying the unique group that contained member m.
# This function is essentially the inverse of group_integers().
def map_members_to_groups(groups, Nmembers=None, unique=False):
	if(Nmembers is None): Nmembers = 1+max((max(group) if (len(group)>0) else 0) for group in groups)
	if(unique):
		member2group = numpy.full(Nmembers,-1)
		for g,group in enumerate(groups):
			member2group[group] = g
		return member2group
	else:
		member2groups = [[] for m in range(Nmembers)]
		for g,group in enumerate(groups):
			for m in group:
				member2groups[m].append(g)
		return member2groups


# given a sequence of mappings X-->Y, where X are non-negative integers, consolidate all mappings into a single mapping X-->Y, i.e., defined over the union of all input X and eliminating any duplications
# The input mappings should be given as a sequence (e.g. list) of 2-tuples, where each tuple is of the form (X,Y), with X & Y being synchronized lists or 1D arrays
# For example, if mappings = [([1,2,3],["a","b","c"]), ([2,5], ["B", "E"])], then this function returns the arrays X=[1,2,3,5], Y=["a","B","c","E"]
# Optionally, one may provide NX, an upper bound for X values (+1) if known beforehand, to save some computation, otherwise it is computed automatically.
def consolidate_mappings(mappings, NX=None):
	if(NX is None): NX = 1+max(max(mapping[0]) for mapping in mappings)
	Y = numpy.empty_like(mappings[0][1], shape=NX)
	mapped = numpy.zeros(NX, dtype=bool)
	for mapping in mappings:
		Y[mapping[0]] = mapping[1]
		mapped[mapping[0]] = True
	X = numpy.where(mapped)[0]
	Y = Y[mapped]
	return X, Y



def get_shell_command():
	arguments = sys.argv;
	arguments = [("'"+a+"'" if (a=="" or re.search(r"[\s*#:;]", a)) else ("'"+'\\\\t'+"'" if (a=='\\t') else a)) for a in arguments];
	return ' '.join(arguments);
	

def get_date_time():
	return time.strftime("%Y.%m.%d") + " " + time.strftime("%H:%M:%S")



# create a random covariance matrix with specific standard deviations, e.g. for debugging purposes
def random_covariance(N, 	# integer, number of desired rows & columns of the covariance matrix
					  std):	# positive number, the standard deviation of each variable, i.e. the square root of each diagonal entry of the returned covariance matrix
	A = numpy.random.normal(loc=0, scale=1, size=(N,N))
	C = A @ A.T
	D = numpy.diag(std/numpy.sqrt(numpy.diag(C)))
	return (D @ C @ D)


# given a 1D numeric numpy array, determine the index of the largest non-nan value
def get_largest_item(values, check_nan=True):
	if(check_nan):
		values = numpy.asarray(values)
		valids = numpy.where(~numpy.isnan(values))[0]
		if(len(valids)==0): return -1
		largest = valids[numpy.argmax(values[valids])]
	elif(len(values)==0):
		return -1
	else:
		largest = numpy.argmax(values)
	return largest


# split a string on a single-character separator such as ':' or ',' or ';', unless it is escaped
def split_escaped(haystack, separator):
	parts = list(re.split(r'(?<!\\)'+separator, haystack)) # don't split at escaped positions
	parts = [part.replace('\\'+separator, separator) for part in parts] # replace escaped separator (e.g. \;) with separator (e.g. ;)
	return parts


# return the set {0,..,N-1} \ subset
# subset may be a slice object or a 1D array/list of integers or a single integer
def integer_complement(N, subset):
	keep = numpy.ones(N, dtype=bool)
	keep[subset,...] = False # use ellipsis at the end to accommodate situations where subset is a tuple
	return numpy.nonzero(keep)[0]


# compute the union of one or more 1D non-negative integer arrays or slices or lists, and return it as a 1D sorted integer array.
# arrays[] must be an exhaustable/finite iterator, yielding 1D integer arrays or integer lists or slices.
# N must be at least as large as the maximum integer in arrays[].
def integer_union(N, arrays):
	keep = numpy.zeros(N, dtype=bool)
	for array in arrays:
		keep[array] = True
	return numpy.nonzero(keep)[0]
	

# given an (n+2)-dimensional array A of size N1 x N2 x ... x Nn x NR x NR, return the upper triangular part of A.
# Note that here "diagonal", "upper triangular" and "lower triangular" refer exclusively to the last two axes of the array A.
# Hence, all elements A[d1,d2,..,dn,r,c] for which r<=c will be returned.
# If flat=True, then the upper-triangular part of A is returned as a completely flattened 1D array of length N1 * .. * Nn * NR*(NR+1)/2.
# if flat=False, then the upper-triangular part of A is returned as an array of size N1 x .. x Nn x (NR*(NR+1)/2), i.e., only the last two axes of A are flattened.
def extract_upper_triangular(A, flat=False):
	NR = A.shape[-1]
	if(A.ndim==2):
		return A[numpy.triu_indices(NR)]
	else:
		Ashape = A.shape # keep a record of the original shape of A
		if(A.ndim>3): A = A.reshape((numpy.prod(A.shape[:-2]),NR,NR)) # first flatten all but the last 2 axes
		upper_indices = numpy.triu_indices(NR)
		F = [A[l,...][upper_indices] for l in range(A.shape[0])]
		if(flat): F = numpy.concatenate(F)
		else: F = numpy.asarray(F).reshape(A.shape[:-2]+(int(NR*(NR+1)/2),))
		return F

# OLD CODE. NOT USED
# # randomly select a training subset, validation subset and test subset from a larger dataset
# def choose_tvt_subsets(	Nsamples, 
# 						training_fraction,
# 						validation_fraction,
# 						test_fraction,
# 						verbose,
# 						verbose_prefix):
# 	if(verbose): print2("%sSplitting data set into training+validation+test subsets.."%(verbose_prefix))
# 	training_samples 		= RNG.choice(Nsamples, size=int(training_fraction*Nsamples), replace=False)
# 	non_training_samples 	= integer_complement(Nsamples, training_samples)
# 	validation_samples 		= RNG.choice(non_training_samples, size=min(len(non_training_samples),int(training_fraction*Nsamples)), replace=False)
# 	if(test_fraction+training_fraction>1):
# 		# also include some samples from the training set
# 		if(verbose): print2("%sWARNING: Some training samples are also used as validation samples, since test_fraction+training_fraction>1"%(verbose_prefix))
# 		validation_samples = numpy.concatenate((validation_samples, RNG.choice(training_samples, size=(int(training_fraction*Nsamples)-len(validation_samples)), replace=False)))
# 	training_validation_samples 	= integer_union(Nsamples, [training_samples, validation_samples])
# 	non_training_validation_samples = integer_complement(Nsamples, training_validation_samples)
# 	test_samples 					= RNG.choice(non_training_validation_samples, size=min(len(non_training_validation_samples),int(test_fraction*Nsamples)), replace=False)
# 	if(len(test_samples)<int(test_fraction*Nsamples)):
# 		# also include some samples from the training/validation sets
# 		if(verbose): print2("%sWARNING: Some training or validation samples are also used as test samples. To avoid this, consider ensuring that test_fraction+training_fraction+test_fraction<=1."%(verbose_prefix))
# 		test_samples = numpy.concatenate((test_samples, RNG.choice(training_validation_samples, size=(int(test_fraction*Nsamples)-len(test_samples)), replace=False)))
# 	if(verbose): print2("%s  Note: Considering %d training samples, %d validation samples and %d test samples"%(verbose_prefix,len(training_samples),len(validation_samples),len(test_samples)))
# 	return training_samples, validation_samples, test_samples
# 
# 
# # randomly select a training subset and validation subset from a larger dataset
# def choose_tv_subsets(	Nsamples, 
# 						training_fraction,
# 						validation_fraction,
# 						verbose,
# 						verbose_prefix):
# 	if(verbose): print2("%sSplitting data set into training+validation subsets.."%(verbose_prefix))
# 	training_samples 		= RNG.choice(Nsamples, size=int(training_fraction*Nsamples), replace=False)
# 	non_training_samples 	= integer_complement(Nsamples, training_samples)
# 	validation_samples 		= RNG.choice(non_training_samples, size=min(len(non_training_samples),int(training_fraction*Nsamples)), replace=False)
# 	if(test_fraction+training_fraction>1):
# 		# also include some samples from the training set
# 		if(verbose): print2("%sWARNING: Some training samples are also used as validation samples, since test_fraction+training_fraction>1"%(verbose_prefix))
# 		validation_samples = numpy.concatenate((validation_samples, RNG.choice(training_samples, size=(int(training_fraction*Nsamples)-len(validation_samples)), replace=False)))
# 	if(verbose): print2("%s  Note: Considering %d training samples & %d validation samples"%(verbose_prefix,len(training_samples),len(validation_samples)))
# 	return training_samples, validation_samples


# return a list of indices corresponding to items to keep
# this function preserves the order in items_to_keep[], or of names[] if (items_to_keep==None)
def filter_name_list(	names,
						start_pool 		= None,		# starting pool to further filter. This will typically be range(len(names)), but can also be a subset thereof in case of multiple filtering. Can also be None (equivalent to range(len(names)))
						only_names		= None, 	# optional comma-separated list of names (or wildcards) to keep, or a list/set of such names/wildcards. If None or "" or [], this filter is ignored.
						omit_names		= None,		# optional comma-separated list of names (or wildcards) to omit, or a list/set of such names/wildcards. If None or "" or [], this filter is ignored.
						case_sensitive	= False,
						allow_wildcards	= True):
	items_to_keep = (list(range(len(names))) if (start_pool is None) else start_pool)
	if(not case_sensitive): names = [name.lower() for name in names];
	if((only_names is not None) and (len(only_names)>0)):
		if(not (isinstance(only_names,list) or isinstance(only_names,set) or isinstance(only_names,numpy.ndarray))): only_names = list(filter(len, only_names.split(',')))
		if(not case_sensitive): only_names = [name.lower() for name in only_names]
		if(allow_wildcards):
			items_to_keep = [i for i in items_to_keep if (next((1 for w in only_names if fnmatch.fnmatchcase(names[i],w)),-1)>=0)]
		else:
			only_names = set(only_names)
			items_to_keep = [i for i in items_to_keep if (names[i] in only_names)]
	if((omit_names is not None) and (len(omit_names)>0)):
		if(not (isinstance(omit_names,list) or isinstance(omit_names,set) or isinstance(omit_names,numpy.ndarray))): omit_names = list(filter(len, omit_names.split(',')))
		if(not case_sensitive): omit_names = [name.lower() for name in omit_names];
		if(allow_wildcards):
			items_to_keep = [i for i in items_to_keep if (next((1 for w in omit_names if fnmatch.fnmatchcase(names[i],w)),-1)<0)]
		else:
			omit_names = set(omit_names)
			items_to_keep = [i for i in items_to_keep if (names[i] not in omit_names)]
	return numpy.array(items_to_keep, dtype=int)




##############################################################
# improved dictionary, with more flexible syntax for assigning & querying members.
# Use e.g., as:
# mydict = IDict(x=10, y=20, city="Vancouver")
# z = mydict.x + mydict.y
# print(mydict.city)
# print(hasattr(mydict, "city")) # this will print True
# Since it is a subclass of the standard dict, one can also use all of classical dict syntax, for example:
# print(mydict.keys())
# print(mydict.values())
# print(mydict["city"])
# print(mydict.get("town", None))
# print("city" in mydict) # this will print True
# One can even pass a IDict as a dictionary of arguments to a function, for example:
# myfunction(**mydict)

class IDict(dict):
	__delattr__ = dict.__delitem__

	# add or modify member elements in-place, and return self
	# this replaces the standard dict.update() method, the only difference being that it returns self instead of None.
	def update(self,**kwargs):
		super().update(kwargs)
		return self

	# function for creating a shallow copy of the IDict instance and adding (or modifying) one or more member elements
	# use for example as:
	#	newdict = olddict.copy_and_update(nearest_town="Whistler")
	#	print(newdict.nearest_town)
	# Note that copying is not deep, i.e., the values stored in the parent & child may both still refer to the same objects.
	def copy_and_update(self,**kwargs):
		new = copy.copy(self)
		new.update(**kwargs)
		return new

	# enable retrieving a member value using the syntax <my_named_list>.<key>
	# standard dictionary attributes take precedence over stored keys, so you should avoid storing keys likely to be standard dict attributes
	def __getattr__(self, key):
# 		if((key=="__array_struct__") or (key=="__array_interface__") or (key=="__array__")):
# 			# explicitly tell numpy that these attributes do not exist, so that it knows the dict is not an array
# 			raise AttributeError
# 		try:
# 			return dict.__getattribute__(self, key)
# 		except:
# 			return self[key]
		if(key.startswith("__")): return dict.__getattribute__(self, key)
		try:
			return dict.__getattribute__(self, key)
		except:
			return self[key]


	# enable setting a member value using the syntax <my_named_list>.<key> = <value>
# 	def __setattr__(self, key, value):
# 		self[key] = value
	__setattr__ = dict.__setitem__


#####################################################
# Checking output files


	
def check_output_file(file,force,verbose,verbose_prefix):
	if(os.path.exists(file)):
		if(force): 
			if(os.path.isdir(file)): shutil.rmtree(file)
			else: os.remove(file)
			if(verbose): print(("%sNote: Replacing output file '%s'" % (verbose_prefix,file)))
		else:
			print(("%sERROR: Output file '%s' already exists\n%s       Cowardly refusing to continue.\n%s       Use --force to ignore this message" % (verbose_prefix,file,verbose_prefix,verbose_prefix)))
			sys.exit(1)
	else:
		parent = os.path.dirname(file);
		if((parent!="") and (not os.path.exists(parent))): os.makedirs(parent)
	
	
def check_output_file_append(file,force,verbose,verbose_prefix):
	if(os.path.exists(file)):
		if(os.path.isdir(file)):
			if(force): 
				shutil.rmtree(file)
				if(verbose): print(("%sNote: Replacing directory '%s' with an output file" % (verbose_prefix,file)))
			else:
				print(("%sERROR: Output file '%s' already exists as a directory\n%s       Cowardly refusing to continue.\n%s       Use --force to ignore this message" % (verbose_prefix,file,verbose_prefix,verbose_prefix)))
				sys.exit(1)
	else:
		parent = os.path.dirname(file);
		if((parent!="") and (not os.path.exists(parent))): os.makedirs(parent)
			
			
			
def check_output_dir(dir,force,verbose,verbose_prefix):
	if(os.path.isfile(dir)):
		if(args.force): 
			shutil.rmtree(dir)
			if(verbose): print(("%sNote: Replacing file '%s' with output directory" % (verbose_prefix,dir)))
		else:
			print(("%sERROR: Output directory '%s' is a file.\n%s       Cowardly refusing to continue.\n%s       Use --force to avoid this message" % (verbose_prefix,dir,verbose_prefix,verbose_prefix)))
			sys.exit(1)
	elif(not os.path.exists(dir)): os.makedirs(dir)



###############################################
# FILE IO


# save an object as a pickled binary file
# This is essentially a wrapper around pickle.dump, but using the parallelized pigz tool for gzip-compression and ensuring existence of the parent directory
def save_object(obj, filepath, compression_level=6):
	parent = os.path.dirname(filepath)
	if(parent!=""): os.makedirs(parent, exist_ok=True)
	if(filepath.lower().endswith(".gz")):
		temp_filepath = filepath[:-3]
		with open(temp_filepath, 'wb') as fout:
			pickle.dump(obj, fout)
		os.system("pigz -%d -c '%s' > '%s'; rm -f '%s'"%(compression_level,temp_filepath,filepath,temp_filepath))
	else:
		with open(filepath, 'wb') as fout:
			pickle.dump(obj, fout)


# load an object from a pickled binary file
# This is essentially a wrapper around pickle.load that can handle remote locations.
# If graceful==False, and filepath is None or the input file is missing or the checksum is inconsistent, an error is raised.
# If graceful==True, and filepath is None or the input file is missing, then default_value is returned.
def load_object(filepath, 				# local or remote filepath. If a remote path, it must be in the form <username>@<hostname or IP>:<remote path>
				graceful		= True, 
				default_value	= None, 
				checksum		= None):
	if(filepath is None):
		if(graceful): return default_value	
		else: abort("ERROR: Missing file path to load")
	elif(not os.path.exists(filepath)):
		if(graceful): return default_value
		else: abort("ERROR: Missing pickle file '%s'"%(filepath))
	try:
		with open_file(filepath, 'rb') as fin:
			obj = pickle.load(file=fin)
	except Exception as e:
		if(graceful): return default_value
		else: abort("ERROR: Failed to load pickle file '%s': %s"%(filepath,e))
	if((checksum is not None) and ((not hasattr(obj,"checksum")) or (obj.checksum!=checksum))):
		if(graceful): return default_value
		else: abort("ERROR: Loaded object checksum differs from expectation")
	return obj


def extract_column_names(line,delimiter,respect_double_quotes_in_column_names):
	if(respect_double_quotes_in_column_names):
		column_names = split_while_preserving_quotes(line,delimiter,['"'])
	else:
		column_names = (line.split() if (delimiter=="") else line.split(delimiter))
		column_names = [name.strip() for name in column_names] # remove flanking whitespace
	return column_names


# open a file for reading/writing, supporting optionally gzip compression
# mode can be 'wt', 'rt' or other standard modes
def open_file(filepath, mode, compresslevel=6):
	if(mode.startswith("w")): 
		parent = os.path.dirname(filepath)
		if(parent!=""): os.makedirs(parent, exist_ok=True)
	return (gzip.open(filepath,mode=mode,compresslevel=compresslevel) if filepath.lower().endswith(".gz") else open(filepath,mode))


def split_comments(line, comment_prefix):
	if(comment_prefix==""): return [line];
	pos = line.find(comment_prefix);
	if(pos<0): return [line];
	return [line[0:pos], line[pos+len(comment_prefix):]];



def get_column_index(column_names, has_column_names, target_column, case_sensitive):
	if(has_column_names):
		# columns have actual names, so search for the target column by name
		if(not case_sensitive): 
			target_column = target_column.lower()
			column_names = [column_name.lower() for column_name in column_names]
		column_index = next((c for c in range(len(column_names)) if (column_names[c]==target_column)), -1);
	else:
		# no column names available, or column names are just column indices
		column_index = int(target_column);
	return column_index


def read_table(	filepath,						# path to classical input table
				delimiter, 						# column delimiter for classical tables. If empty (""), delimiters are elastic whitespace.
				comment_prefix, 				# comment prefix for classical tables.
				column_names_are_in, 			# Only relevant for classical tables. Can be 'none', 'last_comment_line' or 'first_data_line'. Only applies to classical tables.
				only_column_names	= None,		# optional comma-separated list of column names (or wildcards) to keep, or a list of such names or wildcards
				omit_column_names	= None,		# optional comma-separated list of column names (or wildcards) to omit, or a list of such names or wildcards
				verbose				= False,
				verbose_prefix		= ""):
	with open_file(filepath, 'rt') as fin:
		linecount			= 0
		rowcount			= 0
		last_comment_line	= None
		column_names 		= None
		table				= []
		found_column_names 	= False
		keep_columns		= None
		include_row			= True
		for line in fin.readlines():
			linecount += 1
			if(line[-1]=="\n"): line = line[0:-1] # remove trailing newline char
			parts = split_comments(line, comment_prefix)
			if((rowcount==0) and (len(parts)>1) and (parts[0].strip()=='')): 
				# pure comment line in header
				last_comment_line = parts[1];
				last_comment_line_count = linecount;
			line = parts[0] # ignore comments
			if(line == ""): continue
		
			# this line seems to be a data line
			# see if we should interpret it (or the last comment line) as a header
			include_row = True
			if(column_names is None):
				if((rowcount==0) and (not (last_comment_line is None)) and (column_names_are_in=='last_comment_line')):
					# interpret last comment line as header
					column_names = (last_comment_line.split() if (delimiter=="") else last_comment_line.split(delimiter))
					column_names = [name.strip() for name in column_names] # remove flanking whitespace
					found_column_names = True
				elif((rowcount==0) and (column_names_are_in=='first_data_line')):
					# interpret data line as header
					column_names = (line.split() if (delimiter=="") else line.split(delimiter))
					column_names = [name.strip() for name in column_names] # remove flanking whitespace
					found_column_names = True
					include_row = False
				else:
					# set column names to column indices
					column_names = [str(c) for c in range(len(line.split() if (delimiter=="") else line.split(delimiter)))]
				if(((only_column_names is not None) and (only_column_names!="")) or ((omit_column_names is not None) and (omit_column_names!=""))):
					keep_columns = filter_name_list(names 			= column_names,
													only_names		= only_column_names,
													omit_names		= omit_column_names,
													case_sensitive	= False,
													allow_wildcards	= True)
					if(len(keep_columns)==0): raise RuntimeError("All columns from the input table were filtered out")
			if(not include_row): continue
			
			# extract data from this data line
			rowcount+=1
			parts = (line.split(delimiter) if (delimiter!="") else line.split())
			if((column_names is not None) and (len(parts)!=len(column_names))):
				print2("%sERROR: Number of entries (%d) in line %d is inconsistent with number of columns (%d), in file '%s'" % (verbose_prefix,len(parts),linecount,len(column_names),filepath))
				sys.exit(1)								
			table.append(parts if (keep_columns is None) else [parts[k] for k in keep_columns])
	if(keep_columns is not None): column_names = [column_names[k] for k in keep_columns]
	table = numpy.array(table, dtype=object)
	if(verbose): print2("%sLoaded %d rows and %d columns from %d lines in file '%s'" % (verbose_prefix, rowcount, (0 if (column_names is None) else len(column_names)), linecount, filepath))	
	return table, numpy.array(column_names, dtype=str), found_column_names
	

def load_names_from_table(file, delimiter, comment_prefix, header=False):
	with open_file(file,'rt') as fin:
		lines = list(filter(len,[split_comments(line,comment_prefix)[0] for line in fin]))
		if(header): lines = lines[1:] # omit the first data line, as it contains the table header
		names = [(line.split(None,1) if (delimiter=="") else line.split(delimiter,1))[0] for line in lines]
	names = list(filter(len, [name.strip() for name in names])) # remove flanking whitespace and newlines, and omit empty names
	return names


def adjust_predictor_name(filepath, column_name, column_index):
	if(args.predictor_naming=="column_name"): return column_name
	elif(args.predictor_naming=="filebasename.column_name"): return filebasename(filepath)+"."+column_name
	elif(args.predictor_naming=="filebasename.number"): return filebasename(filepath)+"."+str(column_index)
	elif(args.predictor_naming=="filebasename"): return filebasename(filepath)


# load pairwise data from a table file, where each input row is a distinct pair of units specified by name
# to conserve space, this function returns these data in the following form:
#	pairs: A list length NP, each element of which is an integer 2-tuple specifying a pair of units
#	values: A 2D array of size NP x NS, each row of which specifies values extracted for a specific pair. The first axis of values[] is synchronized with pairs[].
#	unit_name2int: A dictionary mapping unit names to unique integer indices. These integers are used in pairs[] to more space-efficiently identify units.
def load_flat_pairwise_data(filepath, 
							only_column_names	= "",
							omit_column_names	= "",
							as_predictors		= False,	# whether the loaded data are predictors, in which case they may be renamed
							omit_gappy_rows		= False, 	# omit any rows with missing entries, i.e. where at least one entry is ""
							numeric				= False,
							unit_name2int		= None, # optional dictionary, mapping unit names to integer indices. Units not already represented in this dictionary will be appended, i.e. with a new int, thus modifying this dictionary in place.
							only_unit_names		= None,	# optional set of strings, unit names to restrict to. May also be a dictionary, in which case the keys should be unit names to restrict to.
							omit_unit_names		= None,	# optional set of strings, unit names to omit. May also be a dictionary, in which case the keys should be unit names to omit.
							only_pairs			= None,	# optional set, each element of which is an integer 2-tuple, specifying unit pairs to restrict to. This can also be a dictionary, in which case the keys must be integer 2-tuples. This only makes sense if unit_name2int was provided.
							omit_pairs			= None,	# optional set, each element of which is an integer 2-tuple, specifying unit pairs to omit. This can also be a dictionary, in which case the keys must be integer 2-tuples. This only makes sense if unit_name2int was provided.
							skip_if_no_columns	= False,# if True, and no columns are kept (other than unit1 & unit2 columns), skip the file and return empty arrays
							verbose				= False,
							verbose_prefix		= ""):
	column_names 		= None
	Nlines 				= 0
	found_column_names	= False
	unit1_column		= None
	unit2_column		= None
	values				= []
	pairs				= []
	if(unit_name2int is None): unit_name2int = {}
	with open_file(filepath,'rt') as fin:
		for line in fin:
			Nlines += 1
			skip_row = False
			if(line[-1]=="\n"): line = line[:-1] # remove trailing newline char
			parts = split_comments(line, args.comment_prefix)
			if(column_names is None):
				if((len(pairs)==0) and (len(parts)>1) and (parts[0].strip()=='')): 
					# pure comment line in header
					last_comment_line = parts[1]
					last_comment_line_number = Nlines
			line = parts[0] # ignore comments
			if(line == ""): continue

			# this line seems to be a data line
		
			# figure out column names and indices if needed
			if(column_names is None):
				if(args.column_names_are_in=='last_comment_line'):
					# interpret last comment line as header
					if(last_comment_line is None):
						abort(("%sERROR: Missing at least one comment line with column names, in file '%s'"%(verbose_prefix,filepath)))
					column_names = extract_column_names(last_comment_line,delimiter=args.delimiter,respect_double_quotes_in_column_names=False)
					found_column_names = True
				elif(args.column_names_are_in=='first_data_line'):
					# interpret data line as header
					column_names = extract_column_names(line,delimiter=args.delimiter,respect_double_quotes_in_column_names=False)
					found_column_names = True
					continue
				else:
					# set column names to column indices
					parts = extract_column_names(line,delimiter=args.delimiter,respect_double_quotes_in_column_names=False)
					column_names = [str(c) for c in range(len(parts))]

			if(unit1_column is None):
				# figure out where unit names are and which columns to keep
				unit1_column = get_column_index(column_names, found_column_names, args.unit1_column, args.case_sensitive)
				unit2_column = get_column_index(column_names, found_column_names, args.unit2_column, args.case_sensitive)
				if(unit1_column<0): abort(("%sERROR: Unknown column '%s' specified for unit1 names in file '%s'" % (verbose_prefix,args.unit1_column,filepath)))
				if(unit2_column<0): abort(("%sERROR: Unknown column '%s' specified for unit2 names in file '%s'" % (verbose_prefix,args.unit2_column,filepath)))
				
				if(as_predictors): column_names = [(name if ((c==unit1_column) or (c==unit2_column)) else adjust_predictor_name(filepath, column_name=name, column_index=c)) for c,name in enumerate(column_names)]

				if((len(only_column_names)>0) or (len(omit_column_names)>0)):
					keep_columns = filter_name_list(names 			= column_names,
													only_names		= only_column_names,
													omit_names		= omit_column_names,
													case_sensitive	= False,
													allow_wildcards	= True)
				else:
					keep_columns = range(len(column_names))
				keep_columns = sorted(set(keep_columns).difference([unit1_column, unit2_column])) # leave out columns containing unit1 & unit2 names
				if((len(keep_columns)==0) and skip_if_no_columns): return [], numpy.empty((0,0)), unit_name2int, [], found_column_names

			# parse this data row
			if(verbose and ((Nlines % 100000) == 0)):
				print("%sInspecting input line %d.."%(args.verbose_prefix+verbose_prefix,Nlines), end='\r')
			parts = (line.split(args.delimiter) if (args.delimiter!="") else line.split())
			if(len(parts)<len(column_names)):
				abort(("%sERROR: Need at least %d columns, but only found %d in line %d in file '%s'" % (verbose_prefix,len(column_names),len(parts),Nlines,filepath)))
			unit1_name = parts[unit1_column]
			unit2_name = parts[unit2_column]
			
			# rename units if needed
			if(old2new_unit_name is not None):
				if(args.missing_unit_name_in_map=="error"):
					if(unit1_name not in old2new_unit_name): raise RuntimeError("Old unit name '%s' not found in unit name map"%(unit1_name))
					if(unit2_name not in old2new_unit_name): raise RuntimeError("Old unit name '%s' not found in unit name map"%(unit2_name))
					unit1_name = old2new_unit_name[unit1_name]
					unit2_name = old2new_unit_name[unit2_name]
				else:
					unit1_name = old2new_unit_name.get(unit1_name,unit1_name)
					unit2_name = old2new_unit_name.get(unit2_name,unit2_name)
			
			# check if this row should be omitted
			if((only_unit1_names is not None) and (unit1_name not in only_unit1_names)): continue
			if((omit_unit1_names is not None) and (unit1_name in omit_unit1_names)): continue
			if((only_unit2_names is not None) and (unit2_name not in only_unit2_names)): continue
			if((omit_unit2_names is not None) and (unit2_name in omit_unit2_names)): continue
			if((only_unit_names is not None) and ((unit1_name not in only_unit_names) or (unit2_name not in only_unit_names))): continue
			if((omit_unit_names is not None) and ((unit1_name in only_unit_names) or (unit2_name in only_unit_names))): continue
			if((only_unit_name_pairs is not None) and ((unit1_name,unit2_name) not in only_unit_name_pairs)): continue
			if(omit_gappy_rows and any((parts[k]=="") for k in keep_columns)): continue
			
			# map unit names to integers
			unit1 = unit_name2int.setdefault(unit1_name, len(unit_name2int))
			unit2 = unit_name2int.setdefault(unit2_name, len(unit_name2int))
			pair = (unit1,unit2)
			if((only_pairs is not None) and (pair not in only_pairs)): continue
			if((omit_pairs is not None) and (pair in omit_pairs)): continue

			values.append([parts[k] for k in keep_columns])
			pairs.append(pair)
	values = numpy.asarray(values, dtype=object)
	if(numeric): values = numpy_float_or_nan(values)
	column_names = [column_names[k] for k in keep_columns]
	return pairs, values, unit_name2int, column_names, found_column_names

	

def save_classical_table(	output_file,		# path to output file. If it exists, it will be silently replaced.
							table,				# 2D numpy object or float or string array of size NR x NC
							row_names				= None,		# list of size NR. Can be None (don't save row names)
							column_names			= None, 	# sequence (e.g. list) of row names
							row_names_column		= None, 	# string, column name under which to store row names. Can be None (don't save row names).
							save_column_names_in	= "first_data_line",	# either 'none' or 'last_comment_line' or 'first_data_line'
							delimiter				= "\t",					# e.g. "\t" or " "
							comment_prefix			= "#",
							comments				= ""):	# optional string of comments to include at the top of the table
	if((row_names is not None) and (len(row_names)!=table.shape[0])): raise ValueError("Number of provided row names (%d) differs from number of rows in table (%d)"%(len(row_names), table.shape[0]))
	if((column_names is not None) and (len(column_names)!=table.shape[1])): raise ValueError("Number of provided column names (%d) differs from number of columns in table (%d)"%(len(column_names), table.shape[1]))
	check_output_file(file=output_file, force=True, verbose=False, verbose_prefix=args.verbose_prefix)
	include_row_names = ((row_names is not None) and (row_names_column is not None))
	include_column_names = ((column_names is not None) and (save_column_names_in!="none"))
	if(include_row_names and (column_names is not None)): column_names = [row_names_column] + column_names # add row-names column
	with open_file(output_file,'wt',compresslevel=args.gzip_compression_level) as fout:
		# write any header comments
		if(comments!=""):
			fout.write(comment_prefix + " " + comments.replace("\n","\n"+comment_prefix+" ") + "\n")
		# write column names
		if(include_column_names):
			if(save_column_names_in=="last_comment_line"): fout.write(comment_prefix+" ")
			fout.write(delimiter.join(column_names)+"\n")
		# write table contents
		if((table.dtype.type is numpy.float_) or (table.dtype.type is numpy.cfloat)):
			for r in range(table.shape[0]):
				row_data = delimiter.join("{:.10g}".format(value) for value in table[r,:])
				fout.write((row_names[r]+delimiter if include_row_names else "") + row_data + "\n")
		else:
			table = numpy_object2str(table)
			for r in range(table.shape[0]):
				row_data = delimiter.join(table[r,:])
				fout.write((row_names[r]+delimiter if include_row_names else "") + row_data + "\n")



def read_rename_list_from_table(file,
								old_name_column,	# index (0-based)
								new_name_column,	# index (0-based)
								delimiter,
								comment_prefix,
								header_is,			# can be 'last_comment_line', 'first_data_line' or 'none'
								verbose_prefix):
	# read classical table file
	linecount			= 0
	rowcount			= 0
	num_columns_needed 	= 1 + max(old_name_column,new_name_column)
	skip_next_row		= (header_is=='first_data_line')
	old2new_names		= {}
	with open_file(file,'rt') as fin:
		for line in fin:
			linecount+=1
			if(line[-1]=="\n"): line = line[0:-1] # remove trailing newline char
			line = split_comments(line, comment_prefix)[0]	# ignore comments
			if(line == ""): continue
		
			# this line seems to be a data line
			if(skip_next_row):
				skip_next_row = False
				continue;
			rowcount+=1
		
			parts = (line.split(delimiter) if (delimiter!="") else line.split())
			if(len(parts)<num_columns_needed):
				print2("ERROR in file '%s': Need at least %d columns, but only found %d in line %d" % (file,num_columns_needed,len(parts),linecount))
				sys.exit(1)
			old2new_names[parts[old_name_column]] = parts[new_name_column]
	return old2new_names




# Symmetrize a dataset defined for a series of unit pairs (a1,b1), (a2,b2),..., by replacing redundant unit pairs (a,b) & (b,a) with a single unit pair (min(a,b), max(a,b))
def symmetrize_data(pairs, 	# list of length NP, with each element being an integer tuple (a,b)
					values,	# 2D numpy array of size NP x N1 x .. x Nn, with rows synchronized to pairs[]
					pair2index = None, # optional dictionary mapping integer tuples (a,b) to their integer position in pairs[]. If None, then it is computed internally.
					binary = False): # if True, values[] are supposed to be binary (False/True). In that case, if multiple equivalent unit pairs have different values (e.g., pair (1,4) is False and pair (4,1) is True), these unit pairs will be omitted.
	if(pair2index is None): pair2index = {pair:p for p,pair in enumerate(pairs)}
	keep_pair = numpy.zeros(len(pairs), dtype=int) # 1:keep, 0:undecided, -1:dont keep
	for p1,pair in enumerate(pairs):
		if(keep_pair[p1]!=0): continue # already encountered an equivalent pair earlier
		if(pair[0]==pair[1]):
			keep_pair[p1] = 1
		elif((p2:=pair2index.get((pair[1],pair[0]),-1))>=0):
			if(p2==p1):
				# we got mapped to the same pair, presumably because this pair was already collapsed during a previous symmetrization
				keep_pair[p1] = 1
				continue
			elif(p2<p1): 
				continue # we already handled this pair
			# found the reversed version of this pair
			if(binary and numpy.any(values[p1,...]!=values[p2,...])):
				keep_pair[p1] = -1
				keep_pair[p2] = -1
				continue
			else:
				values[p1,...] 	= 0.5*(values[p1,...]+values[p2,...])
				pairs[p1] 		= (min(pair), max(pair))
				keep_pair[p1] 	= 1
				keep_pair[p2] 	= -1
		else:			
			if(pair[0]>pair[1]): 
				# switch this pair to ensure that pair[0]<=pair[1]
				pairs[p1] = (pair[1],pair[0])
			keep_pair[p1] = 1
	keep_pair = (keep_pair==1)
	if(numpy.sum(keep_pair)<len(pairs)):
		pairs	= [pair for p,pair in enumerate(pairs) if keep_pair[p]]
		values	= values[keep_pair,...]
	# create a new mapping from pairs to indices, such that both (a,b) & (b,a) are mapped to the same index in pairs[] and values[]
	pair2index = {pair:p for p,pair in enumerate(pairs)} | {(pair[1],pair[0]):p for p,pair in enumerate(pairs)}
	return pairs, values, pair2index
		
		
		
		
def replace_nans_in_data(names, 				 # list of length NC, specifying column/variable names
						values,					 # 2D numeric numpy array of size NR x NC
						replacement_specifiers): # semicolon-separated string of replacement specifiers, each of the form <comma-separated list of names/wildcards>:<replacement value if NaN/Inf>
	Nnonfinite = numpy.sum(~numpy.isfinite(values))
	for replacement_specifier in replacement_specifiers.split(";"):
		parts = replacement_specifier.split(":")
		if(len(parts)!=2): abort("ERROR: Invalid replacement specifier '%s'. Expecting two parts separated by a colon"%(replacement_specifier))
		columns = filter_name_list(names=names, only_names=parts[0], case_sensitive=args.case_sensitive, allow_wildcards=True)
		for k in columns:
			values[~numpy.isfinite(values[:,k]),k] = float(parts[1])
	Nreplaced = Nnonfinite-numpy.sum(~numpy.isfinite(values))
	return values, Nreplaced
		
		
def balance_by_rarefying(classifications):
	classification2subset = group_integers(classifications.astype(int), NG=2) # classification2subset[k] is an integer list specifying which samples have classification k.
	min_subset_size 	  = min(len(subset) for subset in classification2subset)
	balanced_samples 	  = numpy.concatenate(tuple(subset if (len(subset)<=min_subset_size) else RNG.choice(subset, size=min_subset_size, replace=False) for subset in classification2subset))
	return balanced_samples



###################################################
# PLOTTING


def rescale_interval(interval, scale, logarithmic):
	if(logarithmic and (min(interval)>0)):
		return numpy.exp(rescale_interval(numpy.log(interval), scale, False))
	else:
		L = interval[1] - interval[0]
		M = (interval[1] + interval[0])/2.0
		interval = [interval[0]-scale*L/2.0, interval[1]+scale*L/2.0]
		if(interval[0]==interval[1]):
			X = interval[0]
			interval = ([0,1] if X==0 else ([X*0.9,X*1.1] if X>0 else [X*1.1,X*0.9]))
		return interval


# save the current pyplot figure as a standalone file (e.g., PDF, PNG, JPG, SVG), or add it as a new page to an existing PDF file
# this function will also "close" the current pyplot figure at the end
def save_plot(	fig, 					# pyplot figure object to plot
				plot_target	= None,		# either a string specifying a file path to which to plot (e.g., as pdf or png or jpg or svg), or an instantiated pdfpages handle. If a string, the file type is determined based on the extension. If a pdfpages object, the plot is added to the PDF and the caller is responsible for closing the PDF afterwards.
				dpi			= 200,		# resolution of the image, in dots per inch. Note that fonts scale with the inch-size of the figure, so a higher DPI means more smooth text for bitmap images. For PDFs and SVGs this parameter is irrelevant. 
				legend 		= None):	# optional legend object to show
	# reduce the DPI if needed, to avoid exceeding the maximum allowed figure size (pixels per axis)
	(width, height) = fig.get_size_inches() * dpi
	dpi = min(dpi, 65535/max(fig.get_size_inches()))
	# save to file, or add as page to a pdfpages object
	bbox_extra_artists = (() if (legend is None) else (legend,))
	if(isinstance(plot_target,str)):
		# plot to a new file
		check_output_file(file=plot_target,force=True,verbose=False,verbose_prefix=args.verbose_prefix)
		pyplot.savefig(plot_target, dpi=dpi, format=os.path.splitext(plot_target.lower())[1][1:], bbox_extra_artists=bbox_extra_artists, bbox_inches='tight')
	elif(isinstance(plot_target, PdfPages)):
		# plot_target is a pdfpages handle, so add page to an existing pdf
		plot_target.savefig(fig, bbox_extra_artists=bbox_extra_artists, bbox_inches='tight')
	else:
		raise ValueError("Unknown plot_target of class '%s': expected either a file path or a pdfpages handle"%(plot_target.__class__))
	pyplot.close()



def char2hex(c):
	a = ord(c);
	if((a>=48) and (a<=57)): return (a-48);
	elif((a>=65) and (a<=90)): return 10+(a-65);
	elif((a>=97) and (a<=122)): return 10+(a-97);
	else: return numpy.NaN;
	

# takes a matplotlib.cm tuple or hex-color spec (e.g. '#E0A3FF') and returns a grey-scale version of the color in the same format as the original
def color_to_grey_scale(C):
	if(isinstance(C, str)):
		# hex string
		return '#'+("%02x"%(int((char2hex(C[1])*16+char2hex(C[2]))*0.21 + (char2hex(C[3])*16+char2hex(C[4]))*0.72 + (char2hex(C[5])*16+char2hex(C[6]))*0.07)))*3
	else:
		# tuple
		return 	(C[0]*0.21 + C[1]*0.72 + C[2]*0.07);


# generate a contour plot from 2-dim rectangular (non-scatter) grid data
def plot_contour_grid(	plot_target,			# either a string, specifying a file path to which to plot (e.g., as PDF, PNG, JPG or SVG), or a pdfPages handle
						X,						# 1D numeric array of length NX, listing the X-values of the grid in ascending order
						Y,						# 1D numeric array of length NY, listing the Y-values of the grid in ascending order
						Z,						# 2D numeric array of size NX x NY, listing the Z-values on the grid
						show_colorbar		= True,			# include a color bar next to the plot
						colormap			= "hot",		# name of the colormap to use, or a colormap object. For a list of options see: https://matplotlib.org/stable/tutorials/colors/colormaps.html
						minX				= None,			# optional minimum X value to show
						maxX				= None,			# optional maximum X value to show
						minY				= None,			# optional minimum Y value to show
						maxY				= None,			# optional maximum Y value to show
						minZ				= None,			# optional minimum Z value covered by the colormap
						maxZ				= None,			# optional maximum Z value covered by the colormap
						color_levels		= "linear:15",	# how to choose color levels. Either a 1D array of color level values, or one of the strings "auto:N", "linear:N", "logarithmic:N", "equal_quantiles:N", where N is the number of desired contour levels.
						contour_values		= None,			# optional numeric list-like object of Z-values at which to draw contour lines, on top of the contour plot.
						curve_color			= None,			# color specifier for the contour lines. If None, contour line colors are chosen automatically.
						xlabel				= "", 			# e.g. 'market capitalization (USD)'
						ylabel				= "",			# e.g. 'close price (USD)'
						clabel				= "",			# label for the colorbar axis, e.g. 'monthly returns'
						plot_title			= "",
						plot_width			= 4,			# numeric, plot width in inches. To convert to pixels (in the case of bitmap images), multiply this by DPI.
						plot_height			= 4,			# numeric, plot height in inches. To convert to pixels (in the case of bitmap images), multiply this by DPI.
						dpi					= 200):			# resolution of the image, in dots per inch. Note that fonts scale with the inch-size of the figure, so a higher DPI means more smooth text for bitmap images. For PDFs this parameter is irrelevant. 
	# basic options checking
	(NX, NY) = len(X), len(Y)
	if((Z.shape[0]!=NX) or (Z.shape[1]!=NY)): raise ValueError("Input Z has invalid shape %s, expected %d x %d"%(sprint_shape(Z), NX, NY))
	if(not numpy.any(numpy.isfinite(Z))):
		return plot_empty(message="No data available for contours", message_color=DEFAULT_PLOT_ERROR_COLOR, xlabel=xlabel, ylabel=ylabel, plot_title=plot_title, plot_width=plot_width, plot_height=plot_height, plot_target=plot_target, dpi=dpi)

	# instantiate a figure
	fig	= pyplot.figure(figsize=(plot_width+1+(1 if show_colorbar else 0.3), plot_height+1))
	ax	= pyplot.gca()

	# determine axis ranges
	if(minX is None): minX = X[0]
	if(maxX is None): maxX = X[-1]
	if(minY is None): minY = Y[0]
	if(maxY is None): maxY = Y[-1]
			
	# determine contour levels
	if(isinstance(color_levels,str)):
		Nlevels	 = int(color_levels.split(":",1)[1])
		cscaling = color_levels.split(":",1)[0]
		if(minZ is None): minZ = numpy.nanmin(Z[(X>=minX) & (X<=maxX),:][:,(Y>=minY) & (Y<=maxY)])
		if(maxZ is None): maxZ = numpy.nanmax(Z[(X>=minX) & (X<=maxX),:][:,(Y>=minY) & (Y<=maxY)])
		if(cscaling=="auto"):
			color_levels = Nlevels
		elif(cscaling=="linear"):
			# define color levels on linear scale
			if(minZ==maxZ):
				color_levels = ([0,1] if (minZ==0) else sorted([minZ,1.1*minZ]))
			else:
				color_levels = numpy.linspace(minZ, maxZ, num=Nlevels)
				if(color_levels[1]-color_levels[0]>1): color_levels = numpy.round(color_levels).astype(int)
			#if(color_levels[-1]<=maxZ): color_levels = numpy.concatenate((color_levels, [color_levels[-1]+color_levels[-1]-color_levels[-2]])) # add one last contour level to avoid cutting off marginal values at the upper boundary (which are otherwise displayed as white)
		elif(cscaling=="logarithmic"):
			# define contour levels on log scale
			if(minZ==maxZ):
				color_levels = ([0,1] if (minZ==0) else sorted([minZ,1.1*minZ]))
			else:
				color_levels = numpy.exp(numpy.linspace(numpy.log(minZ), numpy.log(maxZ), num=Nlevels+1))[1:]
		elif(cscaling=="equal_quantiles"):
			# define contour levels such that each level has similar number of data points
			inliers = numpy.where(numpy.logical_and(Z<=maxZ,Z>=minZ))[0]
			color_levels = sorted(set(numpy.quantile(a=Z[inliers], q=numpy.linspace(0, 1, num=Nlevels+2)[1:-1], overwrite_input=False)))
		
	# plot filled contours
	cplot = pyplot.contourf(X, Y, Z.T, levels = color_levels, cmap = colormap, vmin=minZ, vmax=maxZ, extend='both')
	
	# add colorbar
	if(show_colorbar): cbar = pyplot.colorbar(cplot)

	# remove white lines appearing between contour levels, due to bugs in some PDF renderers
	for c in cplot.collections:
		c.set_edgecolor("face")
		c.set_linewidth(0.5)
		
	# set axis ranges
	pyplot.xlim((minX,maxX))
	pyplot.ylim((minY,maxY))

	# add contour lines if needed
	if((contour_values is not None) and (len(contour_values)>0)):
		cmap = cplot.get_cmap()
		if(curve_color is None): curve_color = [("white" if (color_to_grey_scale(cmap(value))<0.5) else "black") for value in contour_values] # choose contour curve colors according to the contour colors underneath them
		ccurves = pyplot.contour(X, Y, Z.T, levels=contour_values, colors=curve_color, linewidths=0.7, linestyles="dashed")
		pyplot.clabel(ccurves, inline=True, fontsize="small")

	# titles & labels
	pyplot.title(plot_title+"\n", fontsize="medium", fontname="Helvetica")
	pyplot.xlabel(xlabel, fontsize="medium", fontname="Helvetica")
	pyplot.ylabel(ylabel, fontsize="medium", fontname="Helvetica", rotation=90, horizontalalignment="center")
	if(show_colorbar): cbar.set_label(clabel, rotation=270, labelpad=15, fontsize="medium", fontname="Helvetica")

	save_plot(fig=fig, plot_target=plot_target, dpi=dpi, legend=None)





# plot one or more curves (e.g., time series) or scatterpoint sets in a single 2D plot to a PDF file
# The function can handle both numerical times as well as datetimes.
# In terms of output format, this function can either:
#	1. Plot into a single file (e.g., PDF, PNG, SVG etc). To do so, set plot_target to the appropriate file path.
#	2. Plot into a previously instantiated PdfPages object. To do so, set plot_target to an instantiated pdfpages_handle. This approach allows calling this function multiple times, each time adding a page to the PDF.
#	2. Plot into a previously instantiated PlotCollectionHTML object. To do so, set plot_target to an instantiated PlotCollectionHTML with open block. This approach allows calling this function multiple times, each time adding a plot to the HTML.
#	3. Generate and return a base64-RFC4648 string of the plot in SVG format, which can then be embedded e.g. in HTML pages. To do so, set plot_target to None. This option will not generate any actual file, i.e., the plot is returned as a string that encodes the plot as a SVG image.
def plot_xy(plot_target,						# either a string, specifying a file path to which to plot (e.g., as PDF, PNG, JPG or SVG), or a pdfPages handle
			curves				= None, 		# optional list of size Ncurves, each entry of which is a sublist with two elements [X, Y] specifying a separate curve to be plotted. X may be a list/numpy-array of floats or datetimes. Note that the curves object may be modified in-place, so if you want to keep the original version you should pass a copy.
			curve_names			= None,			# optional list of size Ncurves, listing the name of each curve. May also contain None if a curve should not appear in the legend.
			curve_colors		= None,			# list of Hex colors for each curve, e.g. ['#FF0000', '#A030AA']. If the number of required colors is greater than provided, the list is recycled. If None, colors are chosen automatically from a top_combos.
			curve_widths		= [1],			# numeric list specifying the line width of each curve. If shorter than Ncurves, then curve widths are recycled.
			curve_styles		= ["solid"],	# string list specifying the line style of each curve. If shorter than Ncurves, then curve styles are recycled.
			curve_zorders		= None,			# optional integer list, specifying the zorder of each curve. A lower zorder value means the curve is drawn earlier (i.e., below others). If shorter than Ncurves, then zorders are recycled.
			show_curves			= None,			# optional list of booleans of size Ncurves, indicating whether a curve should be plotted. If None, all curves are plotted.
			scatter				= None,			# optional list of size Nscatter, each entry of which is a sublist with two elements [X, Y] specifying a separate scatterdata set to be plotted. Note that the scatter object may be modified in-place, so if you want to keep the original version you should pass a copy.
			scatter_names		= None,			# optional list of size Nscatter, listing the name of each scatterset. May also contain None if a scatterset should not appear in the legend.
			scatter_colors		= None,			# optional list of Hex colors for each scatterset, either by name, e.g. ['#FF0000', '#A030AA']. If the number of required colors is greater than provided, the list is recycled. If None, a color palette is automatically chosen.
			scatter_edge_colors	= None,			# optional list of Hex edge colors for each scatterset, either by name, e.g. ['#FF0000', '#A030AA']. If None, this is automatically chosen based on scatter_colors.
			scatter_sizes		= None,			# opional numeric list specifying the point size of each scatterset. If shorter than Nscatter, then scatter widths are recycled. If None, scatterpoint sizes are determined automatically.
			scatter_zorders		= None,			# optional integer list, specifying the zorder of each scatterset. If shorter than Nscatter, then zorders are recycled.
			show_scatter		= None,			# optional list of booleans of size Nscatter, indicating whether a scatterset should be plotted. If None, all scattersets are plotted.
			reverse_x			= False,		# boolean, specifying if the x-axis should be plotted reversed
			reverse_y			= False,		# boolean, specifying if the y-axis should be plotted reversed
			plot_minx			= None,			# minimum x-value to plot. If None, this is automatically set by the data
			plot_maxx			= None,			# maximum x-value to plot. If None, this is automatically set by the data
			plot_miny			= None,			# optional lower bound for the plots. If None, this is automatically set by the data
			plot_maxy			= None,			# optional upper bound for the plots. If None, this is automatically set by the data
			xlabel				= "x", 			# e.g. 'time (years)'
			ylabel				= "y",			# e.g. 'stock price (USD)'
			logx				= False,
			logy				= False,
			href				= None,			# optional numeric, specifying a horizontal reference line to show (e.g., at value 0). If None, this is ignored.
			vref				= None,			# optional numeric, specifying a vertical reference line to show (e.g., at value 0). If None, this is ignored.
			show_diagonal		= False,		# if True, the diagonal is added to the plot as a reference
			show_diagonal_R2s	= False,		# if True, the R2 (X vs Y) is computed and shown in the title for each scatter set
			show_correlations	= None,			# either None, "Pearson", or "Spearman". If "Pearson" or "Spearman", correlations between X & Y values (separately for each scatter set) will be shown in the figure's subtitle.
			legend_pos			= "auto",		# (string) specification of legened position, e.g. "outside right" or "far outside right". If "none", no legend will be shown. If "auto", this is automatically chosen.
			show_grid			= True,			# whether to show vertical and horizontal grid lines
			plot_title			= "",
			plot_width			= 4,			# numeric, plot width in inches. To convert to pixels (in the case of bitmap images), multiply this by DPI.
			plot_height			= 3,			# numeric, plot height in inches. To convert to pixels (in the case of bitmap images), multiply this by DPI.
			dpi					= 200):			# resolution of the image, in dots per inch. Note that fonts scale with the inch-size of the figure, so a higher DPI means more smooth text for bitmap images. For PDFs this parameter is irrelevant. 
	# basic options checking & reformating
	Ncurves = (0 if (curves is None) else len(curves))
	if(show_curves is None): show_curves = [True]*Ncurves
	if(len(curve_widths)<Ncurves): curve_widths = numpy.repeat(curve_widths, numpy.ceil(Ncurves/len(curve_widths)))[0:Ncurves]
	if(len(curve_styles)<Ncurves): curve_styles = numpy.repeat(curve_styles, numpy.ceil(Ncurves/len(curve_styles)))[0:Ncurves]
	if(curve_colors is None):
		curve_colors = GENERIC_CURVE_COLOR_PALETTE[numpy.mod(numpy.arange(Ncurves),len(GENERIC_CURVE_COLOR_PALETTE))]
	elif(len(curve_colors)<Ncurves):
		curve_colors = numpy.repeat(curve_colors, numpy.ceil(Ncurves/len(curve_colors)))[0:Ncurves]
	if(curve_names is None): 
		curve_names = [None]*Ncurves
	elif(len(curve_names)<Ncurves): 
		curve_names += [None]*(Ncurves-len(curve_names))
	if(curve_zorders is None): 
		curve_zorders = [None]*Ncurves
	elif(len(curve_zorders)<Ncurves):
		curve_zorders = numpy.repeat(curve_zorders, numpy.ceil(Ncurves/len(curve_zorders)))[0:Ncurves]
	Nscatter = (0 if (scatter is None) else len(scatter))
	if(show_scatter is None): show_scatter = [True]*Nscatter
	if(scatter_colors is None):
		scatter_colors = GENERIC_CURVE_COLOR_PALETTE[numpy.mod(numpy.arange(Nscatter),len(GENERIC_CURVE_COLOR_PALETTE))]
	elif(len(scatter_colors)<Nscatter):
		scatter_colors = numpy.tile(scatter_colors, numpy.ceil(Nscatter/len(scatter_colors)))[0:Nscatter]
	if(scatter_edge_colors is None):
		scatter_edge_colors = scatter_colors
	elif(len(scatter_edge_colors)<Nscatter):
		scatter_edge_colors = numpy.tile(scatter_edge_colors, numpy.ceil(Nscatter/len(scatter_edge_colors)))[0:Nscatter]
	if(scatter_names is None): scatter_names = [None]*Nscatter
	if(legend_pos=="auto"):
		legend_pos = ("none" if((sum((x is not None) for x in curve_names)+sum((x is not None) for x in scatter_names))==0) else "outside right")
	if(scatter_zorders is None): 
		scatter_zorders = [None]*Nscatter
	elif(len(scatter_zorders)<Nscatter):
		scatter_zorders = numpy.repeat(scatter_zorders, numpy.ceil(Nscatter/len(scatter_zorders)))[0:Nscatter]
	
	# inspect and sanitize curve data
	for n in range(Ncurves):
		if(len(curves[n][0])==0): continue
		valids = numpy.where(numpy.isfinite(curves[n][0]) & numpy.isfinite(curves[n][1]))[0]
		if(len(valids)<len(curves[n][0])):
			curves[n][0] = numpy.asarray(curves[n][0])[valids]
			curves[n][1] = numpy.asarray(curves[n][1])[valids]

	# inspect and sanitize scatter data
	for n in range(Nscatter):
		if(len(scatter[n][0])==0): continue
		valids = numpy.where(numpy.isfinite(scatter[n][0]) & numpy.isfinite(scatter[n][1]))[0]
		if(len(valids)<len(scatter[n][0])):
			scatter[n][0] = numpy.asarray(scatter[n][0])[valids]
			scatter[n][1] = numpy.asarray(scatter[n][1])[valids]
	
	# determine X & Y ranges
	data_minx = (None if (vref is None) else vref)
	data_maxx = (None if (vref is None) else vref)
	data_miny = (Inf if (href is None) else href)
	data_maxy = (-Inf if (href is None) else href)
	for n in range(Ncurves+Nscatter):
		if(n<Ncurves):
			if(not show_curves[n]): continue
			X = numpy.asarray(curves[n][0])
			Y = numpy.asarray(curves[n][1])
		else:
			if(not show_scatter[n-Ncurves]): continue
			X = numpy.asarray(scatter[n-Ncurves][0])
			Y = numpy.asarray(scatter[n-Ncurves][1])
		if((plot_minx is not None) or (plot_maxx is not None)): Y = Y[numpy.logical_and((True if (plot_minx is None) else (X>=plot_minx)),(True if (plot_maxx is None) else (X<=plot_maxx)))]
		if(len(Y)>0):
			data_minx = (numpy.min(X) if (data_minx is None) else min(data_minx,numpy.min(X)))
			data_maxx = (numpy.max(X) if (data_maxx is None) else max(data_maxx,numpy.max(X)))
			data_miny = min(data_miny,numpy.min(Y))
			data_maxy = max(data_maxy,numpy.max(Y))
	xlim = [data_minx, data_maxx]
	ylim = rescale_interval(interval=[data_miny, data_maxy], scale=0.1, logarithmic=logy)
	if(plot_minx is not None): xlim[0] = plot_minx
	if(plot_maxx is not None): xlim[1] = plot_maxx
	if(plot_miny is not None): ylim[0] = plot_miny
	if(plot_maxy is not None): ylim[1] = plot_maxy

	# determine scatterpoint sizes
	if(Nscatter>0):
		if(scatter_sizes is None):
			Nscatterpoints = numpy.sum([sum(numpy.logical_and(scatter[n][0]>=xlim[0],scatter[n][0]<=xlim[1])) for n in range(Nscatter)])
			scatter_sizes = numpy.full(Nscatter, max(1,100/max(10,math.sqrt(Nscatterpoints))))
		elif(len(scatter_sizes)<Nscatter):
			scatter_sizes = numpy.tile(scatter_sizes, numpy.ceil(Nscatter/len(scatter_sizes)))[0:Nscatter]

	# initialize figure
	bottom_margin	= 0.5 	# inches
	top_margin		= 0.5	# inches
	left_margin		= 1		# inches
	right_margin	= 0.3 	# inches
	total_width		= plot_width+left_margin+right_margin
	total_height	= plot_height+top_margin+bottom_margin
	fig 			= pyplot.figure(figsize=(total_width, total_height))
	ax				= pyplot.gca()

	# show grid lines	
	if(show_grid):
		pyplot.grid(visible=True, which='major', axis='both', alpha=0.2, color="#404040")

	if(show_diagonal and (plot_minx==plot_miny) and (plot_maxx==plot_maxy)):
		# equalize axis limits when showing diagonal as reference
		xlim = [min(xlim[0],ylim[0]),max(xlim[1],ylim[1])]
		ylim = list(xlim)
	 	
	# plot horizontal reference line
	if((href is not None) and numpy.isfinite(href)):
		pyplot.axhline(y=href, color="#404040", alpha=1, linewidth=1, marker="", label="reference")
	
	# plot vertical reference line
	if((vref is not None) and numpy.isfinite(vref)):
		pyplot.axvline(x=vref, color="#404040", alpha=1, linewidth=1, marker="", label="reference")
		
	# plot diagonal reference curve
	if(show_diagonal):
		pyplot.axline(xy1=[0,0], slope=1, color="#606060", alpha=0.8, linewidth=1.2)
		
	# plot curves
	for n in range(Ncurves):
		if(not show_curves[n]): continue
		pyplot.plot(curves[n][0], curves[n][1], c = curve_colors[n], linewidth = curve_widths[n], linestyle=curve_styles[n], label=curve_names[n], zorder=curve_zorders[n])

	# plot scatterdata
	for n in range(Nscatter):
		if(not show_scatter[n]): continue
		pyplot.scatter(	scatter[n][0], 
						scatter[n][1], 
						color		= scatter_colors[n], 
						edgecolor	= scatter_edge_colors[n], 
						s			= scatter_sizes[n], 
						label		= scatter_names[n], 
						marker		= 'o', 
						clip_on		= True, 
						alpha		= 0.8, 
						zorder		= scatter_zorders[n])
		
	# compute correlations and show in subtitle
	if((Nscatter>0) and (show_correlations is not None)):
		if(show_correlations=="Pearson"):
			correlations = [scipy.stats.pearsonr(scatter[n][0], scatter[n][1]) for n in range(Nscatter)]
		elif(show_correlations=="Spearman"):
			correlations = [scipy.stats.spearmanr(scatter[n][0], scatter[n][1]) for n in range(Nscatter)]
		else:
			raise RuntimeError("Invalid correlation type '%s'"%(show_correlations))
		if(Nscatter==1):
			plot_title += "\n%s correlation = %.4g (P = %.3g)"%(show_correlations,correlations[0].statistic,correlations[0].pvalue)
		else:
			plot_title += "\n"+"\n".join("%s correlation for %s = %.4g (P = %.3g)"%(show_correlations,("scatter.%d"%(n) if (scatter_names[n] is None) else scatter_names[n]),correlation.statistic,correlation.pvalue) for n,correlation in enumerate(correlations))
			
	if((Nscatter>0) and show_diagonal_R2s):
		R2 = [(1 - numpy.nanmean((scatter[n][1]-scatter[n][0])**2)/numpy.nanvar(scatter[n][0])) for n in range(Nscatter)]
		if(Nscatter==1):
			plot_title += "\ndiagonal R2 = %.4g"%(R2[0])
		else:
			plot_title += "\n"+"\n".join("diagonal R2 for %s = %.4g"%(("scatter.%d"%(n) if (scatter_names[n] is None) else scatter_names[n]),R2[n]) for n in range(Nscatter))
		
	# plot labels and titles
	pyplot.title(plot_title+"\n",fontsize="medium", fontname="Helvetica")
	pyplot.xlabel(xlabel, fontsize="medium", fontname="Helvetica")
	horizontal_ylabel = (len(ylabel)<=2)
	pyplot.ylabel(ylabel, fontsize="medium", fontname="Helvetica", rotation=(0 if horizontal_ylabel else 90),horizontalalignment=('right' if horizontal_ylabel else "center"))

	# set axis ranges & directions, frame color
	pyplot.xlim(xlim)
	if(numpy.all(numpy.isfinite(ylim))): pyplot.ylim(ylim)	
	if(reverse_x): pyplot.gca().invert_xaxis()
	if(reverse_y): pyplot.gca().invert_yaxis()
	if(logx): pyplot.xscale('log')
	if(logy): pyplot.yscale('log')

	if(legend_pos.endswith("outside right")):
		legend = pyplot.legend(	bbox_to_anchor	= ((1.1 if (legend_pos=="far outside right") else 1.02), 1.0),
								loc				= 'upper left',
								framealpha		= 1,
								edgecolor		= "#404040",
								markerscale		= 1, 
								frameon			= True)			
	elif(legend_pos!="none"):
		legend = pyplot.legend(	loc			= legend_pos,
								framealpha	= 0.5,
								markerscale	= 1,
								frameon		= True)
	else:
		legend = None
	save_plot(fig=fig, plot_target=plot_target, dpi=dpi, legend=None)




# plot a heatmap of regularly spaced grid data
# The xvalues and yvalues (corresponding to columns & rows) may be discrete labels or numerical, however in the latter case they must be regularly spaced (since heatmap cells will be regularly sized along each axis)
# This function can either:
#	1. Plot into a single file (e.g., PDF, PNG, SVG etc). To do so, set plot_target to the appropriate file path.
#	2. Plot into a previously instantiated PdfPages object. To do so, set plot_target to an instantiated pdfpages_handle. This approach allows calling this function multiple times, each time adding a page to the PDF.
#	3. Generate and return a base64-RFC4648 string of the plot in SVG format, which can then be embedded e.g. in HTML pages. To do so, set plot_target to None. This option will not generate any actual file, i.e., the plot is returned as a string that encodes the plot as a SVG image.
def plot_heatmap(table,							# 2D numeric numpy array of size NR x NC, listing the values determining the heatmap colors. Columns with greater index are plotted further to the right, rows with greater index are plotted further down.
				plot_target,					# either a string, specifying a file path to which to plot (e.g., as PDF, PNG, JPG or SVG), or a pdfPages handle
				xvalues				= None,		# optional list or 1D numpy array of strings or numbers, of length NC, listing labels or x-values for the various table columns. If of type float, these will be interpreted as axis values, otherwise as discrete labels.
				yvalues				= None,		# optional list or 1D numpy array of strings or numbers, of length NR, listing labels or y-values for the various table rows. If of type float, these will be interpreted as axis values, otherwise as discrete labels.
				show_colorbar		= True,
				colormap			= "hot",	# name of the colormap to use. For a list of options see: https://matplotlib.org/stable/tutorials/colors/colormaps.html
				colormap_min		= None,		# optional minimum value covered by the colormap
				colormap_max		= None,		# optional maximum value covered by the colormap
				xlabel				= "", 		# e.g. 'time (seconds)'
				ylabel				= "",		# e.g. 'gene'
				clabel				= "",		# label for the colorbar axis, e.g. 'activity'
				show_values			= False,	# if True, the numerical values are shown in each heatmap cell. If False, no values are shown. If a 2D boolean numpy array (of size NR x NC), values will be displayed in those cells where show_values[r,c] is True.
				value_format		= "%.2g",	# format specifier for converting table values to strings, when displaying them inside cells. Only relevant if show_values!=False
				show_grid			= "auto",	# either "auto" or False/True
				cluster				= "none",	# whether and how to reorder rows & columns via hierarchically clustering. Options are "none", "rows", "columns", "rows_and_columns_joint" and "rows_and_columns_separate".
				plot_title			= "",
				plot_width			= None,		# numeric, plot width in inches. Typical values are 3-4. To convert to pixels (in the case of bitmap images), multiply this by DPI. If None, this is chosen automatically, scaling with the number of columns.
				plot_height			= None,		# numeric, plot height in inches. Typical values are 3-4. To convert to pixels (in the case of bitmap images), multiply this by DPI. If None, this is chosen automatically, scaling with the number of columns.
				dpi					= 200):		# resolution of the image, in dots per inch. Note that fonts scale with the inch-size of the figure, so a higher DPI means more smooth text for bitmap images. For PDFs this parameter is irrelevant. 
	# basic options checking
	(NR, NC) = table.shape
	if(plot_width is None):  plot_width = 0.75 * NC
	if(plot_height is None): plot_height = 0.75 * NR
	if(cluster not in ["none", "rows", "columns", "rows_and_columns_joint", "rows_and_columns_separate"]): raise ValueError("Invalid option '%s' for rows_and_columns_separate"%(rows_and_columns_separate))
	
	# determine the nature and limits of the axes
	xnumeric 	= (xvalues is not None) and ((isinstance(xvalues, numpy.ndarray) and (xvalues.dtype==float)) or (isinstance(xvalues, list) and isinstance(xvalues[0], float)))
	ynumeric	= (yvalues is not None) and ((isinstance(yvalues, numpy.ndarray) and (yvalues.dtype==float)) or (isinstance(yvalues, list) and isinstance(yvalues[0], float)))
	xlim		= ((xvalues[0]-0.5*(xvalues[min(NC-1,1)]-xvalues[0]) if xnumeric else -0.5), (xvalues[-1]+0.5*(xvalues[-1]-xvalues[-min(NC,2)]) if xnumeric else NC-0.5))
	ylim	 	= ((yvalues[0]-0.5*(yvalues[min(NR-1,1)]-yvalues[0]) if ynumeric else -0.5), (yvalues[-1]+0.5*(yvalues[-1]-yvalues[-min(NR,2)]) if ynumeric else NR-0.5))
	data_min 	= numpy.nanmin(table)
	data_max	= numpy.nanmax(table)
	if(colormap_min is None): colormap_min = data_min
	if(colormap_max is None): colormap_max = data_max
	
	# initialize figure
	bottom_margin	= 0.5 	# inches
	top_margin		= 0.5	# inches
	left_margin		= 1		# inches
	right_margin	= (1 if show_colorbar else 0.3) 	# inches
	total_width		= plot_width+left_margin+right_margin
	total_height	= plot_height+top_margin+bottom_margin
	fig 			= pyplot.figure(figsize=(total_width, total_height))
	ax				= pyplot.gca()
	
	# cluster rows & columns if needed
	if(("rows" in cluster) and (NR>1)):
		row_order = scipy.cluster.hierarchy.leaves_list(scipy.cluster.hierarchy.optimal_leaf_ordering(scipy.cluster.hierarchy.ward(table),table))
		if(yvalues is not None): yvalues = yvalues[row_order]
		table = table[row_order,:]
		if(isinstance(show_values,numpy.ndarray)): show_values = show_values[row_order,:]
	if(("columns" in cluster) and (NC>1)):
		if(cluster=="rows_and_columns_joint"):
			column_order = row_order
		else:
			column_order = scipy.cluster.hierarchy.leaves_list(scipy.cluster.hierarchy.optimal_leaf_ordering(scipy.cluster.hierarchy.ward(table.T),table.T))
		if(xvalues is not None): xvalues = xvalues[column_order]
		table = table[:,column_order]
		if(isinstance(show_values,numpy.ndarray)): show_values = show_values[:,column_order]

	# plot heatmap
	colormap = matplotlib.colormaps[colormap]
	im = pyplot.imshow(	table, 
						cmap			= colormap,
						aspect			= (plot_height/(1 if (NR==1) else (ylim[1]-ylim[0])))/(plot_width/(1 if (NC==1) else (xlim[1]-xlim[0]))),
						interpolation	= "nearest", 
						vmin			= colormap_min, 
						vmax			= colormap_max,
						extent 			= (xlim[0], xlim[1], ylim[1], ylim[0]))
						
	# set axis ticks
	if(not xnumeric): 
		if(xvalues is None):
			ax.set_xticks([], minor=False)
		else:
			rotate_xticks = (next((1 for x in xvalues if (len(x)>2)), -1)>0)
			ax.set_xticks(numpy.arange(0,NC,1), labels=xvalues, minor=False, rotation=(45 if rotate_xticks else 0), horizontalalignment=("right" if rotate_xticks else "center"))
		ax.set_xticks(numpy.arange(xlim[0],xlim[1],numpy.sign(xlim[1]-xlim[0])), minor=True)
	if(not ynumeric): 
		if(yvalues is None):
			ax.set_yticks([], minor=False)
		else:
			ax.set_yticks(numpy.arange(0,NR,1), labels=yvalues, minor=False)
		ax.set_yticks(numpy.arange(ylim[0],ylim[1],numpy.sign(ylim[1]-ylim[0])), minor=True)
	if(not (xnumeric and ynumeric)): 
		ax.tick_params(which='minor', length=0, color='black') # hide minor ticks
		
	# show grid lines if needed
	if(((show_grid=="auto") and (not (xnumeric and ynumeric))) or (show_grid==True)):
		ax.grid(visible=True, axis=('y' if xnumeric else ('x' if ynumeric else 'both')), color='w', linestyle='-', linewidth=1, which='minor') # show white grid lines at minor tick locations (to separate heatmap cells)
				
	# add colorbar
	if(show_colorbar):
		colorbar = pyplot.colorbar(location = 'right', fraction = 0.5/plot_width, shrink=0.9, pad=0.2/plot_width)
		colorbar.ax.set_ylabel(clabel, rotation=-90, va="bottom")
		
	# show values in grid cells
	if(show_values is not False):
		relative_values = (table - colormap_min)/(colormap_max-colormap_min)
		for r in range(NR):
			for c in range(NC):
				if((show_values is True) or show_values[r,c]):
					ax.text((xvalues[c] if xnumeric else c), 
							(yvalues[r] if ynumeric else r), 
							value_format%(table[r,c]), 
							ha="center", 
							va="center", 
							color=("white" if (color_to_grey_scale(colormap(relative_values[r,c]))<0.5) else "black"), 
							fontsize="small")

	# plot labels and titles
	pyplot.title(plot_title+"\n",fontsize="medium", fontname="Helvetica")
	if(xlabel!=""): pyplot.xlabel(xlabel, fontsize="medium", fontname="Helvetica")
	horizontal_ylabel = (len(ylabel)<=2)
	if(ylabel!=""): pyplot.ylabel(ylabel, fontsize="medium", fontname="Helvetica", rotation=(0 if horizontal_ylabel else 90), horizontalalignment=('right' if horizontal_ylabel else "center"))

	save_plot(fig=fig, plot_target=plot_target, dpi=dpi, legend=None)



			
###################################################
# MAIN


if __name__ == '__main__':
	
	# parse command line arguments
	allowed_grouping_methods = ["none","greedy_seeds","greedy_groups","greedy_groups_refined","AffinityPropagation","DBSCAN","HDBSCAN"]
	parser = argparse.ArgumentParser(description="Estimate the number of clusters based on pairwise predictor variables (e.g., pairwise differences & similarities) between units.", epilog="")

	# main inputs
	parser.add_argument('-c','--in_classifications', default="", type=str, help='Optional path to one or more classical input tables, containing classifications for unit pairs being in the same cluster/group or not. This table must have at least 3 columns (unit1 & unit2 names, ingroup classification). Valid classifications are "True"/"yes"/"1" (=ingroup) or "False"/"no"/"0" (=outgroup), case-insensitive; any other strings are interpreted as an unknown/missing classification. The loaded classifications are used for supervised model training, testing and evaluation, as well as potentially for predictions where applicable');
	parser.add_argument('-p','--in_predictors', default="", type=str, help='Paths (or shell wildcards) to one or more classical input tables, containing pairwise predictor variables between units. Multiple paths (or wildcards) are allowed and must be separated by semicolon. Each input table must contain at least 2 columns (unit1 & unit2 names) and an arbitrary number of data columns (pairwise predictor variables). If a pre-calibrated model is not provided, then only units also represented in the --in_classifications table are considered.');
	parser.add_argument('-m','--in_classifier', default="", type=str, help='Optional path to a calibrated model for classifying unit pairs. This will typically have been generated by a previous execution of this script.');
	parser.add_argument('-g','--in_Kgroups', default="", type=str, help='Optional path to a classical input table containing known group assignments for a subset of units. This may be used, for example, to infer known classifications for a subset of unit pairs, or to generate test data for assessing the accuracy of the group-counting algorithm. These group assignments are not used to constrain or directly guide the group-counting or group-estimation algorithm.');

	# main outputs
	parser.add_argument('--out_classifier', default="", type=str, help='Optional path to save the calibrated classifier model, as a pickled file. The recommended file extension is .pickle.');
	parser.add_argument('--out_test_results', default="", type=str, help='Optional path to save the scatter test results, i.e., known vs predicted classifications for the chosen classifier across the test set. The classifier used is the one re-trained on the training+validation set after hyperparameter selection, but not yet trained on the test set. Not relevant if a classifier is already provided.');
	parser.add_argument('--out_predicted_classifications', default="", type=str, help='Optional path to save the predicted pairwise classifications.');
	parser.add_argument('--out_group2units', default="", type=str, help='Optional path to save the estimated group-->units mapping (representative units in the 1st column, group sizes in the 2nd column, member units in the 3rd column as a comma-separated list).');
	parser.add_argument('--out_unit2group', default="", type=str, help='Optional path to save the estimated unit-->group mapping. Note that this file will contain the same information as --out_group2units.');
	parser.add_argument('--out_forward_sensitivities', default="", type=str, help='Optional path to save the forward sensitivities in tabular format.');
	parser.add_argument('--out_backward_sensitivities', default="", type=str, help='Optional path to save the backward sensitivities in tabular format.');
	parser.add_argument('--plot_affinity_matrix', default="", type=str, help="Optional path to plot the affinity matrix, listing known and/or predicted classifications. If --probabilistic is passed, then entries of this matrix are interpreted as probabilities, with a value of 1 suggesting that two units are surely in the same group. Only relevant if --diarize is not 'none'.");
	parser.add_argument('--out_classifier_phase_diagram', default="", type=str, help="Optional path to save the classifier's phase diagram, i.e., showing the predicted classification as a function of predictor values, evaluated on a regular meshgrid. This file will contain one column for each predictor and a column listing the corresponding predicted classifications.");
	parser.add_argument('--plot_classifier_phase_diagram', default="", type=str, help="Optional path to plot the classifier's phase diagram as a contour plot, i.e., showing the predicted classification as a function of predictor values. Only applicable to classifiers using exactly two predictors.");
	parser.add_argument('--scatterplot_test_results', default="", type=str, help='Optional path to a PDF for scatterplotting test results, i.e., known vs predicted classifications for the chosen classifier, as a function of any two chosen predictors. All combinations of predictors of the selected classifier will be considered, and the multiple generated plots are saved as distinct pages in the same PDF.');
	parser.add_argument('--scatterplot_input_data', default="", type=str, help='Optional path to a PDF for scatterplotting input data, i.e. known classifications as a function of any two considered predictors, with points colored according to classification. All combinations of predictors in the input data will be considered, and the multiple generated plots are saved as distinct pages in the same PDF.');
	parser.add_argument('--out_log', default="", type=str, help='Optional path to text file for logging messages.');

	# input data format
	parser.add_argument('--delimiter', default='\\t', help='Column delimiter in classical input/output files. For tabs use \\t (default). For elastic whitespace use "" (empty).');
	parser.add_argument('--comment_prefix', action='store', default="#", help='Comment prefix in classical input/output tables (default: %(default)s). Set to "" (empty) if no comments should be allowed.');
	parser.add_argument('--column_names_are_in', default='first_data_line', choices=['none','last_comment_line','first_data_line'], help="Where the table header (listing column names) is located in the input/output tables. The option 'last_comment_line' means that column names are in the last comment line prior to any data. If set to 'none', then column names are column indices (starting at 0).");
	parser.add_argument('--case_sensitive', action='store_true', dest="case_sensitive", default=False, help="Sample & observation filtering (by name or metadata) should be case-sensitive (by default it is case-insensitive).");
	parser.add_argument('--unit1_column', default="unit1", type=str, help="Column containing unit1 names in the pairwise input tables (e.g. --in_classifications & --in_predictors). If column names are available, this specifies a column by name, otherwise it specifies a column by index (starting at 0).");
	parser.add_argument('--unit2_column', default="unit2", type=str, help="Column containing unit2 names in the pairwise input tables (e.g. --in_classifications & --in_predictors). If column names are available, this specifies a column by name, otherwise it specifies a column by index (starting at 0).");
	parser.add_argument('--classification_column', default=None, help="Column containing classifications of each pair as in-group ('True'/'False'/'') in the input classifications table. If column names are available, this specifies a column by name, otherwise it specifies a column by index (starting at 0).");
	parser.add_argument('--predictor_naming', default='column_name', choices=['column_name','filebasename.column_name','filebasename.number','filebasename'], help="How to name predictor variables loaded from input files. 'filebasename.column_name' means predictor names will be of the format <input file basename>.<column name in input file>. Predictor naming happens before predictor filtering, i.e., --only_predictors and --omit_predictors should refer to the final predictor names.");
	parser.add_argument('--symmetrize', action='store_true', dest="symmetrize", default=False, help='Symmetrize input data (classifications and predictor values), i.e., define all data for non-ordered unit pairs {a,b} instead of ordered pairs (a,b). If data are found for both versions (a,b and b,a), the average is kept. Note that for grouping/group counting, the affinity matrix is always symmetrized.');
	parser.add_argument('--unit_column', default="unit", type=str, help="Column containing unit names in the unit-wise input/output tables (e.g., --in_Kgroups). If column names are available, this specifies a column by name, otherwise it specifies a column by index (starting at 0).");
	parser.add_argument('--group_column', default="group", help="Column containing group names in the --in_Kgroups table. If column names are available, this specifies a column by name, otherwise it specifies a column by index (starting at 0).");
	
	# output data format
	parser.add_argument('--group_member_separator', default=',', help='Separator to use between group members (unit names) in the --out_group2units file. (default: %(default)s)');

	# renaming units
	parser.add_argument('--in_unit_name_map', default="", type=str, help='Optional path to a lookup table that maps input unit names (as encountered in --in_classifications and --in_predictors) to new names. This may be useful for example if --in_classifications lists abbreviated unit names for space efficiency.');
	parser.add_argument('--missing_unit_name_in_map', default='error', choices=["error","keep_old"], help='How to handle missing unit names in the optional unit name mapping (see option --in_unit_name_map). (default: %(default)s)');
	parser.add_argument('--old_unit_name_column', default=0, type=int, help='Column index (0-based) where old unit names are listed, in the unit name mapping table. (default: %(default)s)');
	parser.add_argument('--new_unit_name_column', default=1, type=int, help='Column index (0-based) where new unit names are listed, in the unit name mapping table. (default: %(default)s)');

	# filtering input data
	# Note that if you want to only filter the data for classifier selection/training but not for subsequent predictions, you will need to run this script twice, saving the classifier in the first run and then using it for predictions in the 2nd run.
	parser.add_argument('--only_unit1_names_in_file', default="", type=str, help="Optional input file listing unit1 names (in the first column) to restrict to.")
	parser.add_argument('--omit_unit1_names_in_file', default="", type=str, help="Optional input file listing unit1 names (in the first column) to omit.")
	parser.add_argument('--only_unit2_names_in_file', default="", type=str, help="Optional input file listing unit2 names (in the first column) to restrict to.")
	parser.add_argument('--omit_unit2_names_in_file', default="", type=str, help="Optional input file listing unit2 names (in the first column) to omit.")
	parser.add_argument('--only_unit_name_pairs_in_file', default="", type=str, help="Optional input file listing unit1 & unit2 names (in the first 2 columns) to restrict to.")

	parser.add_argument('--only_predictors', default="", type=str, help="Optional comma-separated list of column names (or indices, if column names are not available) to restrict to as predictors. By default, all columns other than the unit1 & unit2 columns will be considered as possible predictors.")
	parser.add_argument('--omit_predictors', default="", type=str, help="Optional comma-separated list of column names (or indices, if column names are not available) to omit from the list of considered predictors.")

	# model evaluation options
	# If training_fraction+validation_fraction+test_fraction>1, then they are normalized so that they sum to 1
	# There are multiple options for re-balancing data for training & model evaluation: --rarefy_to_balance_for_fitting, --SGD_balanced_training_weights, --cv_scoring
	parser.add_argument('--training_fraction', default=0.7, type=float, help="Fraction of data points (unit pairs) to use for model training/fitting. Must be within (0,1]. (default: %(default)s)");
	parser.add_argument('--validation_fraction', default=0.2, type=float, help="Fraction of data points (unit pairs) to use for model cross-validation and hyperparameter selection. Must be within (0,1]. (default: %(default)s)");
	parser.add_argument('--test_fraction', default=0.1, type=float, help="Fraction of data points (unit pairs) to use for model testing, after hyperparameter selection and fitting. Must be within (0,1]. (default: %(default)s)");
	parser.add_argument('--NCV', default=5, type=int, help="Number of cross-validations to perform when selecting model hyperparameters and for testing the final model. With each cross-validation, the training/validation/test sets are randomly rechosen. (default: %(default)s)");
	parser.add_argument('--rarefy_to_balance_for_fitting', action='store_true', dest="rarefy_to_balance_for_fitting", default=False, help="Rarefy datasets used for classifier fitting to achieve a balance between classifications, by only subsampling the dominant classification. Hence, the data subset with the lesser-common classification will be rarefied to be of the same size as the more-common classification. For example, if a training dataset comprises 100 ingroup samples and 500 outgroup samples, then only 100 randomly chosen outgroup samples and all 100 ingroup samples will be kept for training. Rarefaction is applied in all cases where the classifier is fitted, including the final fit after classifier selection. Rarefaction is not applied to the validation or test sets (but see option --cv_scoring).");
	parser.add_argument('--max_training_count', default=None, type=int, help="Optional maximum number of samples to use for training. Training samples are subsampled randomly down to this number if needed, after potentially rarefying to balance (see option --rarefy_to_balance_for_fitting).");
	parser.add_argument('--SGD_balanced_training_weights', action='store_true', dest="SGD_balanced_training_weights", default=False, help="Weigh samples inversely proportional to the frequency of their classification, when training SGD models. For example, if the training data comprise more outgroups than ingroups, ingroup samples will be given proportionally more weight in the loss function. This option is irrelevant when using --rarefy_to_balance_for_fitting. Note that this option is separate from how trained models are evaluated on the validation sets (see option --cv_scoring).");
	parser.add_argument('--cv_scoring', default=None, type=str, choices=["accuracy","balanced_accuracy","average_precision","neg_brier_score","neg_log_loss","precision","recall","jaccard"], help="How to score models on test/validation sets, during cross-validation. For example, 'balanced_accuracy' scores models based on the average of the true positive rate and true negative rate; in that case, consider also balancing the training sets using --rarefy_to_balance_for_fitting or --SGD_balanced_training_weights. If left unspecified, the default scoring method of the considered model type will be used. Naming is as for scikit-learn's classifier scoring types, listed at: https://scikit-learn.org/stable/modules/model_evaluation.html#scoring-parameter (default: %(default)s)");
	parser.add_argument('--fit_final_using', default="all_data", type=str, choices=["all_data","nontest_data"], help="Which part of the data to train the final chosen classifier with, i.e. before saving it as a pickle and before using it for predictions. 'nontest_data' means the non-test data subset. Note that if --rarefy_to_balance_for_fitting is set, the data are always rarefied to balance prior to fitting, even for the final fit. (default: %(default)s)");

	# model type & hyperparameters to explore
	parser.add_argument('--model_type', default="MLP", choices=["MLP", "RF", "SGD"], help='What type of classifier model to use. MLP:multilayer perceptron, aka. feed-forward neural network, RF:Random Forest, SGD:Stochastic Gradient Descend. Irrelevant if a calibrated classifier is provided.');
	parser.add_argument('--hidden_layer_sizes', default="", type=str, help="Optional comma-separated list of hidden layer sizes to explore for neural classifiers. Each element in this list should be a tuple of positive integers. For example, '(3,1), (4,2), (6,4,4)' defines three alternative neural network structures, with either 2, 2 or 3 hidden layers, respectively. If empty, this list is chosen automatically based on heuristics. Irrelevant if a claibrated model is already provided.");
	parser.add_argument('--neural_activations', default="identity,logistic,tanh", type=str, help="Comma-separated list of activation functions to explore for neural classifiers. Irrelevant if a claibrated model is already provided. (default: %(default)s)");
	parser.add_argument('--SGD_learning_rates', default="optimal,constant,invscaling,adaptive", type=str, help="Comma-separated list of learning rate rules to consider for SGD classifiers. Irrelevant if a claibrated model is already provided. (default: %(default)s)");
	parser.add_argument('--RF_criteria', default="gini,entropy,log_loss", type=str, help="Comma-separated list of split ciriteria to consider for RF classifiers. Irrelevant if a claibrated model is already provided. (default: %(default)s)");

	# predictor selection algorithm
	parser.add_argument('--predictor_selection_minNX', default=1, type=int, help='Minimum number of predictors to consider. (default: %(default)s)');
	parser.add_argument('--predictor_selection_maxNX', default=None, type=int, help='Optional maximum number of predictors to consider. If left unspecified, all provided predictors could in principle serve as predictors for the model.');
	parser.add_argument('--predictor_selection_max_added', default=2, type=int, help='Maximum number of predictors that may be added in each iteration, except for the first one. In the first iteration at least minNX predictors will be included regardless of this limit. A greater value means more alternative options are explored in each iteration, but at the expense of increased computation time. (default: %(default)s)');
	parser.add_argument('--predictor_selection_max_removed', default=2, type=int, help='Maximum number of predictors that may be removed in each iteration. In any case, the number of removed predictors may be smaller, for example in order to ensure at least minNX predictors. A greater value means more alternative options are explored in each iteration, but at the expense of increased computation time. If 0, the selection algorithm becomes a pure forward feature selection algorithm. (default: %(default)s)');
	parser.add_argument('--predictor_selection_max_iterations', default=1000, type=int, help='Maximum number of allowed iterations, as a safety valve in case convergence takes tool long. (default: %(default)s)');
	parser.add_argument('--predictor_selection_score_change_threshold', default=0, type=float, help='Optional score change threshold, below which predictor selection is considered to have converged and is therefore halted. If <=0, then predictor selection halts if no score improvement is achieved (compared to the previous iteration) or if the maximum number of iterations was reached.');

	parser.add_argument('--analyze_forward_sensitivities', action='store_true', dest="analyze_forward_sensitivities", default=False, help="Perform forward sensitivity analysis of predictors, i.e. considering only one predictor at a time and examining the performance of the best classifier based on that single predictor. Only relevant if a classifier wasn't provided.");
	parser.add_argument('--analyze_backward_sensitivities', action='store_true', dest="analyze_backward_sensitivities", default=False, help="Perform backward sensitivity analysis of the selected classifier, i.e. leaving out one predictor at a time and examining its impact on the classifier's performance. Only relevant if a classifier wasn't provided.");

	# classification options
	parser.add_argument('--probabilistic', action='store_true', dest="probabilistic", default=False, help='Predicted classifications (ingroup vs outgroup) should be probabilistic, i.e., given in terms of probabilities instead of True/False.');
	parser.add_argument('--diarize', default="all", type=str, choices=["only_pairs_with_predictors","all","none"], help="Which subset of the input data (which subset of unit pairs) to diarize. 'all' means all unit pairs for which predictor data and/or classifications are provided. 'only_pairs_with_predictors' means only unit pairs listed in --in_predictors. 'none' means that no prediction/diarization will be performed, so this script stops after the model calibration.");
	parser.add_argument('--reclassify_knowns', action='store_true', dest="reclassify_knowns", default=False, help='When classifying samples (unit pairs) with known classifications, predict classifications again anyway. By default, known classification are used where available.');

	# diarization/grouping options
	parser.add_argument('--counting_method', default="none", type=str, help="Method for counting groups among all input units, based on the ingroup/outgroup classifications of unit pairs. 'greedy_seeds' uses the greedy_seeds grouping method (see option --grouping_method), repeated multiple times to reduce variance. Quick tests indicate that 'HDBSCAN' and 'greedy_groups_refined' work best, depending on situation, followed by 'AffinityPropagation'. Group counting is done across the units specified by --diarize. (default: %(default)s)")
	parser.add_argument('--greedy_counting_Nrepeats', default=10, type=int, help="Number of random repeats for counting groups, when using a greedy method (e.g., 'greedy_seeds' or 'greedy_groups').");
	parser.add_argument('--grouping_method', default="none", type=str, help="Method for grouping units based on their pairwise ingroup/outgroup classifications. 'greedy_seeds' sequentially assigns each unit to an existing group or to a new group (the first unit in a newly formed group is a 'seed' or 'centroid' for that group, and new candidate seeds are compared to previous seeds). 'greedy_groups' is similar to greedy_seeds, but each new candidate seed is first compared to all members of all groups constructed thus far, in terms of the average affinity between that unit and each group's members; 'greedy_groups' takes an optional affinity_threshold parameter, specified as 'greedy_groups_'<affinity_threshold>. 'greedy_groups_refined' is similar to 'greedy_groups', but the groups are iteratively refined (calculating average affinities to groups defined in the previous iteration) until no more changes are needed; 'greedy_groups_refined' takes an optional affinity_threshold parameter, specified as 'greedy_groups_refined_'<affinity_threshold>. Grouping is done across the units specified by --diarize. (default: %(default)s)")

	# handling missing data
	parser.add_argument('--replace_nan_predictors', default="", type=str, help="Optional semicolon-separated list of replacement specifiers, specifying replacement values for NaN/Inf predictors. Each specifier is of the format <comma-separated list of predictor names or wildcards>:<replacement value if NaN/Inf>. For example, 'GCD:1;AAI,ANI:0' specifies that any non-finite GCD values should be replaced with 1, while any non-finite AAI & ANI values should be replaced with 0.")
	parser.add_argument('--remaining_nan_predictors', default="error", type=str, choices=["error","omit_samples","omit_predictors","zero"], help="How to handle remaining NaN/Inf predictor values, both for model selection/training as well as predictions, after all NaN/Inf-replacements are done. 'zero' means that non-finite predictor values are replaced with zeros. Note that if a calibrated model uses a specific set of predictors, it generally won't make sense to omit predictors during predictions because of NaNs; in that case consider using 'omit_samples'. (default: %(default)s)")
	parser.add_argument('--missing_pairs', default="error", type=str, choices=["error","outgroup","ingroup"], help="How to treat unit pairs missing from the predictions, when counting groups or grouping. 'ingroup' means that missing pairs are assumed to be in-groups (i.e., belong to the same group), while 'outgroup' means that missing pairs are assumed to be out-groups (i.e., belong to a different group). 'error' means that an error is raised. Regardless, self-comparisons are always classified as ingroup during grouping or group counting. (default: %(default)s)")
	
	# options for testing the diarization accuracy
	parser.add_argument('--test_counting_methods', default="", type=str, help="Optional comma-separated list of counting methods to test (see --counting_method for options). Only available if --in_Kgroups was provided.")
	parser.add_argument('--reserve_Kgroups_for_tests', action='store_true', dest="reserve_Kgroups_for_tests", default=False, help="Only use the information provided by --in_Kgroups for testing the accuracy of various procedures (notably the counting procedure), but don't use that information to improve/complement predictions.")
	parser.add_argument('--counting_Ntests', default=10, type=int, help="Number of accuracy tests to perform for the group counting algorithm. Only available if --in_Kgroups was provided.")
	parser.add_argument('--out_counting_test_results', default="", type=str, help="Optional path to output classical table for saving the counting test results (true vs predicted number of groups).")
	parser.add_argument('--plot_counting_test_results', default="", type=str, help="Optional path to output PDF for scatterplotting the counting test results (true vs predicted number of groups).")


	# options for generating fake data, e.g. for debugging purposes
	parser.add_argument('--out_fake_basepath', default="", type=str, help='Optional basepath to output files that will contain fake unit groups, pairwise predictor values and pairwise classifications, for debugging purposes.');
	parser.add_argument('--fake_Npredictors', default=2, type=int, help='Number of true predictor variables to include in the fake data, i.e. fake variables that follow a bimodal distribution and indeed correlate with the fake classifications. Only relevant if --out_fake_basepath is specified.');
	parser.add_argument('--fake_Nnoise', default=2, type=int, help='Number of noise (spurious predictor) variables to include in the fake data, i.e. fake variables that follow a unimodal distribution and do not actually correlate with the fake classifications. Only relevant if --out_fake_basepath is specified.');
	parser.add_argument('--fake_Nunits', default=10, type=int, help='Number of units to generate pairwise fake data for.');
	parser.add_argument('--fake_Ngroups', default=3, type=int, help='Number of distinct groups/clusters in the fake data.');
	parser.add_argument('--fake_Fclassified', default=0.5, type=float, help='Fraction of fake samples (unit pairs) for which to include the classifications in the output file. For all other fake samples, classifications will be omitted (i.e., unknown).');

	# general options
	parser.add_argument('--seed', default=None, type=int, help="Optional seed for random number generator. If left unspecified, the generator is seeded automatically.");
	parser.add_argument('--Nthreads', default=0, type=int, help="Number of threads to use (if strictly positive), or number of threads to use below the number of available cores (if non-positive). If negative or zero, use the number of available cores minus this value.");
	parser.add_argument('--gzip_compression_level', action='store', type=int, default=6, choices=[1,2,3,4,5,6,7,8,9], help="Compression level to use for gzipped output files. For levels above 6 the computation time grows rapidly with diminishing returns. (default: '%(default)s)'");
	parser.add_argument('--max_Nscatter', default=20000, type=int, help="Maximum number of points to include in scatterplots, for example when showing test results. If the number of scatterpoints exceeds this limit, a random subset is shown instead. (default: '%(default)s)'");

	parser.add_argument('--unit_type', default="unit", type=str, help="What each unit corresponds to (e.g. 'MAG' or 'OTU' or 'recording'). This only affects output formatting, for example column names. (default: %(default)s)");
	parser.add_argument('--group_type', default="group", type=str, help="What each group corresponds to (e.g. 'genus' or 'phylum' or 'individual' or 'taxon'). This only affects output formatting, for example column names. (default: %(default)s)");
	parser.add_argument('--group_naming', default="unit_rep", choices=["unit_rep", "enumerate"], help="How to name inferred groups. 'unit_rep means the name of the representative/centroid unit will be used. 'enumerate' means that groups are named sequentially following the pattern <group type><number>. (default: %(default)s)");
	parser.add_argument('--case_tag', default="", type=str, help="Optional short text to identify this particular run. This will be added to plot subtitles and as a comment to output files.");

	parser.add_argument('--plot_width', default=3, type=float, help="Width of generated plots, in inches.");
	parser.add_argument('--plot_height', default=3, type=float, help="Height of generated plots, in inches.");

	parser.add_argument('-f','--force', action='store_true', dest="force", default=False, help='Replace existing output file without asking.');
	parser.add_argument('-v','--verbosity', action='store', default=0, type=int, help="Non-negative integer, specifying the verbosity level. A value of 0 means no messages will be printed except for errors.");
	parser.add_argument('--verbose_prefix', action='store', default="  ", type=str, help="Prefix to be used for any messages. This might be useful if you're integrating this script in another pipeline.");
	args = parser.parse_args()

	def abort(error_msg, error_code=1):
		print(args.verbose_prefix+error_msg.replace("\n","\n"+args.verbose_prefix))
		sys.exit(error_code)

	def print2(message):
		print(args.verbose_prefix+message.replace("\n","\n"+args.verbose_prefix))
		if(args.out_log!=""):
			with open_file(args.out_log,'at') as fout:
				fout.write(message+"\n")
			
	# basic option checking
	if(args.out_log!=""): check_output_file(file=args.out_log,force=args.force,verbose=(args.verbosity>=2),verbose_prefix=args.verbose_prefix)
	if(args.delimiter=='\\t'): args.delimiter='\t'
	if((args.in_classifier=="") and (args.in_classifications=="") and (args.diarize!="none")): 
		abort("ERROR: Neither --in_classifier nor --in_classifications were provided, so diarization is not possible")
	if((args.in_classifications!="") and (args.classification_column is None)):
		abort("ERROR: Missing --classification_column, needed for loading classifications")
	args.training_fraction 		= min(1,max(0,args.training_fraction))
	args.validation_fraction 	= min(1,max(0,args.validation_fraction))
	args.test_fraction 			= min(1,max(0,args.test_fraction))
	if(args.seed is None):
		numpy.random.seed()
		args.seed = numpy.random.randint(1,100000)
		if(args.verbosity): print2("Seeding random number generator with seed %d"%(args.seed))
	random.seed(args.seed)
	numpy.random.seed(args.seed)
	CPP.seed_random_number_generators(args.seed)
	RNG = numpy.random.default_rng(args.seed)
	if(args.Nthreads<=0): args.Nthreads = max(1, os.cpu_count()+args.Nthreads)
	if((args.counting_method!="none") and (args.diarize=="none")):
		abort("ERROR: Predictions are needed for counting groups, however --classify was set to 'none'")
	elif((args.grouping_method!="none") and (args.diarize=="none")):
		abort("ERROR: Predictions are needed for grouping, however --classify was set to 'none'")
	if(((args.out_group2units!="") or (args.out_unit2group!="")) and (args.grouping_method=="none")):
		abort("ERROR: Grouping method cannot be 'none', when requesting to save groups to a file")
	if((not args.analyze_forward_sensitivities) and (args.out_forward_sensitivities!="")):
		abort("ERROR: Forward sensitivities cannot be saved to a file when forward sensitivity analysis is disabled")
	if((not args.analyze_backward_sensitivities) and (args.out_backward_sensitivities!="")):
		abort("ERROR: Forward sensitivities cannot be saved to a file when backward sensitivity analysis is disabled")
	if((args.test_counting_methods!="") and (args.in_Kgroups=="")):
		abort("ERROR: Cannot test the accuracy of the group-counting process, when no group assignments are provided")
	args.training_fraction, args.validation_fraction, args.test_fraction = normalize_downward(args.training_fraction, args.validation_fraction, args.test_fraction)
	args.predictor_selection_score_change_threshold = max(0, args.predictor_selection_score_change_threshold)
	test_counting_methods = args.test_counting_methods.split(",")
	def aux_valid_grouping_method(method):
		return (method in allowed_grouping_methods) or method.startswith("greedy_groups_")
	if(not aux_valid_grouping_method(args.counting_method)):
		abort("ERROR: Unknown counting method '%s'"%(args.counting_method))
	if(not aux_valid_grouping_method(args.grouping_method)):
		abort("ERROR: Unknown grouping method '%s'"%(args.grouping_method))
	if((invalid_method:=next((method for method in test_counting_methods if (not aux_valid_grouping_method(method))),""))!=""):
		abort("ERROR: Unknown counting method '%s' requested for testing"%(invalid_method))
	if(args.counting_method.startswith("best_") and (args.test_counting_methods=="")):
		abort("ERROR: Counting method can only be set to '%s' if count-testing is enabled (option --test_counting_methods)"%(args.counting_method))
	if(args.grouping_method.startswith("best_") and (args.test_counting_methods=="")):
		abort("ERROR: Grouping method can only be set to '%s' if count-testing is enabled (option --test_counting_methods)"%(args.grouping_method))
	
	classifier 			  = None
	classifications 	  = None
	string2classification = {"false":0, "0":0, "no":0, "true":1, "1":1, "yes":1}
	
	##########################################
	# Generate fake data if requested
	
	if(args.out_fake_basepath!=""):
		if(args.verbosity): print2("Generating fake data (%d predictors, %d noise variables, %d units).."%(args.fake_Npredictors,args.fake_Nnoise,args.fake_Nunits))
		fake_unit_names 	= ["unit_%d"%(u) for u in range(args.fake_Nunits)]
		fake_unit_groups	= RNG.choice(a=args.fake_Ngroups, size=args.fake_Nunits, replace=True)
		fake_ingroup 		= (fake_unit_groups.reshape((args.fake_Nunits,1)) == fake_unit_groups.reshape((1,args.fake_Nunits)))
		# generate random predictors for all unit pairs
		fake_predictors 					= numpy.empty((args.fake_Nunits, args.fake_Nunits, args.fake_Npredictors))
		fake_predictors[fake_ingroup,...]  	= 10+RNG.multivariate_normal(mean=numpy.ones(args.fake_Npredictors), cov=random_covariance(N=args.fake_Npredictors, std=1), size=(args.fake_Nunits,args.fake_Nunits))[fake_ingroup,...]
		fake_predictors[~fake_ingroup,...] 	= 10+RNG.multivariate_normal(mean=numpy.zeros(args.fake_Npredictors), cov=random_covariance(N=args.fake_Npredictors, std=1), size=(args.fake_Nunits,args.fake_Nunits))[~fake_ingroup,...]
		fake_noise							= 10+RNG.multivariate_normal(mean=numpy.zeros(args.fake_Npredictors), cov=random_covariance(N=args.fake_Nnoise, std=1), size=(args.fake_Nunits,args.fake_Nunits))
		fake_predictors 					= numpy.concatenate((fake_predictors,fake_noise), axis=2)
		# flatten fake predictors & classifications w.r.t. the unit axes, keeping only upper-triangular entries
		fake_predictors 	 = numpy.moveaxis(extract_upper_triangular(numpy.moveaxis(fake_predictors, source=2, destination=0), flat=False), source=0, destination=1)
		fake_classifications = numpy_object2str(fake_ingroup[numpy.triu_indices(args.fake_Nunits)])
		fake_classifications_masked = fake_classifications.copy()
		fake_classifications_masked[RNG.choice(fake_classifications.size, size=int((1-args.fake_Fclassified)*fake_classifications.size), replace=False)] = "" # hide a subset of classifications from the output file
		# save table of pairwise comparisons between fake units
		save_classical_table(output_file			= args.out_fake_basepath+"_pairs.tsv",
							table					= numpy.hstack(([(fake_unit_names[u1],fake_unit_names[u2]) for u1 in range(args.fake_Nunits) for u2 in range(u1,args.fake_Nunits)],numpy.expand_dims(fake_classifications,axis=1),numpy.expand_dims(fake_classifications_masked,axis=1),fake_predictors)),
							column_names			= [args.unit1_column, args.unit2_column, "classification", "classification_masked"] + ["predictor_%d"%(s) for s in range(fake_predictors.shape[1])],
							delimiter				= args.delimiter,
							comment_prefix			= args.comment_prefix,
							comments				= "Fake pairwise predictors & classifications (ingroup:1/True, outgroup:0/False), generated between %d units (%d predictors correlated to classifications, %d noise variables uncorrelated to classifications, %d groups)\nGenerated on: %s\nUsed command: %s"%(args.fake_Nunits,args.fake_Npredictors,args.fake_Nnoise,args.fake_Ngroups,get_date_time(),get_shell_command()))
		# save table of fake unit groups
		save_classical_table(output_file			= args.out_fake_basepath+"_groups.tsv",
							table					= numpy.column_stack((fake_unit_names,fake_unit_groups)),
							column_names			= [args.unit_column, "group"],
							delimiter				= args.delimiter,
							comment_prefix			= args.comment_prefix,
							comments				= "Fake units & groups\nGenerated on: %s\nUsed command: %s"%(get_date_time(),get_shell_command()))
		


	##########################################
	# Load data
	
	# load unit name mapping
	if(args.in_unit_name_map!=""):
		if(args.verbosity>=1): print2("Loading unit name mapping..")
		old2new_unit_name = read_rename_list_from_table(file			= args.in_unit_name_map,
														old_name_column	= args.old_unit_name_column,
														new_name_column	= args.new_unit_name_column,
														delimiter		= args.delimiter,
														comment_prefix	= args.comment_prefix,
														header_is		= 'none',
														verbose_prefix	= "  ")
		if(args.verbosity>=1): print2("  Note: Loaded %d old-->new unit name pairs"%(len(old2new_unit_name)))
	else:
		old2new_unit_name = None
		
		
	# load name lists for filtering
	if(args.only_unit1_names_in_file!=""):
		only_unit1_names = load_names_from_table(file=args.only_unit1_names_in_file, delimiter=args.delimiter, comment_prefix=args.comment_prefix, header=False)
	else:
		only_unit1_names = None
	if(args.omit_unit1_names_in_file!=""):
		omit_unit1_names = load_names_from_table(file=args.omit_unit1_names_in_file, delimiter=args.delimiter, comment_prefix=args.comment_prefix, header=False)
	else:
		omit_unit1_names = None
	if(args.only_unit2_names_in_file!=""):
		only_unit2_names = load_names_from_table(file=args.only_unit2_names_in_file, delimiter=args.delimiter, comment_prefix=args.comment_prefix, header=False)
	else:
		only_unit2_names = None
	if(args.omit_unit2_names_in_file!=""):
		omit_unit2_names = load_names_from_table(file=args.omit_unit2_names_in_file, delimiter=args.delimiter, comment_prefix=args.comment_prefix, header=False)
	else:
		omit_unit2_names = None

	# load name-pairs for filtering
	if(args.only_unit_name_pairs_in_file!=""):
		if(args.verbosity): print2("Loading list of unit pairs to restrict calculations to..")
		only_unit_name_pairs = set()
		with open_file(args.only_unit_name_pairs_in_file,'rt') as fin:
			for line in fin:
				if(line.endswith("\n")): line = line[:-1]
				line = split_comments(line,args.comment_prefix)[0]
				if(line==""): continue
				parts = line.split(args.delimiter)
				if(len(parts)<2): abort("  ERROR: Found %d parts in line %d of input file, but expected at least 2"%(len(parts),Nlines))
				only_unit_name_pairs.add((parts[0],parts[1])) 
		if(args.verbosity): print2("  Note: Loaded %d unit name pairs to restrict calculations to"%(len(only_unit_name_pairs)))
	else:
		only_unit_name_pairs = None
		
	# find all input predictor tables
	if(((args.in_classifications!="") and (args.in_classifier=="")) or (args.diarize!="none")):
		predictor_filepaths = [path for paths in split_escaped(args.in_predictors, ";") for path in glob.glob(paths)]
		if(len(predictor_filepaths)==0): 
			abort("ERROR: No input predictor tables found")
		elif(args.verbosity):
			print2("Note: Found %d input predictor tables"%(len(predictor_filepaths)))

	# load pairwise classifications as ingroup vs outgroup
	# To save RAM, we represent units using unique integers, rather than by the unit names. unit_name2int will define the mapping between unit names & integers.
	# Unit integer IDs have no relationship to the order in which the loaded data are stored, they are merely used as alternative (more memory-efficient) representations of the units.
	if((args.in_classifications!="") and (args.in_classifier=="")):
		if(args.verbosity): print2("Loading classifications..")
		classified_pairs,\
		classifications,\
		unit_name2int,\
		dummy,\
		found_column_names = load_flat_pairwise_data(	filepath			= args.in_classifications,
														only_column_names	= args.classification_column,
														numeric				= False,
														verbose				= args.verbosity,
														verbose_prefix		= "  ")
		classifications = numpy.asarray([string2classification.get(cl[0].lower(),-1) for cl in classifications], dtype=int)
		# omit any missing/unknown classifications
		keep = numpy.where(classifications>=0)[0]
		if(len(keep)<len(classifications)):
			classifications	 = classifications[keep]
			classified_pairs = [classified_pairs[k] for k in keep]
		if(args.verbosity): print2("  Note: Loaded classifications for %d pairs covering %d distinct units"%(len(classified_pairs),len(unit_name2int)))
		# create mapping of pair-->index
		classified_pair2index = {pair:p for p,pair in enumerate(classified_pairs)}
		if(args.symmetrize): 
			# symmetrize loaded classifications, i.e., so that only pairs (a,b) with a<=b are kept
			# classified_pair2index will then map both (a,b) and (b,a) to the same index
			classified_pairs, classifications, classified_pair2index = symmetrize_data(pairs=classified_pairs, values=classifications, pair2index = classified_pair2index, binary=True)
			if(args.verbosity>=2): print2("  Note: After symmetrizing, kept %d unique unit pairs"%(len(classified_pairs)))
		unit_names = list(unit_name2int.keys()) # since unit_name2int grew incrementally with increasing int values, unit_names[unit_name2int[name]] will be equal to name.
		classifications = classifications.astype(bool) # ensure known classifications are booleans

		# load pairwise predictor data and synchronize it with the master set (which is based on the classifications data)
		if(args.verbosity): print2("Loading input predictors from %d files.."%(len(predictor_filepaths)))
		Xdata  = numpy.empty((len(classified_pairs),0), dtype=float) # prepare array to which loaded predictors will be appended, one input file at a time
		Xnames = []
		for ifp,input_filepath in enumerate(predictor_filepaths):
			if(args.verbosity>=2): print2("  Loading input predictors table #%d ('%s').."%(ifp,input_filepath))
			pairs,\
			values,\
			dummy,\
			column_names,\
			found_column_names = load_flat_pairwise_data(	filepath			= input_filepath,
															only_column_names	= args.only_predictors,
															omit_column_names	= args.omit_predictors,
															as_predictors		= True,
															numeric				= True,
															only_unit_names		= unit_name2int,
															unit_name2int		= unit_name2int,
															only_pairs			= classified_pair2index,
															verbose				= args.verbosity-1,
															verbose_prefix		= "    ")
			if(args.verbosity): print2("    Note: Loaded %d predictors for %d samples (unit pairs), covering %d distinct units"%(len(column_names),len(pairs),len(unit_name2int)))
			if(len(column_names)>0):
				if(args.symmetrize):
					pairs, values, dummy = symmetrize_data(pairs=pairs, values=values, binary=False)
					if(args.verbosity>=2): print2("    Note: After symmetrizing, kept predictors for %d unique unit pairs"%(len(pairs)))
				Xdata = numpy.pad(Xdata, pad_width=((0,0),(0,len(column_names))), constant_values=NaN)
				rows  = [classified_pair2index[pair] for pair in pairs]
				Xdata[rows,-len(column_names):] = values
				Xnames += column_names
				if((args.verbosity>=2) and (len(pairs)<len(classified_pairs))): 
					first_missing_pair = classified_pairs[integer_complement(len(classified_pairs), rows)[0]]
					print2("    Note: Predictors were missing for %d unit pairs, for example (%s, %s)"%(len(classified_pairs)-len(pairs),unit_names[first_missing_pair[0]],unit_names[first_missing_pair[1]]))
		

		if(args.verbosity>=1): print2("Checking non-finite predictor values..")
		if(args.replace_nan_predictors!=""):
			Xdata, Nreplaced = replace_nans_in_data(names=Xnames, values=Xdata, replacement_specifiers=args.replace_nan_predictors)
			if(args.verbosity>=2): print2("  Note: Replaced %d non-finite predictor values"%(Nreplaced))

		# check for any remaining predictor values that are non-finite
		if(args.remaining_nan_predictors=="error"):
			if(not numpy.all(numpy.isfinite(Xdata))):
				completely_nan = numpy.where(numpy.all(~numpy.isfinite(Xdata),axis=0))[0]
				if(len(completely_nan)>0): abort("ERROR: %d out of %d predictors were NaN/Inf in all samples. Are you sure these are supposed to be predictors?"%(len(completely_nan),Xdata.shape[1]))
				else: abort("ERROR: Found %d non-finite values in the input predictors. Consider adjusting --remaining_nan_predictors."%(numpy.sum(~numpy.isfinite(Xdata))))
		elif(args.remaining_nan_predictors=="omit_samples"):
			# only keep complete samples, i.e., with no missing predictor values
			keep_samples = numpy.where(numpy.all(numpy.isfinite(Xdata),axis=1))[0]
			if(len(keep_samples)<Xdata.shape[0]):
				if(args.verbosity>=2): print2("  Note: Omitting %d out of %d samples (unit pairs) due to missing/non-finite predictor data"%(Xdata.shape[0]-len(keep_samples),Xdata.shape[0]))
				Xdata 				= Xdata[keep_samples,:]
				classifications 	= classifications[keep_samples]
				classified_pairs 	= [classified_pairs[k] for k in keep_samples]
				classified_pair2index = {pair:p for p,pair in enumerate(classified_pairs)}
		elif(args.remaining_nan_predictors=="omit_predictors"):
			# only keep complete predictors, i.e., with no missing values
			keep_predictors = numpy.where(numpy.all(numpy.isfinite(Xdata),axis=0))[0]
			if(len(keep_predictors)<Xdata.shape[1]):
				if(args.verbosity>=2): print2("  Note: Omitting %d out of %d predictors due to missing/non-finite values"%(Xdata.shape[1]-len(keep_predictors),Xdata.shape[1]))
				Xdata 	= Xdata[:,keep_predictors]
				Xnames	= [Xnames[k] for k in keep_predictors]
		elif(args.remaining_nan_predictors=="zero"):
			# replace NaN/Inf predictor values with zero
			Ninvalids = numpy.sum(~numpy.isfinite(Xdata))
			if(Ninvalids>0):
				if(args.verbosity>=2): print2("  Note: Replacing %d non-finite predictor values with zero"%(Ninvalids))
				Xdata[~numpy.isfinite(Xdata)] = 0

		Nsamples = Xdata.shape[0]
		Xnames   = numpy.asarray(Xnames, dtype=object)
		if(args.verbosity>=1): print2("Note: Sanitized input data comprises %d samples (unit pairs) with known classification and %d predictor variables\n      %d out of %d classifications (%.4g %%) are ingroup"%(Nsamples,len(Xnames),sum(classifications),Nsamples,100*sum(classifications)/Nsamples))
	
		# at this point:
		#	Xdata is a 2D numeric array of size Nsamples x NX, listing predictor values for Nsamples unit pairs
		#	Xnames is a 1D array of length NX, listing the names of predictor variables
		#	classifications is a 1D boolean array of size Nsamples, listing binary classifications (ingroup True or False) for the various unit pairs
		#	classified_pairs is a list of length Nsamples, every element of which is an integer 2-tuple specifying the integer IDs of the compared units
	
		if((args.scatterplot_input_data!="") and (Xdata.shape[1]>=2)):
			if(args.verbosity>=1): print2("Scatterplotting input data..")
			shown_samples = (range(Nsamples) if (Nsamples<=args.max_Nscatter) else RNG.choice(Nsamples, size=args.max_Nscatter, replace=False))
			scatter_size  = max(10,500/max(10,math.sqrt(len(shown_samples))))
			check_output_file(file=args.scatterplot_input_data,force=args.force,verbose=(args.verbosity>=2),verbose_prefix=args.verbose_prefix+"  ")
			plot_target = PdfPages(args.scatterplot_input_data)
			for x1,Xname1 in enumerate(Xnames):
				for x2,Xname2 in enumerate(Xnames):
					if(x1==x2): continue
					fig = pyplot.figure(figsize=(args.plot_width+1.5, args.plot_height+1.5))
					pyplot.grid(visible=True, which='major', axis='both', alpha=0.2, color="#404040")
					pyplot.scatter(	Xdata[shown_samples,x1],
									Xdata[shown_samples,x2], 
									color		= [(SCATTER_COLOR_INGROUP if cl else SCATTER_COLOR_OUTGROUP) for cl in classifications[shown_samples]],
									s			= scatter_size,
									linewidth	= 0.5,
									marker		= 'o',
									alpha		= 0.9)
					pyplot.title("Input classifications vs predictors '%s' & '%s'\nN=%d%s\n%s"%(Xname1,Xname2,len(shown_samples),("" if (Nsamples<=args.max_Nscatter) else " (subsampled down from %d)"%(Nsamples)),args.case_tag),fontsize="medium", fontname="Helvetica")
					pyplot.xlabel(Xname1, fontsize="medium", fontname="Helvetica")
					pyplot.ylabel(Xname2, fontsize="medium", fontname="Helvetica", rotation=90,horizontalalignment="center")
					save_plot(fig=fig, plot_target=plot_target)				
			plot_target.close()			
					
			# also write plot data to a table file
			if(args.verbosity>=1): print2("Writing input data to a condensed file..")
			save_classical_table(output_file			= os.path.splitext(args.scatterplot_input_data)[0]+"_data.tsv.gz",
								table					= numpy.hstack((Xdata, numpy.expand_dims(classifications,axis=1))),
								column_names			= list(Xnames) + ["classification"],
								delimiter				= "\t",
								comment_prefix			= args.comment_prefix,
								comments				= "Input Xdata & classifications\nGenerated on: %s\nUsed command: %s"%(get_date_time(),get_shell_command()))


	
	
	##########################################
	# Load or select classifier model

	if(args.in_classifier!=""):
		if(args.verbosity): print2("Loading calibrated classifier from pickle file..")
		classifier = load_object(args.in_classifier, graceful=False)
		if(args.verbosity>=2): 
			print2("  Note: Loaded %s has true positive rate %.4g, true negative rate %.4g, accuracy %.4g, balanced accuracy %.4g"%(classifier.__class__.__name__,classifier.true_positive_rate,classifier.true_negative_rate,classifier.accuracy,classifier.balanced_accuracy))
			print2("  Loaded classifier uses the following %d predictors:\n    %s"%(len(classifier.Xnames),("\n    ").join(classifier.Xnames)))
			print2("  Loaded classifier structure:\n    %s"%(str(classifier).replace("\n","\n    ")))
		
	
	elif(args.in_classifications!=""):
		start_time	= time.time()

		NX = Xdata.shape[1]

		# rescale & centralize the input data for the model
		Xcenter = numpy.mean(Xdata,axis=0,keepdims=True)
		Xscale  = numpy.std(Xdata,axis=0,keepdims=True)
		Xmin	= numpy.min(Xdata,axis=0)
		Xmax	= numpy.max(Xdata,axis=0)
		if(numpy.any(Xscale==0)): abort("  ERROR: %d out of %d predictors are completely zero"%(numpy.sum(Xscale==0),Xscale.size))
		Xdata0 	= (Xdata-Xcenter)/Xscale # standardized version of the data
		
		# determine training/validation/test subsets
		# test_samples[] will be an integer list with values in 0,..,Nsamples-1, listing which samples to reserve for final testing
		# nontest_samples[] is the complement of test_samples[]
		# tv_subsets will be a list of integer 2-tuples of the form (training_samples, validation_samples), where the training_samples[] & validation_samples[] are disjoing integer lists with values in 0,..,len(nontest_samples), specifying which nontest_samples to use for training & validation in each cross-validation round
		if(args.verbosity): print2("Preparing training/validation/test sets..")
		test_samples	= RNG.choice(Nsamples, size=int(args.test_fraction*Nsamples), replace=False)
		nontest_samples = integer_complement(Nsamples, test_samples)
		tv_subsets		= [None]*args.NCV
		for cv in range(args.NCV):
			training_samples 	= RNG.choice(len(nontest_samples), size=int(args.training_fraction*Nsamples), replace=False)
			validation_samples 	= nontest_samples[integer_complement(len(nontest_samples), training_samples)]
			training_samples	= nontest_samples[training_samples]
			if(args.training_fraction+args.validation_fraction+args.test_fraction<1):
				# further subset validation_samples
				validation_samples = RNG.choice(validation_samples, size=int(args.validation_fraction*Nsamples), replace=False)
			if(args.rarefy_to_balance_for_fitting):
				if((args.verbosity>=2) and (cv==0)): print2("  Note: Rarefying training data to balance")
				training_samples = training_samples[balance_by_rarefying(classifications[training_samples])]
			if((args.max_training_count is not None) and (args.max_training_count<len(training_samples))):
				# further subsample training samples
				if((args.verbosity>=2) and (cv==0)): print2("  Note: Further rarefying training data down to %d samples"%(args.max_training_count))
				training_samples = RNG.choice(training_samples, size=args.max_training_count, replace=False)
			tv_subsets[cv] = (training_samples, validation_samples)
		nontest_samples_for_refitting = (nontest_samples[balance_by_rarefying(classifications[nontest_samples])] if args.rarefy_to_balance_for_fitting else nontest_samples)
		if(args.verbosity>=1): print2("  Test set comprises %d samples (%.4g %% of which are ingroup)\n  Training sets each comprise %d samples (%.4g %% of which are ingroup)\n  Validation sets each comprise %d samples (%.4g %% of which are ingroup)"%(len(test_samples),100*sum(classifications[test_samples])/len(test_samples),len(tv_subsets[0][0]),100*sum(classifications[tv_subsets[0][0]])/len(tv_subsets[0][0]),len(tv_subsets[0][1]),100*sum(classifications[tv_subsets[0][1]])/len(tv_subsets[0][1])))


		def try_predictor_subset(predictor_subset, tag=None, verbosity=0, verbose_prefix=""):
			if(verbosity>=1):
				if(tag is None): print2("%sTrying %d predictors {%s}.."%(verbose_prefix,len(predictor_subset),", ".join(Xnames[list(predictor_subset)])))
				else: print2("%sTrying predictor subset %s, with %d predictors: %s.."%(verbose_prefix,tag,len(predictor_subset),", ".join(Xnames[list(predictor_subset)])))
			if(args.model_type=="MLP"):
				# multilayer-perceptron, aka. feed-forward neural network
				if(args.hidden_layer_sizes==""):
					hidden_layer_sizes = [(n,) for n in numpy.unique(numpy.maximum(2,(numpy.linspace(0.1,1,5)*2*len(predictor_subset)).astype(int)))]
					hidden_layer_sizes += [(n1[0],n2) for n1 in hidden_layer_sizes for n2 in numpy.unique(numpy.maximum(1,numpy.linspace(0.1,1,5)*n1[0]).astype(int))]
				else:
					hidden_layer_sizes = list(ast.literal_eval(args.hidden_layer_sizes))
				classifier_template = MLPClassifier(random_state=args.seed, tol=1e-5, solver="lbfgs", max_iter=1000000, max_fun=100000)
				selection = sklearn.model_selection.GridSearchCV(	estimator 	= classifier_template,
																	param_grid 	= {	"activation":			args.neural_activations.split(","), 
																					"alpha":				[1e-5,1e-4,1e-3,1e-2,0.1,1,10], 
																					"hidden_layer_sizes":	hidden_layer_sizes},
																	scoring		= args.cv_scoring,
																	cv			= tv_subsets,
																	verbose 	= 0,
																	n_jobs		= Nthreads_per_candidate,
																	refit 		= False)															
			elif(args.model_type=="SGD"):
				# Stochastic Gradient Descend (regularized linear model)
				classifier_template = SGDClassifier(random_state=args.seed, fit_intercept=True, max_iter=10000000, tol=1e-5, class_weight=("balanced" if args.SGD_balanced_training_weights else None))
				selection = sklearn.model_selection.GridSearchCV(	estimator 	= classifier_template,
																	param_grid 	= {	"loss":				(["log_loss","modified_huber"] if args.probabilistic else ["squared_error","hinge","log_loss","modified_huber","squared_hinge","huber"]), 
																					"alpha":			[1e-5,1e-4,1e-3,1e-2,1e-1], 
																					"learning_rate":	args.SGD_learning_rates.split(","), 
																					"eta0":				[0.1,1,10]},
																	scoring		= args.cv_scoring,
																	cv			= tv_subsets,
																	verbose 	= 0,
																	n_jobs		= Nthreads_per_candidate,
																	refit 		= False)															
			elif(args.model_type=="RF"):
				# Random Forest
				classifier_template = RandomForestClassifier(random_state=args.seed)
				selection = sklearn.model_selection.GridSearchCV(	estimator 	= classifier_template,
																	param_grid 	= {	"criterion":		args.RF_criteria.split(","),
																					"max_features":		[None, "sqrt", "log2"],
																					"n_estimators":		[10, 100, 1000],
																					"min_samples_split":[2, 5, 10],
																					"max_depth":		[None, 10, 100]},
																	scoring		= args.cv_scoring,
																	cv			= tv_subsets,
																	verbose 	= 0,
																	n_jobs		= Nthreads_per_candidate,
																	refit 		= False)
			# train classifier and select hyperparameters
			selection.fit(Xdata0[:,predictor_subset],classifications)
			classifier = classifier_template.set_params(**selection.best_params_) # define uncalibrated classifier with selected hyperparameters
		
			# add some auxiliary info to the chosen classifier
			classifier.Xcenter 		= Xcenter[...,list(predictor_subset)]
			classifier.Xscale  		= Xscale[...,list(predictor_subset)]
			classifier.Xmin			= Xmin[list(predictor_subset)]
			classifier.Xmax			= Xmax[list(predictor_subset)]
			classifier.Xnames		= Xnames[list(predictor_subset)]
			classifier.score		= selection.best_score_
			classifier.Xsubset		= predictor_subset
			classifier.hyperparams	= selection.best_params_
			return classifier

	
		# auxiliary function for printing basic information about a classifier
		def aux_sprint_classifier(classifier):
			S = "score %.5g, %d predictors {%s}"%(classifier.score,len(classifier.Xnames),", ".join(classifier.Xnames))
			return S

		subset2selection 			= {} # every key in subset2selection[] will specify a unique subset of predictors examined as an integer tuple, and map to the chosen classifier (i.e., with optimized hyperparameters)
		best_classifier	 			= IDict(score=-Inf, Xsubset=(), Xnames=[])
		previous_iteration_score 	= -Inf
		score_change 				= NaN
		maxNX 			 			= (NX if (args.predictor_selection_maxNX is None) else min(NX,args.predictor_selection_maxNX))
		minNX 			 			= max(0,min(maxNX,args.predictor_selection_minNX))
		if(args.verbosity>=1): print2("Selecting and training binary classifier..")
		if(args.verbosity>=2): print2("  Note: Selecting between %d and %d predictors, from a set of %d candidates"%(minNX,maxNX,NX))
		for iteration in range(args.predictor_selection_max_iterations):
			if(args.verbosity>=3): print2("    Iteration %d.."%(iteration))
			if(args.verbosity>=4): print2("      Note: Current classifier %s"%(aux_sprint_classifier(best_classifier)))
			score_change  = 0
			# Stage 1: Add predictors
			# determine all possible subsets of predictors that we may add, and from those construct the corresponding subsets of predictors to consider.
			# candidate_subsets[] will be a list specifying all possible combinations of predictors that may be considered in this iteration. Hence, each element in candidate_subsets[] is an integer tuple listing predictor indices that may be considered.
			# Each combination of added predictors must satisfy various constraints: the number of added predictors must be between min_addedX & max_addedX, and the total number of predictors added must be <=max_addedX.
			availableX = integer_complement(NX,best_classifier.Xsubset)
			min_addedX = max(0, minNX-len(best_classifier.Xsubset)) # add at least as many X predictors as needed to reach minNX
			max_addedX = max(min_addedX, min(maxNX - len(best_classifier.Xsubset), args.predictor_selection_max_added))
			candidate_additions = list(itertools.chain.from_iterable(itertools.combinations(availableX, nax) for nax in range(min_addedX,max_addedX+1)))
			# convert candidate predictor additions to candidate predictor subsets in the form of integer tuples, omitting any trivial candidates (i.e., those that are not actually adding any new predictors).
			candidate_subsets = [tuple(sorted(best_classifier.Xsubset + candidate_addition)) for candidate_addition in candidate_additions if (len(candidate_addition)>0)]
			# omit any candidate subsets already examined previously
			candidate_subsets = [candidate_subset for candidate_subset in candidate_subsets if (candidate_subset not in subset2selection)]
			Ncandidates = len(candidate_subsets)
			# examine all candidate subsets
			if(Ncandidates==0):
				if(args.verbosity>=4): print2("      No predictor additions to try in this iteration (all have been previously examined)")
			else:
				if(args.verbosity>=4): print2("      Trying %d different predictor additions.."%(Ncandidates))
				Nthreads_across_candidates	= min(Ncandidates,args.Nthreads)
				Nthreads_per_candidate 		= max(1,int(args.Nthreads/Nthreads_across_candidates))
				selections 					= run_parallel(function=(lambda ca: try_predictor_subset(candidate_subsets[ca],tag="#%d"%(ca),verbosity=args.verbosity-3,verbose_prefix="      ")), arguments=range(Ncandidates), Nthreads=Nthreads_across_candidates, fast=False)
				subset2selection.update({subset:selection for subset,selection in zip(candidate_subsets,selections)}) # update the set of examined subsets
				best_selection = get_largest_item([selection.score for selection in selections])
				if((best_selection>=0) and (selections[best_selection].score>best_classifier.score)):
					best_classifier	= selections[best_selection]
					if(args.verbosity>=4): print2("      Note: Best predictor combo so far: %s"%(aux_sprint_classifier(best_classifier)))
				elif(args.predictor_selection_max_removed==0):
					if(args.verbosity>=4): print2("      Note: None of the added predictor sets improved the score, and backtracking is disabled, so halting search")
					break

			# Stage 2: Remove predictors, if applicable
			# determine all possible subsets of predictors that we may remove
			# candidate_subsets[] will be a list specifying all possible combinations of predictors that may be considered in this iteration. Hence, each element in candidate_subsets[] is an integer tuple listing predictor indices that may be considered.
			if(args.predictor_selection_max_removed>0):
				max_removedX 		= min(len(best_classifier.Xsubset)-minNX, args.predictor_selection_max_removed)
				candidate_removals 	= list(itertools.chain.from_iterable(itertools.combinations(best_classifier.Xsubset, nrx) for nrx in range(max_removedX+1)))
				# convert candidate predictor removals to candidate predictor subsets as integer tuples, omitting any trivial candidates (i.e., those that are not actually removing any predictors).
				candidate_subsets = [tuple(sorted(set(best_classifier.Xsubset).difference(candidate_removal))) for candidate_removal in candidate_removals if (len(candidate_removal)>0)]
				# omit any candidate subsets already examined previously
				candidate_subsets = [candidate_subset for candidate_subset in candidate_subsets if (candidate_subset not in subset2selection)]
				Ncandidates = len(candidate_subsets)
				# examine all candidate subsets			
				if(Ncandidates==0):
					if(args.verbosity>=4): print2("      No predictor removals to try in this iteration (all have been previously examined)")
				else:
					if(args.verbosity>=4): print2("      Trying %d different predictor removals.."%(Ncandidates))
					Nthreads_across_candidates	= min(Ncandidates,args.Nthreads)
					Nthreads_per_candidate 		= max(1,int(args.Nthreads/Nthreads_across_candidates))
					selections					= run_parallel(function=(lambda ca: try_predictor_subset(candidate_subsets[ca],tag="#%d"%(ca),verbosity=args.verbosity-3,verbose_prefix="      ")), arguments=range(Ncandidates), Nthreads=Nthreads_across_candidates, fast=False)
					best_selection 				= get_largest_item([selection.score for selection in selections])
					subset2selection.update({subset:selection for subset,selection in zip(candidate_subsets,selections)}) # update the set of examined subsets
					if((best_selection>=0) and (selections[best_selection].score>best_classifier.score)):
						best_classifier	= selections[best_selection]
					else:
						if(args.verbosity>=4): print2("      Note: None of the predictor removals improved the score")
				if(args.verbosity>=4): print2("      Note: Best predictor combo so far: %s"%(aux_sprint_classifier(best_classifier)))

			# check if we should do another iteration
			score_change = best_classifier.score - previous_iteration_score
			previous_iteration_score = best_classifier.score
			if(score_change<=args.predictor_selection_score_change_threshold):
				if(args.verbosity>=2): print2("  Search converged at %s, last score change %g"%(aux_sprint_classifier(best_classifier),score_change))
				break
			elif(iteration==args.predictor_selection_max_iterations-1):
				if(args.verbosity>=2): print2("  Reached maximum number of iterations (%d). Stopping at %s"%(max_iterations,aux_sprint_classifier(best_classifier)))
		if(not numpy.isfinite(best_classifier.score)): abort("  ERROR: Predictor selection failed for an unknown reason")
		classifier = best_classifier
		
		if(args.verbosity>=1):
			print2("Refitting best classifier with non-test set (%d samples%s).."%(len(nontest_samples_for_refitting),(", rarefied down from %d samples to balance"%(len(nontest_samples)) if args.rarefy_to_balance_for_fitting else "")))
			classifier.fit(Xdata0[numpy.ix_(nontest_samples_for_refitting,classifier.Xsubset)],classifications[nontest_samples_for_refitting])
		
		if(args.verbosity>=1): 
			print2("Chosen best classifier has score %.5g, comprises the following %d predictors:\n  %s"%(classifier.score,len(classifier.Xsubset),("\n  ").join(classifier.Xnames)))
			print2("Classifier structure:\n  %s"%(str(classifier).replace("\n","\n  ")))
			print2("Note: Model selection took %g seconds"%(time.time()-start_time))

		# auxiliary function for evaluating a classifier on the test set
		test_classifications = classifications[test_samples]
		def evaluate_on_test_set(classifier):
			test_predictions 	 = classifier.predict(Xdata0[test_samples,:][:,classifier.Xsubset])
			true_negative_rate	 = ratio_nansafe(numpy.sum(~test_predictions[~test_classifications]),numpy.sum(~test_classifications))
			true_positive_rate	 = ratio_nansafe(numpy.sum(test_predictions[test_classifications]),numpy.sum(test_classifications))
			accuracy			 = numpy.sum(test_predictions==test_classifications)/len(test_classifications)
			balanced_accuracy	 = 0.5*(true_negative_rate+true_positive_rate)
			return IDict(test_predictions	= test_predictions, 
						 true_negative_rate	= true_negative_rate, 
						 true_positive_rate	= true_positive_rate,
						 accuracy			= accuracy,
						 balanced_accuracy	= balanced_accuracy,
						 Nsamples			= len(test_predictions))

		if(args.verbosity>=1): print2("Evaluating best classifier on test set (%d samples).."%(len(test_samples)))
		evaluation = evaluate_on_test_set(classifier)
		if(args.verbosity>=1): print2("  --> True positive rate (sensitivity) %.5g, true negative rate (specificity) %.5g, accuracy %.5g, balanced accuracy %.5g"%(evaluation.true_positive_rate,evaluation.true_negative_rate,evaluation.accuracy,evaluation.balanced_accuracy))
		classifier.true_positive_rate 	= evaluation.true_positive_rate
		classifier.true_negative_rate 	= evaluation.true_negative_rate
		classifier.accuracy 			= evaluation.accuracy
		classifier.balanced_accuracy	= evaluation.balanced_accuracy
		
		if(args.verbosity>=1): 
			# examine the accuracy that would be achieved by a random biased classifier (null model), i.e. randomly classifying unit pairs by drawing from a Bernoulli distribution whose probability reflects the classification frequencies in the data
			positive_rate		 = numpy.mean(classifications)
			test_set_probability = numpy.mean(classifications[test_samples])
			null_evaluation 	 = IDict(true_positive_rate	= positive_rate,
										 true_negative_rate	= (1-positive_rate), 
										 accuracy			= (1-test_set_probability)*(1-positive_rate) + test_set_probability*positive_rate,
										 balanced_accuracy	= 0.5)
			print2("Note: Biased random classifier (with positive rate %g) would achieve true positive rate %.5g, true negative rate %.5g, accuracy %.5g, balanced accuracy %.5g"%(positive_rate,null_evaluation.true_positive_rate,null_evaluation.true_negative_rate,null_evaluation.accuracy,null_evaluation.balanced_accuracy))

		if(args.verbosity>=1): 
			# examine the accuracy that would be achieved by a Dirac classifier, i.e. classifying all unit pairs as the most common state seen in the data
			positive_rate		 = (1 if (numpy.mean(classifications)>=0.5) else 0)
			test_set_probability = numpy.mean(classifications[test_samples])
			null_evaluation 	 = IDict(true_positive_rate	= positive_rate,
										 true_negative_rate	= (1-positive_rate), 
										 accuracy			= (1-test_set_probability)*(1-positive_rate) + test_set_probability*positive_rate,
										 balanced_accuracy	= 0.5)
			print2("Note: Dirac classifier (with positive rate %d) would achieve true positive rate %.5g, true negative rate %.5g, accuracy %.5g, balanced accuracy %.5g"%(positive_rate,null_evaluation.true_positive_rate,null_evaluation.true_negative_rate,null_evaluation.accuracy,null_evaluation.balanced_accuracy))
		
		# save test results to table file
		if(args.out_test_results!=""):
			if(args.verbosity>=1): print2("Saving test results as scatterpoints..")
			test_table = numpy.column_stack(([unit_names[classified_pairs[s][0]] for s in test_samples],[unit_names[classified_pairs[s][1]] for s in test_samples],test_classifications.astype(int),evaluation.test_predictions.astype(int)))
			if(args.probabilistic): test_table = numpy.hstack((test_table, classifier.predict_proba(Xdata0[test_samples,:][:,classifier.Xsubset])[:,[1]]))
			test_table = numpy.hstack((test_table,Xdata[test_samples,:][:,classifier.Xsubset]))
			save_classical_table(output_file			= args.out_test_results,
								table					= test_table,
								column_names			= ["unit1","unit2","true_classification","predicted_classification"]+(["predicted_classification_probability"] if args.probabilistic else [])+list(classifier.Xnames),
								delimiter				= args.delimiter,
								comment_prefix			= args.comment_prefix,
								comments				= join_nonempty(["Test results for calibrated classifier, after re-training on the non-test-data (%d samples%s)\nAlso included are the predictor values used as input to the classifier for the predictions\nGenerated on: %s\nUsed command: %s"%(len(nontest_samples_for_refitting),(", rarefied down from %d samples to balance"%(len(nontest_samples)) if args.rarefy_to_balance_for_fitting else ""),get_date_time(),get_shell_command()), args.case_tag], sep="\n"))
		
		# scatterplot test results, coloring points according to their true classifications (fill color) and predicted classifications (edge color)
		if((args.scatterplot_test_results!="") and (len(classifier.Xsubset)>=2)):
			if(args.verbosity>=1): print2("Scatterplotting test results..")
			shown_test_samples = (range(len(test_samples)) if (len(test_samples)<=args.max_Nscatter) else RNG.choice(len(test_samples), size=args.max_Nscatter, replace=False))
			scatter_size= max(10,500/max(10,math.sqrt(len(shown_test_samples))))
			check_output_file(file=args.scatterplot_test_results,force=args.force,verbose=(args.verbosity>=2),verbose_prefix=args.verbose_prefix+"  ")
			plot_target = PdfPages(args.scatterplot_test_results)
			for x1,Xname1 in zip(classifier.Xsubset,classifier.Xnames):
				for x2,Xname2 in zip(classifier.Xsubset,classifier.Xnames):
					if(x1==x2): continue
					fig = pyplot.figure(figsize=(args.plot_width+1.5, args.plot_height+1.5))
					pyplot.grid(visible=True, which='major', axis='both', alpha=0.2, color="#404040")
					pyplot.scatter(	Xdata[test_samples[shown_test_samples],x1],
									Xdata[test_samples[shown_test_samples],x2], 
									color		= [(SCATTER_COLOR_INGROUP if cl else SCATTER_COLOR_OUTGROUP) for cl in evaluation.test_predictions[shown_test_samples]],
									s			= scatter_size,
									linewidth	= 0,
									marker		= 'o',
									alpha		= 0.9)
					pyplot.scatter(	Xdata[test_samples[shown_test_samples],x1],
									Xdata[test_samples[shown_test_samples],x2], 
									color		= [(SCATTER_COLOR_INGROUP if cl else SCATTER_COLOR_OUTGROUP) for cl in test_classifications[shown_test_samples]],
									s			= 0.2*scatter_size,
									linewidth	= 0,
									marker		= 'o',
									alpha		= 0.9)
					pyplot.title(join_nonempty(["Classifier test results, vs predictors '%s' & '%s'\nN=%d (subsampled to %d), accuracy=%.4g, balanced accuracy=%.4g\nFill colors: known classifications, edge colors: predictions"%(Xname1,Xname2,evaluation.Nsamples,len(shown_test_samples),evaluation.accuracy,evaluation.balanced_accuracy)], sep="\n"),fontsize="medium", fontname="Helvetica")
					pyplot.xlabel(Xname1, fontsize="medium", fontname="Helvetica")
					pyplot.ylabel(Xname2, fontsize="medium", fontname="Helvetica", rotation=90,horizontalalignment="center")
					save_plot(fig=fig, plot_target=plot_target)				
			plot_target.close()			
	
		if(args.fit_final_using=="all_data"):
			if(args.rarefy_to_balance_for_fitting):
				balanced_samples = balance_by_rarefying(classifications)
				if(args.verbosity>=1): print2("Refitting classifier with full dataset (%d samples, rarefied from originally %d samples to balance).."%(len(balanced_samples),Xdata.shape[0]))
				classifier.fit(Xdata0[numpy.ix_(balanced_samples,classifier.Xsubset)], classifications[balanced_samples])
			else:
				if(args.verbosity>=1): print2("Refitting classifier with full dataset (%d samples).."%(Xdata.shape[0]))
				classifier.fit(Xdata0[:,classifier.Xsubset], classifications)

		if((args.out_classifier!="") and (os.path.abspath(args.out_classifier)!=os.path.abspath(args.in_classifier))):
			if(args.verbosity>=1): print2("Saving final classifier to file..")
			check_output_file(args.out_classifier, force=args.force, verbose=(args.verbosity>=1), verbose_prefix=args.verbose_prefix)
			save_object(obj=classifier, filepath=args.out_classifier)


		# Forward sensitivity analysis
		if(args.analyze_forward_sensitivities): 
			if(args.verbosity>=1): print2("Performing forward sensitivity analysis (considering only a single predictor at a time)..")
			single_predictor_evaluations = [None]*Xdata.shape[1]
			for p in range(Xdata.shape[1]):
				if(args.verbosity>=2): print2("  Considering only predictor '%s'.."%(Xnames[p]))
				if((p,) in subset2selection):
					single_predictor_classifier = subset2selection[(p,)]
				else:
					single_predictor_classifier = try_predictor_subset((p,),verbosity=args.verbosity-2,verbose_prefix="    ")
				single_predictor_classifier.fit(Xdata0[numpy.ix_(nontest_samples_for_refitting,single_predictor_classifier.Xsubset)],classifications[nontest_samples_for_refitting])
				single_predictor_evaluation 	= evaluate_on_test_set(single_predictor_classifier)
				single_predictor_evaluations[p] = single_predictor_evaluation
				if(args.verbosity>=2):
					print2("    Performance of best classifier, selected with only predictor '%s', evaluated on test set:"%(Xnames[p]))
					print2("      True positive rate (sensitivity) %.5g, true negative rate (specificity) %.5g, accuracy %.5g, balanced accuracy %.5g"%(single_predictor_evaluation.true_positive_rate,single_predictor_evaluation.true_negative_rate,single_predictor_evaluation.accuracy,single_predictor_evaluation.balanced_accuracy))
			if(args.out_forward_sensitivities!=""):
				if(args.verbosity>=1): print2("Saving forward sensitivities to file..")
				save_classical_table(output_file			= args.out_forward_sensitivities,
									table					= numpy.asarray([(ev.true_positive_rate,ev.true_negative_rate,ev.accuracy,ev.balanced_accuracy,1-ev.true_negative_rate,1-ev.true_positive_rate) for ev in single_predictor_evaluations]),
									row_names				= Xnames,
									column_names			= ["true_positive_rate","true_negative_rate","accuracy","balanced_accuracy","false_positive_rate","false_negative_rate"],
									row_names_column		= "predictor",
									delimiter				= args.delimiter,
									comment_prefix			= args.comment_prefix,
									comments				= join_nonempty(["Forward sensitivities w.r.t. %d predictors, i.e., performances of single-predictor models\nGenerated on: %s\nUsed command: %s"%(Xdata.shape[1],get_date_time(),get_shell_command()), args.case_tag], sep="\n"))
				
			
		# Backward sensitivity analysis: Omit one predictor at a time from the classifier and assess the impact on the classifier's performance (evaluated on the test set)
		# Hyperparameters are selected/optimized from sratch each time
		if(args.analyze_backward_sensitivities and (len(classifier.Xsubset)>1)):
			if(args.verbosity>=1): print2("Performing backward sensitivity analysis (leaving out one predictor at a time)..")
			reduced_evaluations = [None]*len(classifier.Xsubset)
			for p in range(len(classifier.Xsubset)):
				if(args.verbosity>=2): print2("  Leaving out predictor '%s'.."%(classifier.Xnames[p]))
				reduced_Xsubset			= classifier.Xsubset[:p]+classifier.Xsubset[(p+1):]
				if(reduced_Xsubset in subset2selection):
					reduced_classifier = subset2selection[reduced_Xsubset]
				else:
					reduced_classifier = try_predictor_subset(reduced_Xsubset,verbosity=args.verbosity-2,verbose_prefix="    ")
				reduced_classifier.fit(Xdata0[numpy.ix_(nontest_samples_for_refitting,reduced_classifier.Xsubset)],classifications[nontest_samples_for_refitting])
				reduced_evaluation		= evaluate_on_test_set(reduced_classifier)
				reduced_evaluations[p] 	= reduced_evaluation
				if(args.verbosity>=2):
					print2("    Performance of best classifier, selected with modified (reduced) predictor set, evaluated on test set:")
					print2("      True positive rate (sensitivity) %.5g, true negative rate (specificity) %.5g, accuracy %.5g, balanced accuracy %.5g"%(reduced_evaluation.true_positive_rate,reduced_evaluation.true_negative_rate,reduced_evaluation.accuracy,reduced_evaluation.balanced_accuracy))
					print2("      True positive rate decreased by %.5g, true negative rate by %.5g, accuracy by %.5g, balanced accuracy by %.5g"%(evaluation.true_positive_rate-reduced_evaluation.true_positive_rate,evaluation.true_negative_rate-reduced_evaluation.true_negative_rate,evaluation.accuracy-reduced_evaluation.accuracy,evaluation.balanced_accuracy-reduced_evaluation.balanced_accuracy))
			if(args.out_backward_sensitivities!=""):
				if(args.verbosity>=1): print2("Saving backward sensitivities to file..")
				save_classical_table(output_file			= args.out_backward_sensitivities,
									table					= numpy.asarray([(ev.true_positive_rate,ev.true_negative_rate,ev.accuracy,ev.balanced_accuracy,1-ev.true_negative_rate,1-ev.true_positive_rate,evaluation.true_positive_rate-ev.true_positive_rate,evaluation.true_negative_rate-ev.true_negative_rate,evaluation.accuracy-ev.accuracy,evaluation.balanced_accuracy-ev.balanced_accuracy) for ev in reduced_evaluations]),
									row_names				= classifier.Xnames,
									column_names			= ["true_positive_rate","true_negative_rate","accuracy","balanced_accuracy","false_positive_rate","false_negative_rate","reduction_in_true_positive_rate","reduction_in_true_negative_rate","reduction_in_accuracy","reduction_in_balanced_accuracy"],
									row_names_column		= "removed_predictor",
									delimiter				= args.delimiter,
									comment_prefix			= args.comment_prefix,
									comments				= join_nonempty(["Backward sensitivities w.r.t. %d predictors, i.e., performances of modified model after removing one predictor at a time re-selecting hyperparameters\nGenerated on: %s\nUsed command: %s"%(Xdata.shape[1],get_date_time(),get_shell_command()), args.case_tag], sep="\n"))
	else:
		classifier = None	
	
	if((classifier is not None) and (args.plot_classifier_phase_diagram!="")):
		if(len(classifier.Xsubset)!=2):
			if(args.verbosity>=1): print2("Note: Cannot plot the classifier's phase diagram, because the classifier uses %d predictors instead of 2"%(len(classifier.Xsubset)))
		else:
			if(args.verbosity>=1): print2("Plotting the classifier's phase diagram..")
			X1grid	   = numpy.linspace(classifier.Xmin[0],classifier.Xmax[0],num=200)
			X2grid	   = numpy.linspace(classifier.Xmin[1],classifier.Xmax[1],num=200)
			Xmeshgrid  = numpy.column_stack(tuple(xm.flat for xm in numpy.meshgrid(X1grid, X2grid, copy=False, indexing="ij")))
			Xmeshgrid0 = (Xmeshgrid - classifier.Xcenter)/classifier.Xscale # Apply rescalings prior to predictions with new data
			plot_contour_grid(	plot_target			= args.plot_classifier_phase_diagram,
								X					= X1grid,
								Y					= X2grid,
								Z					= (classifier.predict_proba(Xmeshgrid0)[:,1] if args.probabilistic else classifier.predict(Xmeshgrid0)).reshape((X1grid.size,X2grid.size)),
								show_colorbar		= args.probabilistic,
								colormap			= LinearSegmentedColormap.from_list("mycmap", [SCATTER_COLOR_OUTGROUP,SCATTER_COLOR_INGROUP]+([] if args.probabilistic else [SCATTER_COLOR_INGROUP])),
								minZ				= 0,
								maxZ				= 1,
								color_levels		= ("linear:15" if args.probabilistic else [0,1]),
								contour_values		= ([0.1,0.5,0.9] if args.probabilistic else None),
								curve_color			= None,
								xlabel				= classifier.Xnames[0],
								ylabel				= classifier.Xnames[1],
								clabel				= ("probability" if args.probabilistic else "classification"),
								plot_title			= join_nonempty([classifier.__class__.__name__+" phase diagram",args.case_tag], sep="\n"),
								plot_width			= args.plot_width,
								plot_height			= args.plot_height)
			del X1grid,X2grid,Xmeshgrid,Xmeshgrid0
								
	if((classifier is not None) and (args.out_classifier_phase_diagram!="")):
		if(args.verbosity>=1): print2("Saving the classifier's phase diagram to file..")
		Xgrids 	   = [numpy.linspace(classifier.Xmin[k],classifier.Xmax[k],num=200) for k in range(len(classifier.Xsubset))]
		Xmeshgrid  = numpy.column_stack(tuple(xm.flat for xm in numpy.meshgrid(*Xgrids, copy=False, indexing="ij")))
		Xmeshgrid0 = (Xmeshgrid - classifier.Xcenter)/classifier.Xscale # Apply rescalings prior to predictions with new data
		save_classical_table(	output_file				= args.out_classifier_phase_diagram,
								table					= numpy.hstack((Xmeshgrid0, (classifier.predict_proba(Xmeshgrid0)[:,1] if args.probabilistic else classifier.predict(Xmeshgrid0)).reshape((Xmeshgrid.shape[0],1)))),
								column_names			= list(classifier.Xnames)+["classification"],
								delimiter				= args.delimiter,
								comment_prefix			= args.comment_prefix,
								comments				= join_nonempty(["Phase diagram of %s, evaluated on a %dD regular meshgrid\nGenerated on: %s\nUsed command: %s"%(classifier.__class__.__name__,len(classifier.Xsubset),get_date_time(),get_shell_command()), args.case_tag], sep="\n"))
		del Xgrids,Xmeshgrid,Xmeshgrid0

	##########################################################################################
	# Predict ingroup/outgroup classifications with previously calibrated or loaded classifier
	# Predictor data and any available classifications are reloaded, to avoid any filters that may not be relevant anymore
	
	if(args.diarize=="none"):
		if(args.verbosity): print2("No diarization requested, so we're done.")
		sys.exit(0)
	elif(classifier is None):
		abort("ERROR: No classifier available, so diarization is not possible.")

	# load pairwise predictor data
	if(args.verbosity>=1): print2("Loading predictors used by classifier, including for any non-classified unit pairs, from %d files.."%(len(predictor_filepaths)))
	all_pairs		= [None]*len(predictor_filepaths)
	all_values		= [None]*len(predictor_filepaths)
	all_varnames	= [None]*len(predictor_filepaths)
	prediction_pairs= set()
	unit_name2int 	= {}
	for i,input_filepath in enumerate(predictor_filepaths):
		if(args.verbosity>=2): print2("  Loading input predictor table '%s'.."%(input_filepath))
		all_pairs[i],\
		all_values[i],\
		unit_name2int,\
		all_varnames[i],\
		found_column_names = load_flat_pairwise_data(	filepath			= input_filepath,
														only_column_names	= classifier.Xnames,
														as_predictors		= True,
														numeric				= True,
														unit_name2int		= unit_name2int,
														skip_if_no_columns	= True,
														verbose				= args.verbosity,
														verbose_prefix		= "  ")
		if(len(all_varnames[i])>0): prediction_pairs.update(all_pairs[i])
		if(args.verbosity>=2): 
			if(len(all_varnames[i])==0): print2("    Note: No data loaded from this file")
			else: print2("    Note: Loaded %d predictors for %d unit pairs, %d units"%(len(all_varnames[i]),len(all_pairs[i]),len(unit_name2int)))

	# Merge all loaded predictor values as a 2D master array, whose rows correspond to samples (unit pairs) and whose columns correspond to predictor variables (synchronized with classifier.Xnames)
	if(args.verbosity): print2("Consolidating all loaded predictors across %d unit pairs.."%(len(prediction_pairs)))
	varnames = [varname for varnames in all_varnames for varname in varnames]
	if(len(duplicate_varnames:=find_duplicates(varnames,indices=False))>0): abort("ERROR: Found duplicate predictor names: %s"%(duplicate_varnames))
	elif((varnames_set:=set(varnames))!=set(classifier.Xnames)): abort("ERROR: Failed to load all predictor names needed for classifier:\n  %s"%(("\n  ").join((name+(": found" if (name in varnames_set) else ": not found")) for name in classifier.Xnames)))
	varname2int			= {varname:v for v,varname in enumerate(classifier.Xnames)}
	prediction_pairs	= sorted(prediction_pairs)
	pair2sample			= {pair:p for p,pair in enumerate(prediction_pairs)}
	Xdata 				= numpy.full((len(prediction_pairs),len(classifier.Xnames)), NaN, dtype=float)
	for pairs,values,varnames in zip(all_pairs,all_values,all_varnames):
		Xdata[numpy.ix_([pair2sample[pair] for pair in pairs],[varname2int[varname] for varname in varnames])] = values
	del varname2int, varnames, all_pairs, all_values, all_varnames # free some RAM and avoid bugs later on

	# check for non-finite predictor values
	if(args.remaining_nan_predictors=="error"):
		if(not numpy.all(numpy.isfinite(Xdata))): abort("ERROR: Found %d non-finite values in the input predictors. Consider adjusting --remaining_nan_predictors."%(numpy.sum(~numpy.isfinite(Xdata))))
	elif(args.remaining_nan_predictors=="omit_samples"):
		# only keep complete samples, i.e., with no missing predictor values
		keep_samples = numpy.where(numpy.all(numpy.isfinite(Xdata),axis=1))[0]
		if(len(keep_samples)<Xdata.shape[0]):
			if(args.verbosity>=1): print2("Note: Omitting %d out of %d samples due to missing/non-finite predictor data"%(Xdata.shape[0]-len(keep_samples),Xdata.shape[0]))
			Xdata 				= Xdata[keep_samples,:]
			prediction_pairs	= [prediction_pairs[s] for s in keep_samples]
			pair2sample			= {pair:p for p,pair in enumerate(prediction_pairs)}
	elif(args.remaining_nan_predictors=="omit_predictors"):
		# requested to only keep complete predictors, i.e., with no missing values
		incomplete_predictors = numpy.where(~numpy.all(numpy.isfinite(Xdata),axis=0))[0]
		if(len(incomplete_predictors)>0): abort("ERROR: %d out of %d predictors (e.g. '%s') contain non-finite values and their removal was requested, however this makes it impossible to apply the calibrated classifier"%(len(incomplete_predictors),Xdata.shape[1],classifier.Xnames[incomplete_predictors[0]]))
	elif(args.remaining_nan_predictors=="zero"):
		# replace NaN/Inf predictor values with zero
		Ninvalids = numpy.sum(~numpy.isfinite(Xdata))
		if(Ninvalids>0):
			if(args.verbosity>=1): print2("Note: Replacing %d non-finite predictor values with zero"%(Ninvalids))
			Xdata[~numpy.isfinite(Xdata)] = 0

	if(args.symmetrize): 
		# symmetrize loaded predictors, i.e., so that only pairs (a,b) with a<=b are kept
		# pair2sample will then map both (a,b) and (b,a) to the same integer
		prediction_pairs, Xdata, pair2sample = symmetrize_data(pairs=prediction_pairs, values=Xdata, pair2index = pair2sample, binary=False)
		if(args.verbosity): print2("Note: After symmetrizing, kept predictors for %d unique samples"%(len(prediction_pairs)))
		
	# load any available known classifications and synchronize with predictors
	# classified_samples[] will be an integer array, with values in 0,..,Xdata.shape[0], specifying for which unit pairs we were able to load classifications for
	# classifications[] will be an integer array with values in {0,1}, synchronized with classified_samples[], specifying the known classifications of the corresponding unit pairs
	if((args.in_classifications=="") or ((args.diarize=="all") and args.reclassify_knowns)):
		# no need to load known classifications, or none available
		classified_samples 	= []
		classifications 	= numpy.empty(0, dtype=float)
	else:
		if(args.verbosity>=1): print2("Loading known classifications from file..")
		classified_pairs,\
		classifications,\
		unit_name2int,\
		dummy,\
		found_column_names = load_flat_pairwise_data(filepath			= args.in_classifications,
													only_column_names	= args.classification_column,
													numeric				= False,
													only_unit_names		= (unit_name2int if (args.diarize=="only_pairs_with_predictors") else None),
													unit_name2int		= unit_name2int,
													only_pairs			= (pair2sample if (args.diarize=="only_pairs_with_predictors") else None),
													verbose				= max(0,args.verbosity-1),
													verbose_prefix		= "  ")
		classifications = numpy.asarray([string2classification.get(cl[0].lower(),-1) for cl in classifications], dtype=int)
		# omit any missing/unknown classifications
		keep = numpy.where(classifications>=0)[0]
		if(len(keep)<len(classifications)):
			classifications  = classifications[keep]
			classified_pairs = [classified_pairs[k] for k in keep]
		if(args.verbosity>=2): print2("  Note: Loaded classifications for %d unit pairs"%(classifications.size))
		if(args.symmetrize): 
			# symmetrize loaded classifications, i.e., so that only pairs (a,b) with a<=b are kept
			classified_pairs, classifications, dummy = symmetrize_data(pairs=classified_pairs, values=classifications, binary=True)
			if(args.verbosity>=2): print2("  Note: After symmetrizing, kept predictors for %d unique unit pairs"%(len(classified_pairs)))
		# map classifications to rows of Xdata where possible, expand Xdata if needed for new unit pairs
		if(args.diarize=="all"):
			# we may have encountered classifications for new units and new unit pairs, i.e., not yet represented in Xdata & pair2sample
			new_pairs = [p for p,pair in enumerate(classified_pairs) if (pair not in pair2sample)]
			if(len(new_pairs)>0):
				if(args.verbosity>=2): print2("  Note: %d of the loaded classifications correspond to new units and/or unit pairs, i.e., not encountered in in_predictors"%(len(new_pairs)))
				if(args.reclassify_knowns): abort("ERROR: Reclassification is requested for all unit pairs, but this is not possible for those without predictors")
				Xdata 			  = numpy.pad(Xdata, pad_width=((0,len(new_pairs)),(0,0)), constant_values=NaN)
				prediction_pairs += [classified_pairs[p] for p in new_pairs]
				pair2sample		  = {pair:p for p,pair in enumerate(prediction_pairs)}
		elif(args.diarize=="only_pairs_with_predictors"):
			# we can assume that classifications were only loaded for unit pairs with predictors, i.e., already represented in Xdata & pair2sample
			pass
		classified_samples = [pair2sample[pair] for pair in classified_pairs] # 1D array of size len(classifications), listing the row indices of Xdata for which classifications are known/provided
		del classified_pairs

	unit_names = list(unit_name2int.keys()) # since unit_name2int grew incrementally with increasing int values, unit_names[unit_name2int[name]] will be equal to name.
	Nunits = len(unit_names)
	
	
	# load any available known unit-->group assignments and synchronize with our inventory of units
	if(args.in_Kgroups!=""):
		if(args.verbosity>=1): print2("Loading known group assignments from file..")
		Kgroup_table,\
		column_names,\
		found_column_names = read_table(filepath			= args.in_Kgroups,
										delimiter			= args.delimiter,
										comment_prefix		= args.comment_prefix,
										column_names_are_in	= args.column_names_are_in,
										only_column_names	= [args.unit_column, args.group_column],
										verbose				= (args.verbosity>=2),
										verbose_prefix		= "  ")
		if((unit_names_column:=get_column_index(column_names, found_column_names, args.unit_column, case_sensitive=args.case_sensitive))<0): abort("  ERROR: Missing column '%s' in groups table"%(args.unit_column))
		if((group_names_column:=get_column_index(column_names, found_column_names, args.group_column, case_sensitive=args.case_sensitive))<0): abort("  ERROR: Missing column '%s' in groups table"%(args.group_column))
		Kgroup_table = Kgroup_table[Kgroup_table[:,group_names_column]!="",:] # omit assignments with empty group name, presumably being unknowns
		if(args.verbosity>=2): print2("  Note: Loaded Kgroup assignments for %d units"%(Kgroup_table.shape[0]))
		# map group assignments to our existing unit inventory
		# Kgrouped_units[] will be an integer array with values in 0,..,Nunits-1, specifying for which units we have known group assignments for
		# Kgroup_assignments[] will be an integer array with values in 0,..,NKG-1, synchronized with Kgrouped_units[], specifying which group each of the assigned units belongs to
		# Kgroup2units[] will be a list of size NKG, with each element being an integer array with values in 0,..,Nunits-1, specifying the units in each Kgroup
		Kgroup_names 		= sorted(set(Kgroup_table[:,group_names_column]))
		name2Kgroup  		= {name:g for g,name in enumerate(Kgroup_names)}
		Kgrouped_units 		= numpy.asarray([unit_name2int.get(unit_name,-1) for unit_name in Kgroup_table[:,unit_names_column]], dtype=int)
		Kgroup_assignments	= numpy.asarray([name2Kgroup[group_name] for group_name in Kgroup_table[:,group_names_column]], dtype=int)
		# remove any unit->Kgroup assignments whose unit was not found in our inventory
		keep_assignments = numpy.where(Kgrouped_units>=0)[0]
		if(len(keep_assignments)<len(Kgrouped_units)):
			if(args.verbosity>=2): print2("  Note: Only keeping %d out of %d unit->Kgroup assignments, because %d encountered units were unknown"%(len(keep_assignments),len(Kgrouped_units),len(Kgrouped_units)-len(keep_assignments)))
			Kgroup_assignments = Kgroup_assignments[keep_assignments]
			Kgrouped_units 	= Kgrouped_units[keep_assignments]
		Kgroup2units = [Kgrouped_units[members] for members in group_integers(Kgroup_assignments)] # Kgroup2units[] will be a list of length NKG, with each element being an integer list with values in 0,..,Nunits-1, specifying which units were assigned to each Kgroup
		NKG = len(Kgroup2units)
		if(args.verbosity>=1): print2("  Note: Got %d Kgroups, with sizes ranging from %d to %d"%(NKG,min(len(members) for members in Kgroup2units),max(len(members) for members in Kgroup2units)))
		# use known group assignments to expand set of known classifications
		if(not args.reserve_Kgroups_for_tests):
			if(args.verbosity): print2("Expanding known classifications based on loaded Kgroup assignments..")
			derived_classifications 	= [(g1==g2) for u1,g1 in zip(Kgrouped_units,Kgroup_assignments) for u2,g2 in zip(Kgrouped_units,Kgroup_assignments) if ((not args.symmetrize) or (u1<=u2))]
			derived_classified_samples 	= [pair2sample[(u1,u2)] for u1 in Kgrouped_units for u2 in Kgrouped_units if ((not args.symmetrize) or (u1<=u2))]
			classified_samples, classifications = consolidate_mappings([(classified_samples, classifications), (derived_classified_samples,derived_classifications)], NX=Xdata.shape[0])
			if(args.verbosity>=2): print2("  Note: Now got known classifications for %d unit pairs"%(len(classified_samples)))
		

	# further filter the samples to make predictions for, if requested
# 	if(args.diarize=="classified"):
# 		if(len(classified_samples)<Xdata.shape[0]):
# 			if(args.verbosity>=1): print2("Note: Predictions will only be performed for already classified samples (%d out of %d)"%(len(classified_samples),Xdata.shape[0]))
# 			Xdata 				= Xdata[classified_samples,:]
# 			prediction_pairs	= [prediction_pairs[p] for p in classified_samples]
# 			classified_samples 	= range(Xdata.shape[0])
# 	elif(args.diarize=="nonclassified"):
# 		if(len(classified_samples)>0):
# 			nonclassified_samples = integer_complement(Xdata.shape[0], classified_samples)
# 			if(args.verbosity>=1): print2("Note: Predictions will only be performed for not-yet classified samples (%d out of %d)"%(len(nonclassified_samples),Xdata.shape[0]))
# 			Xdata 				= Xdata[nonclassified_samples, :]
# 			prediction_pairs	= [prediction_pairs[p] for p in nonclassified_samples]
# 			classified_samples	= []
# 			classifications		= numpy.empty(0, dtype=float)
# 	elif(args.diarize=="all"):
# 		pass
# 	if(Xdata.shape[0]==0): abort("ERROR: No samples left for predictions")
# 	pair2sample = {pair:p for p,pair in enumerate(prediction_pairs)}
		
		
	# at this point:
	#	Xdata is a 2D numeric array of size Nsamples x NX, listing predictor values for Nsamples unit pairs, where NX is the number of predictors considered by the classifier. The columns of Xdata are synchronized with classifier.Xnames.
	#	prediction_pairs is a list of length Nsamples, each element of which is an integer 2-tuple listing the integer IDs of the two units being compared
	#	pair2sample is a dictionary that maps integer 2-tuples (pairs) to the corresponding position in prediction_pairs[] and the corresponding row in Xdata[]
	#	unit_name2int is a dictionary mapping unit names to their integer IDs, which are indices in unit_names[]
	#	classifications is a 1D boolean array, listing binary classifications (ingroup True or False) for a subset of samples
	#	classified_samples is a 1D integer array synchronized with classifications, listing which samples the classifications are available for
	

	# Perform predictions
	Xdata0 = (Xdata - classifier.Xcenter)/classifier.Xscale # Apply rescalings prior to predictions with new data
	if(args.reclassify_knowns or (len(classified_samples)==0)):
		if(args.verbosity>=1): print2("Predicting for %d samples (pairs)%s.."%(Xdata.shape[0],(", including for %d samples with provided classification"%(len(classified_samples)) if (len(classified_samples)>0) else "")))
		predicted_classifications = (classifier.predict_proba(Xdata0)[:,1] if args.probabilistic else classifier.predict(Xdata0))
	else:
		nonclassified_samples	  = integer_complement(Xdata.shape[0], classified_samples)
		predicted_classifications = numpy.empty(Xdata.shape[0], dtype=(float if args.probabilistic else bool))
		if(len(nonclassified_samples)>0):
			if(args.verbosity>=1): print2("Predicting for %d out of %d samples (pairs), keeping known classifications for the remaining %d samples.."%(len(nonclassified_samples),Xdata.shape[0],len(classified_samples)))
			predicted_classifications[nonclassified_samples] = (classifier.predict_proba(Xdata0[nonclassified_samples,:])[:,1] if args.probabilistic else classifier.predict(Xdata0[nonclassified_samples,:]))
		predicted_classifications[classified_samples] = classifications
	del classifications, classified_samples, Xdata, Xdata0

	if(args.out_predicted_classifications!=""):
		if(args.verbosity>=1): print2("Saving predicted classifications..")
		predictions_table = numpy.column_stack(([unit_names[pair[0]] for pair in prediction_pairs],[unit_names[pair[1]] for pair in prediction_pairs],(predicted_classifications>=0.5).astype(int)))
		if(args.probabilistic): predictions_table = numpy.hstack((predictions_table, numpy.expand_dims(predicted_classifications,axis=1)))
		save_classical_table(output_file			= args.out_predicted_classifications,
							table					= predictions_table,
							column_names			= ["unit1","unit2","classification"]+(["classification_probability"] if args.probabilistic else []),
							delimiter				= args.delimiter,
							comment_prefix			= args.comment_prefix,
							comments				= join_nonempty(["Predicted classifications (ingroup=True=1, outgroup=False=0)\nGenerated on: %s\nUsed command: %s"%(get_date_time(),get_shell_command()), args.case_tag], sep="\n"))
			
	# at this point:
	#	prediction_pairs is a list of length Nsamples, each element of which is an integer 2-tuple listing the integer IDs of the two units being compared
	#	unit_name2int is a dictionary mapping unit names to their integer IDs, which are indices in unit_names[]
	#	predicted_classifications is a 1D boolean array of length Nsamples, listing predicted classifications (binary or probabilistic)

	

	##########################################################
	# Diariation/grouping based on pairwise classifications
	

	# Symmetrize predicted classifications if needed (i.e., such that (a,b) has the same classification as (b,a)) and remove redundant pairs (i.e., only keep pairs (a,b) for which a<=b.
	# Hence, prediction_pairs will be such that every element (a,b) satisfies a<=b, implicitly assuming the same classification for the transposed (b,a).
	# This is because our diarization/group counting algorithms only make sense if the affinity matrix is symmetric
	# pair2sample will map both (a,b) and (b,a) to the same sample index
	if(args.verbosity>=1): print2("Symmetrizing classifications and removing redundant pairs..")
	prediction_pairs, predicted_classifications, pair2sample = symmetrize_data(pairs=prediction_pairs, values=predicted_classifications, pair2index = pair2sample)

	# auxiliary function for getting the classifications for multiple unit pairs forming a row in the affinity matrix, i.e. (a,b1), (a,b2), .., (a,bn)
	def pairs2classifications(a,bs):
		samples = numpy.asarray([pair2sample.get((a,b),-1) for b in bs], dtype=int)
		classifications = numpy.empty(len(bs), dtype=(float if args.probabilistic else bool))
		classifications[samples>=0] = predicted_classifications[samples[samples>=0]]
		if(args.missing_pairs=="ingroup"):
			classifications[samples<0] = 1
		else:
			classifications[samples<0] = 0
		classifications[numpy.asarray(bs)==a] = 1 # self-comparisons are always ingroup
		return classifications
			

	if((args.counting_method!="none") or (args.test_counting_methods!="") or (args.grouping_method!="none") or (args.plot_affinity_matrix!="")):
		# check if any pairs are missing
		missing = [(a,b) for a in range(Nunits) for b in range(a,Nunits) if ((a,b) not in pair2sample)]
		if(len(missing)>0): 
			if(args.missing_pairs=="error"):
				abort("ERROR: Missing classification for %d pairs, for example (%s, %s)"%(len(missing),unit_names[missing[0][0]],unit_names[missing[0][1]]))
			elif(args.verbosity>=1):
				print2("WARNING: %d out of %d unit pairs have unknown/unpredicted classification, so they will be assumed to be %ss"%(len(missing),Nunits*(Nunits+1)/2,args.missing_pairs))

		# construct full affinity matrix if needed (2D array of size Nunits x Nunits)
		# Note that prediction_pairs by construction only cover the upper-triangular portion of the affinity matrix
		if(args.verbosity): print2("Constructing full affinity matrix..")
		affinity_matrix = numpy.full((Nunits,Nunits), (1 if (args.missing_pairs=="ingroup") else 0), dtype=float)
		rows, columns = [pair[0] for pair in prediction_pairs], [pair[1] for pair in prediction_pairs]
		affinity_matrix[rows,columns] = predicted_classifications
		affinity_matrix[columns,rows] = predicted_classifications
		affinity_matrix[range(Nunits),range(Nunits)] = 1 # each unit must be in the same group as itself
			

	if(args.plot_affinity_matrix!=""):
		if(args.verbosity): print2("Plotting %d x %d affinity matrix.."%(Nunits,Nunits))
		plot_heatmap(table				= affinity_matrix,
					show_colorbar		= False,
					colormap			= "Greys",
					colormap_min		= 0,
					colormap_max		= 1,
					xlabel				= args.unit_type,
					ylabel				= args.unit_type,
					cluster				= "rows_and_columns_joint",
					plot_title			= join_nonempty(["Adjacency matrix between %d units"%(Nunits),args.case_tag], sep="\n"),
					plot_width			= args.plot_width,
					plot_height			= args.plot_height,
					show_grid			= False,
					plot_target			= args.plot_affinity_matrix)
					
					
	# define auxiliary function for estimating groups based on an affinity (aka. adjacency) matrix or sub-matrix
	# group members will be indices with respect to the provided subaffinity_matrix
	# If count_only, this function returns a single integer (number of groups)
	# If count_only==False, this function returns two synchronized lists: group_reps & group_members
	def aux_estimate_groups(method, subaffinity_matrix, count_only):
		if(numpy.all(subaffinity_matrix>=0.9)): 
			# all units are very close to each other, so they all must be in the same group
			if(count_only):
				return 1
			else:
				return [0], [list(range(subaffinity_matrix.shape[0]))]
		elif(method=="greedy_seeds"):
			# use the greedy clustering algorithm to construct groups
			group_reps    = []
			group_members = []
			for n in RNG.permutation(subaffinity_matrix.shape[0]):
				affinities = subaffinity_matrix[n,group_reps]
				nearest_group = get_largest_item(affinities, check_nan=False)
				if((nearest_group>=0) and (affinities[nearest_group]>=0.5)):
					# add this unit as member to the closest group
					if(not count_only): group_members[nearest_group].append(n)
				else:
					# create a new group with this unit as representative (seed)
					group_reps.append(n)
					if(not count_only): group_members.append([n])
			if(count_only):
				return len(group_reps)
			else:
				return group_reps, group_members
		elif(method=="AffinityPropagation"):
			# use AffinityPropagation algorithm. Note that for some strange reason one gets much better results when passing an upper-triangular version of the affinity matrix, rather than the full affinity matrix
			clustering = sklearn.cluster.AffinityPropagation(affinity="precomputed",max_iter=10000, random_state=args.seed).fit(numpy.triu(subaffinity_matrix))
			if(count_only):
				return len(clustering.cluster_centers_indices_)			
			else:
				group_reps = clustering.cluster_centers_indices_
				return group_reps, group_integers(clustering.labels_, NG=len(group_reps))
		elif(method=="DBSCAN"):
			clustering = sklearn.cluster.DBSCAN(eps=0.9, metric="precomputed",n_jobs=-1).fit(1-subaffinity_matrix)
			if(count_only):
				return 1+max(clustering.labels_)
			else:
				group_members = group_integers(clustering.labels_, NG=None)
				group_reps	  = [gm[0] for gm in group_members]
				return group_reps, group_members
		elif(method=="HDBSCAN"):
			clustering = sklearn.cluster.HDBSCAN(metric="precomputed",n_jobs=-1).fit(1-subaffinity_matrix)
			if(count_only):
				return 1+max(clustering.labels_)
			else:
				group_members = group_integers(clustering.labels_, NG=None)
				group_reps	  = [gm[0] for gm in group_members]
				return group_reps, group_members
		elif(method.startswith("greedy_groups_refined")):
			# expect method to be either 'greedy_groups_refined' or 'greedy_groups_refined_'<affinity_threshold>
			if(method=="greedy_groups_refined"): 
				affinity_threshold = 0.5
			else: 
				affinity_threshold = float_or_nan(method.rsplit("_",1)[1])
				if(not numpy.isfinite(affinity_threshold)): abort("Invalid affinity threshold specified in grouping method '%s'"%(method))
			grouping = CPP.cluster_greedy_refined(Nnodes=subaffinity_matrix.shape[0], affinity_matrix=subaffinity_matrix.flatten(), count_only=count_only, random_seed=RNG.integers(0,100000,size=1)[0], max_iterations=100, affinity_threshold=affinity_threshold)
			if(count_only):
				return grouping["Ngroups"]
			else:
				group_members = grouping["group_members"]
				group_reps = [group[0] for group in group_members]
				return group_reps, group_members
			
			# OLD CODE. Pure python, so slower.
# 			group_members = []
# 			for n in RNG.permutation(subaffinity_matrix.shape[0]):
# 				# check if this unit matches any of the existing groups (in terms of its average affinity), and add it to the closest group if applicable
# 				affinities = [numpy.mean(subaffinity_matrix[n,group]) for group in group_members] # affinities[g] is the average affinity of the focal unit to group g
# 				nearest_group = get_largest_item(affinities, check_nan=False)
# 				if((nearest_group>=0) and (affinities[nearest_group]>=affinity_threshold)):
# 					# add this unit to the closest group
# 					group_members[nearest_group].append(n)
# 				else:
# 					# none of the existing groups are close to this unit, so create a new group
# 					group_members.append([n])
# 			member2group = map_members_to_groups(group_members, Nmembers=subaffinity_matrix.shape[0], unique=True)
# 			# refine groups, by re-evaluating unit-to-group affinities using the current group configurations, and merging groups that appear close to each other
# 			for iteration in range(100):
# 				new_group_members = [[] for g in range(len(group_members))] # reserve space for revised groups, based on how many groups we got in the first round
# 				for n in RNG.permutation(subaffinity_matrix.shape[0]):
# 					affinities = [mean_without(N=len(group), Xmean=numpy.mean(subaffinity_matrix[n,group]), Xomitted=([1] if (member2group[n]==g) else [])) for g,group in enumerate(group_members)]\
# 								  + [numpy.mean(subaffinity_matrix[n,group]) for group in new_group_members[len(group_members):]]
# 					nearest_group = get_largest_item(affinities, check_nan=True)
# 					if((nearest_group>=0) and (affinities[nearest_group]>=affinity_threshold)):
# 						# add this unit to the closest group
# 						new_group_members[nearest_group].append(n)
# 					else:
# 						# none of the existing groups are close to this unit, so create a new group
# 						new_group_members.append([n])
# 				# replace old groups with new groups, omitting groups that don't have any members left
# 				group_members = list(filter(len,new_group_members))
# 				# check if we should merge some groups, using an algorithm similar to "greedy_groups", but applied at the group level
# 				group_mergings = []
# 				for g, group in enumerate(group_members):
# 					affinities = [numpy.mean(subaffinity_matrix[numpy.ix_(group,[m for g2 in group_merging for m in group_members[g2]])]) for group_merging in group_mergings] # affinities[gm] is the average affinity of the focal group to all members of all groups in group merging gm
# 					nearest_group_merging = get_largest_item(affinities, check_nan=False)
# 					if((nearest_group_merging>=0) and (affinities[nearest_group_merging]>=affinity_threshold)):
# 						# add this group to the closest group merging
# 						group_mergings[nearest_group_merging].append(g)
# 					else:
# 						# none of the existing group mergings are close to this group, so create a new group merging
# 						group_mergings.append([g])
# 				# construct new groups based on the group mergings, i.e. all groups from a given group-merging will be merged into a single group
# 				group_members = [[m for g in group_merging for m in group_members[g]] for group_merging in group_mergings]
# 				new_member2group = map_members_to_groups(group_members, Nmembers=subaffinity_matrix.shape[0], unique=True)
#				if(numpy.all(new_member2group==member2group)): break # algorithm converged
# 				member2group = new_member2group
# 			if(count_only):
# 				return len(group_members)
# 			else:
# 				group_reps = [group[0] for group in group_members]
# 				return group_reps, group_members

		elif(method.startswith("greedy_groups")):
			if(method=="greedy_groups"): 
				affinity_threshold = 0.5
			else: 
				affinity_threshold = float_or_nan(method.rsplit("_",1)[1])
				if(not numpy.isfinite(affinity_threshold)): abort("Invalid affinity threshold specified in grouping method '%s'"%(method))
			group_members = []
			for n in RNG.permutation(subaffinity_matrix.shape[0]):
				# check if this unit matches any of the existing groups (in terms of its average affinity), and add it to the closest group if applicable
				affinities = [numpy.mean(subaffinity_matrix[n,group]) for group in group_members] # affinities[g] is the average affinity of the focal unit to group g
				nearest_group = get_largest_item(affinities, check_nan=False)
				if((nearest_group>=0) and (affinities[nearest_group]>=affinity_threshold)):
					# add this unit to the closest cluster
					group_members[nearest_group].append(n)
				else:
					# none of the existing group are close to this unit, so create a new group
					group_members.append([n])
			if(count_only):
				return len(group_members)
			else:
				group_reps = [group[0] for group in group_members]
				return group_reps, group_members

					
	# define auxiliary function for estimating the number of groups spanned by all or a subset of units
	def aux_estimate_Ngroups(method, unit_subset=None):
		subaffinity_matrix = (affinity_matrix if (unit_subset is None) else affinity_matrix[numpy.ix_(unit_subset,unit_subset)])
		if(method.startswith("greedy")):
			Ngroups = numpy.mean(values:=[aux_estimate_groups(method=method, subaffinity_matrix=subaffinity_matrix, count_only=True) for r in range(args.greedy_counting_Nrepeats)])
		else:
			Ngroups = aux_estimate_groups(method=method, subaffinity_matrix=subaffinity_matrix, count_only=True)
		return Ngroups


	if(args.test_counting_methods!=""):
		if(args.verbosity>=1): 
			print2("Testing the accuracy of group counting methods (%s) using the loaded Kgroup assignments.."%(", ".join(test_counting_methods)))
			if(not (args.reserve_Kgroups_for_tests or args.reclassify_knowns)): print2("WARNING: Loaded Kgroup assignments were also used to supplement the predicted classifications, so test results may be unreliable. Consider using --reserve_Kgroups_for_tests.")
		# prepare test sets
		# counting_tests_true_NKG[], counting_tests_Kgroup_subsets[], counting_tests_unit_subsets[] and counting_tests_Nunits[] will be synchronized lists or arrays, each of length args.counting_Ntests
		counting_tests_true_NKG 	  = numpy.linspace(1,NKG,num=args.counting_Ntests).astype(int) # number of Kgroups to consider for the counting tests
		counting_tests_Kgroup_subsets = [random.sample(range(NKG), k=true_NKG) for true_NKG in counting_tests_true_NKG] # counting_tests_Kgroup_subsets[ct] specifies a subset of Kgroups, whose units will be considered in the ct-th test
		counting_tests_unit_subsets	  = [[unit for g in Kgroup_subset for unit in Kgroup2units[g]] for Kgroup_subset in counting_tests_Kgroup_subsets] # counting_tests_unit_subsets[ct] specifies the subset of units considered in the ct-th test
		counting_tests_Nunits		  = [len(unit_subset) for unit_subset in counting_tests_unit_subsets]
		# omit very small datasets, as this causes issues with some of the clustering algorithms
		valids = [t for t,Nunits in enumerate(counting_tests_Nunits) if (Nunits>=5)]
		if(len(valids)<args.counting_Ntests):
			print2("  Note: Omitting %d out of %d tests due to insufficient number of included units"%(args.counting_Ntests-len(valids),args.counting_Ntests))
			counting_tests_true_NKG 		= counting_tests_true_NKG[valids]
			counting_tests_Kgroup_subsets 	= [counting_tests_Kgroup_subsets[t] for t in valids]
			counting_tests_unit_subsets		= [counting_tests_unit_subsets[t] for t in valids]
			counting_tests_Nunits			= [counting_tests_Nunits[t] for t in valids]
			args.counting_Ntests			= len(valids)
		# prepare array for storing test results (predicted group counts)
		counting_tests_predicted_NKG = numpy.empty((len(counting_tests_true_NKG),len(test_counting_methods)), dtype=float)
		counting_tests_MREs			 = numpy.empty(len(test_counting_methods), dtype=float)
		counting_tests_R2s			 = numpy.empty(len(test_counting_methods), dtype=float)
		if(args.plot_counting_test_results!=""): plot_target = PdfPages(args.plot_counting_test_results)
		for m,method in enumerate(test_counting_methods):
			if(args.verbosity>=2): print2("  Testing the accuracy of group-counting method '%s'.."%(method))
			start_time						  = time.time()
			counting_tests_predicted_NKG[:,m] = run_parallel(function=(lambda unit_subset: aux_estimate_Ngroups(method=method, unit_subset=unit_subset)), Nthreads=args.Nthreads, arguments=counting_tests_unit_subsets, fast=False, randomize=True)
			counting_tests_MREs[m] 			  = numpy.median(numpy.abs(counting_tests_predicted_NKG[:,m]-counting_tests_true_NKG)/counting_tests_true_NKG) # median relative error
			counting_tests_R2s[m]			  = 1 - numpy.mean((counting_tests_predicted_NKG[:,m]-counting_tests_true_NKG)**2)/numpy.nanvar(counting_tests_true_NKG) # fraction of variance explained by the predictions
			if(args.verbosity>=3): print2("    --> MRE = %.4g, R2 = %.4g, runtime = %g sec"%(counting_tests_MREs[m],counting_tests_R2s[m],time.time()-start_time))
			# plot predicted vs true number of groups
			if(args.plot_counting_test_results!=""):
				plot_xy(plot_target			= plot_target,
						scatter				= [[counting_tests_true_NKG, counting_tests_predicted_NKG[:,m]]],
						plot_minx			= 0,
						plot_miny			= 0,
						xlabel				= "true number of %ss"%(args.group_type),
						ylabel				= "predicted number of %ss"%(args.group_type),
						show_diagonal		= True,
						show_correlations	= "Spearman",
						show_grid			= True,			# whether to show vertical and horizontal grid lines
						plot_title			= join_nonempty(["True vs predicted number of %ss, method '%s'\nbased on %s classifications"%(args.group_type,method,("probabilistic" if args.probabilistic else "binary")), args.case_tag, "MRE = %.4g, R2 = %.4g, N = %d"%(counting_tests_MREs[m],counting_tests_R2s[m],len(counting_tests_true_NKG))], sep="\n"),
						plot_width			= args.plot_width,
						plot_height			= args.plot_height)
		if(args.plot_counting_test_results!=""): plot_target.close()
		# determine the most accurate counting method, in terms of various metrics
		best_method_MRE = numpy.argmin(counting_tests_MREs)
		if(args.verbosity>=2): print2("  Note: Most accurate group-counting method (in terms of median relative error) was %s (MRE=%.3g)"%(test_counting_methods[best_method_MRE],counting_tests_MREs[best_method_MRE]))
		best_method_R2 = numpy.argmax(counting_tests_R2s)
		if(args.verbosity>=2): print2("  Note: Most accurate group-counting method (in terms of R2) was %s (R2=%.3g)"%(test_counting_methods[best_method_R2],counting_tests_R2s[best_method_R2]))
		
		# save counting test results to a tabular file
		if(args.out_counting_test_results!=""):
			if(args.verbosity): print2("Saving test results for all %d examined group-counting methods to file.."%(len(test_counting_methods)))
			counting_tests_true_NKG = numpy.expand_dims(counting_tests_true_NKG,axis=1)
			save_classical_table(output_file			= args.out_counting_test_results,
								table					= numpy.hstack((counting_tests_true_NKG,numpy.expand_dims(counting_tests_Nunits,axis=1),counting_tests_predicted_NKG,(counting_tests_predicted_NKG-counting_tests_true_NKG)/counting_tests_true_NKG)),
								column_names			= ["true_Ngroups","Nunits"] + ["predicted_Ngroups."+method for method in test_counting_methods] + ["relative_error."+method for method in test_counting_methods],
								delimiter				= args.delimiter,
								comment_prefix			= args.comment_prefix,
								comments				= join_nonempty(["Test results for %d group-counting methods (%s)\nGenerated on: %s\nUsed command: %s"%(len(test_counting_methods),", ".join(test_counting_methods),get_date_time(),get_shell_command()), args.case_tag], sep="\n"))


	if(args.counting_method!="none"):
		if(args.verbosity>=1): print2("Estimating the number of groups among %d units based on known and/or predicted %s classifications (method '%s').."%(len(unit_names),("probabilistic" if args.probabilistic else "binary"),args.counting_method))
		Ngroups = aux_estimate_Ngroups(method=args.counting_method)
		if(args.verbosity>=1): print2("  --> Estimated Ngroups = %g"%(Ngroups))


	if(args.grouping_method!="none"):
		if(args.verbosity>=1): print2("Estimating groups among %d units based on known and/or predicted %s classifications (method '%s').."%(len(unit_names),("probabilistic" if args.probabilistic else "binary"),args.grouping_method))
		group_reps, group_members = aux_estimate_groups(method=args.grouping_method, subaffinity_matrix=affinity_matrix, count_only=False)
		group_sizes = [len(gm) for gm in group_members]
		group_internal_affinities = [numpy.mean(affinity_matrix[numpy.ix_(gm,gm)]) for gm in group_members]
		Ngroups = len(group_reps)
		if(args.verbosity>=1): print2("  --> Estimated %d groups, sizes ranging from %d to %d members, interval affinities range from %g to %g"%(Ngroups,min(group_sizes),max(group_sizes),min(group_internal_affinities),max(group_internal_affinities)))
		if(args.group_naming=="unit_rep"):
			group_names = [unit_names[u] for u in group_reps]
		elif(args.group_naming=="enumerate"):
			group_names = ["%s%s"%(args.group_type,int2zero_padded_str(g,Ngroups)) for g in range(Ngroups)]

		if(args.out_group2units!=""):
			if(args.verbosity>=1): print2("Saving group2units mapping to file..")
			save_classical_table(output_file			= args.out_group2units,
								table					= numpy.column_stack((group_names,group_internal_affinities,group_sizes,[args.group_member_separator.join(unit_names[u] for u in gm) for gm in group_members])),
								column_names			= ["representative","mean_internal_affinity","Nmembers","members"],
								delimiter				= args.delimiter,
								comment_prefix			= args.comment_prefix,
								comments				= join_nonempty(["%s-to-%ss (group-to-units) mapping, for %d estimated %ss among %d %ss\nGenerated on: %s\nUsed command: %s"%(args.group_type,args.unit_type,Ngroups,args.group_type,Nunits,args.unit_type,get_date_time(),get_shell_command()), args.case_tag], sep="\n"))
								
		if(args.out_unit2group!=""):
			if(args.verbosity>=1): print2("Saving unit2group mapping to file..")
			unit2group = map_members_to_groups(groups=group_members, Nmembers=Nunits)
			save_classical_table(output_file			= args.out_unit2group,
								table					= numpy.column_stack((unit_names,[group_names[g[0]] for g in unit2group])),
								column_names			= [args.unit_type,args.group_type],
								delimiter				= args.delimiter,
								comment_prefix			= args.comment_prefix,
								comments				= join_nonempty(["%s-to-%s (unit-to-group) mapping, for %d estimated %ss among %d %ss\nGenerated on: %s\nUsed command: %s"%(args.unit_type,args.group_type,Ngroups,args.group_type,Nunits,args.unit_type,get_date_time(),get_shell_command()), args.case_tag], sep="\n"))

